MODULE dynnxt_tam
#ifdef key_tam
   !!======================================================================
   !!                       ***  MODULE  dynnxt_tam  ***
   !! Ocean dynamics: time stepping
   !!                 Tangent and Adjoint Module
   !!======================================================================
   
   !!----------------------------------------------------------------------
   !!   dyn_nxt_tan  : update the horizontal velocity from the momentum trend
   !!   dyn_nxt_adj  : update the horizontal velocity from the momentum trend
   !!----------------------------------------------------------------------
   !! * Modules used
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 & 
      & jpkm1,               & 
      & jpiglo
   USE oce_tam       , ONLY: & ! ocean dynamics and tracers
      & un_tl,               &
      & vn_tl,               &
      & ub_tl,               &
      & vb_tl,               &
      & ua_tl,               &
      & va_tl,               &
      & un_ad,               &
      & vn_ad,               &
      & ub_ad,               &
      & vb_ad,               &
      & ua_ad,               &
      & va_ad
   USE dom_oce       , ONLY: & ! ocean space and time domain
      & umask,               &
      & vmask,               &
      & lk_vvl,              &
      & neuler,              &
      & rdt,                 & 
      & atfp,                & 
      & atfp1,               & 
      & e1u,                 &
      & e2u,                 &
      & e1v,                 &
      & e2v,                 &
#if defined key_zco
      & e3t_0,               &
#else
      & e3v,                 &
      & e3u,                 &
#endif
      & mig,                 &
      & mjg,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej   
   USE in_out_manager, ONLY: & ! I/O manager
      & lwp,                 &
      & nit000,              &
      & nitend,              &
      & numout,              &
      & ctl_stop
   USE dynspg_oce    , ONLY: & ! type of surface pressure gradient
      & lk_dynspg_flt
   USE lbclnk        , ONLY: & ! lateral boundary condition (or mpp link)
      & lbc_lnk
   USE lbclnk_tam    , ONLY: & ! lateral boundary condition (or mpp link)
      & lbc_lnk_adj
   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld,     ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE tstool_tam    , ONLY: &
      & prntst_adj,          & !
                               ! random field standard deviation for:
      & stdu,                & !   u-velocity
      & stdv                   !   v-velocity

!VERIF
!   USE bdy_oce         ! unstructured open boundary conditions
!   USE bdydta          ! unstructured open boundary conditions
!   USE bdydyn          ! unstructured open boundary conditions
!!!!!

   IMPLICIT NONE
   PRIVATE

   !! * Accessibility
   PUBLIC dyn_nxt_tan            ! routine called by step.F90
   PUBLIC dyn_nxt_adj            ! routine called by step.F90
   PUBLIC dyn_nxt_adj_tst        ! routine called by step.F90
   !! * Substitutions
#  include "domzgr_substitute.h90"
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE dyn_nxt_tan ( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_nxt_tan  ***
      !!                   
      !! ** Purpose of the direct routine:
      !!      Compute the after horizontal velocity from the 
      !!      momentum trend.
      !!
      !! ** Method  :   Apply lateral boundary conditions on the trends (ua,va) 
      !!      through calls to routine lbc_lnk.
      !!      After velocity is compute using a leap-frog scheme environment:
      !!         (ua,va) = (ub,vb) + 2 rdt (ua,va)
      !!      Note that if lk_dynspg_flt=T, the time stepping has already been
      !!      performed in dynspg module
      !!      Time filter applied on now horizontal velocity to avoid the
      !!      divergence of two consecutive time-steps and swap of dynamics
      !!      arrays to start the next time step:
      !!         (ub,vb) = (un,vn) + atfp [ (ub,vb) + (ua,va) - 2 (un,vn) ]
      !!         (un,vn) = (ua,va) 
      !!
      !! ** Action : - Update ub,vb arrays, the before horizontal velocity
      !!             - Update un,vn arrays, the now horizontal velocity
      !!
      !! History of the direct routine:
      !!        !  87-02  (P. Andrich, D. L Hostis)  Original code
      !!        !  90-10  (C. Levy, G. Madec)
      !!        !  93-03  (M. Guyon)  symetrical conditions
      !!        !  97-02  (G. Madec & M. Imbard)  opa, release 8.0
      !!        !  97-04  (A. Weaver)  Euler forward step
      !!        !  97-06  (G. Madec)  lateral boudary cond., lbc routine
      !!   8.5  !  02-08  (G. Madec)  F90: Free form and module
      !!        !  02-10  (C. Talandier, A-M. Treguier) Open boundary cond.
      !!   9.0  !  05-11  (V. Garnier) Surface pressure gradient organization
      !!        !  07-07  (D. Storkey) Calls to BDY routines. 
      !! History of the TAM routine:
      !!   9.0  !  08-06  (A. Vidard) Skeleton
      !!        !  08-08  (A. Vidard) tangent of the 05-11 version
      !!        !  08-08  (A. Vidard) tangent of the 07-07 version
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt      ! ocean time-step index

      !! * Local declarations
      INTEGER  ::   ji, jj, jk   ! dummy loop indices
      REAL(wp) ::   z2dt         ! temporary scalar
      REAL(wp) ::   zsshun1, zsshvn1         ! temporary scalar
      !!----------------------------------------------------------------------
      !!  OPA 9.0 , LOCEAN-IPSL (2005) 
      !! $Header$ 
      !! This software is governed by the CeCILL licence see modipsl/doc/NEMO_CeCILL.txt 
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn_nxt_tan : time stepping'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~'
      ENDIF

      ! Local constant initialization
      z2dt = 2. * rdt
      IF( neuler == 0 .AND. kt == nit000 )  z2dt = rdt

      !! Explicit physics with thickness weighted updates
      IF( lk_vvl .AND. .NOT. lk_dynspg_flt ) THEN
         CALL ctl_stop ('vvl not available for tangent')
      ENDIF

      ! Lateral boundary conditions on ( ua, va )
      CALL lbc_lnk( ua_tl, 'U', -1. )
      CALL lbc_lnk( va_tl, 'V', -1. )

      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
         ! Next velocity
         ! -------------
#if defined key_dynspg_flt
         ! Leap-frog time stepping already done in dynspg.F routine
#else
         DO jj = 1, jpj                      ! caution: don't use (:,:) for this loop
            DO ji = 1, jpi                   ! it causes optimization problems on NEC in auto-tasking
               ! Leap-frog time stepping
               ua_tl(ji,jj,jk) = ( ub_tl(ji,jj,jk) + z2dt * ua_tl(ji,jj,jk) ) * umask(ji,jj,jk)
               va_tl(ji,jj,jk) = ( vb_tl(ji,jj,jk) + z2dt * va_tl(ji,jj,jk) ) * vmask(ji,jj,jk)
            END DO
         END DO

         IF( lk_vvl ) THEN
            CALL ctl_stop ('key_vvl not available yet for tangent')
         ENDIF
# if defined key_obc
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============
      CALL ctl_stop ('key_obc not available yet for tangent')


      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
# elif defined key_bdy 
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============
      CALL ctl_stop ('key_bdy not available yet for tangent')

      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
# endif
# if defined key_agrif
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============
      CALL ctl_stop ('key_agrif not available yet for tangent')
      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
# endif
#endif

         ! Time filter and swap of dynamics arrays
         ! ------------------------------------------
         IF( neuler == 0 .AND. kt == nit000 ) THEN
            IF( (lk_vvl .AND. .NOT. lk_dynspg_flt) ) THEN      ! Varying levels
               CALL ctl_stop ('vvl not available for tangent')
            ELSE                                               ! Fixed levels
               DO jj = 1, jpj
                  DO ji = 1, jpi
                     ! Euler (forward) time stepping
                     ub_tl(ji,jj,jk) = un_tl(ji,jj,jk)
                     vb_tl(ji,jj,jk) = vn_tl(ji,jj,jk)
                     un_tl(ji,jj,jk) = ua_tl(ji,jj,jk)
                     vn_tl(ji,jj,jk) = va_tl(ji,jj,jk)
                  END DO
               END DO
            ENDIF
         ELSE
            IF( (lk_vvl .AND. .NOT. lk_dynspg_flt) ) THEN      ! Varying levels
               CALL ctl_stop ('vvl not available for tangent')
            ELSE                                               ! Fixed levels
               DO jj = 1, jpj
                  DO ji = 1, jpi
                     ! Leap-frog time stepping
                     ub_tl(ji,jj,jk) = atfp  * ( ub_tl(ji,jj,jk) + ua_tl(ji,jj,jk) ) + atfp1 * un_tl(ji,jj,jk)
                     vb_tl(ji,jj,jk) = atfp  * ( vb_tl(ji,jj,jk) + va_tl(ji,jj,jk) ) + atfp1 * vn_tl(ji,jj,jk)
                     un_tl(ji,jj,jk) = ua_tl(ji,jj,jk)
                     vn_tl(ji,jj,jk) = va_tl(ji,jj,jk)
                  END DO
               END DO
            ENDIF
         ENDIF
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============

#if defined key_agrif
      CALL ctl_stop ('key_agrif not available yet for tangent')
      
!      IF (.NOT.Agrif_Root())    CALL Agrif_Update_Dyn( kt )
#endif      

   END SUBROUTINE dyn_nxt_tan
   SUBROUTINE dyn_nxt_adj ( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_nxt_adj  ***
      !!                   
      !! ** Purpose of the direct routine:
      !!      Compute the after horizontal velocity from the 
      !!      momentum trend.
      !!
      !! ** Method  :   Apply lateral boundary conditions on the trends (ua,va) 
      !!      through calls to routine lbc_lnk.
      !!      After velocity is compute using a leap-frog scheme environment:
      !!         (ua,va) = (ub,vb) + 2 rdt (ua,va)
      !!      Note that if lk_dynspg_flt=T, the time stepping has already been
      !!      performed in dynspg module
      !!      Time filter applied on now horizontal velocity to avoid the
      !!      divergence of two consecutive time-steps and swap of dynamics
      !!      arrays to start the next time step:
      !!         (ub,vb) = (un,vn) + atfp [ (ub,vb) + (ua,va) - 2 (un,vn) ]
      !!         (un,vn) = (ua,va) 
      !!
      !! ** Action : - Update ub,vb arrays, the before horizontal velocity
      !!             - Update un,vn arrays, the now horizontal velocity
      !!
      !! History of the direct routine:
      !!        !  87-02  (P. Andrich, D. L Hostis)  Original code
      !!        !  90-10  (C. Levy, G. Madec)
      !!        !  93-03  (M. Guyon)  symetrical conditions
      !!        !  97-02  (G. Madec & M. Imbard)  opa, release 8.0
      !!        !  97-04  (A. Weaver)  Euler forward step
      !!        !  97-06  (G. Madec)  lateral boudary cond., lbc routine
      !!   8.5  !  02-08  (G. Madec)  F90: Free form and module
      !!        !  02-10  (C. Talandier, A-M. Treguier) Open boundary cond.
      !!   9.0  !  05-11  (V. Garnier) Surface pressure gradient organization
      !! History of the TAM routine:
      !!   9.0  !  08-06  (A. Vidard) Skeleton
      !!        !  08-08  (A. Vidard) Adjoint of the 02-10 version
      !!        !  08-11  (A. Vidard) Adjoint of the 05-12 version
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt      ! ocean time-step index

      !! * Local declarations
      !! * Local declarations
      INTEGER  ::   ji, jj, jk   ! dummy loop indices
      REAL(wp) ::   z2dt         ! temporary scalar

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn_nxt_adj : time stepping'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~'
      ENDIF

      ! Local constant initialization
      z2dt = 2. * rdt
      IF( neuler == 0 .AND. kt == nit000 )  z2dt = rdt

#if defined key_agrif
      CALL ctl_stop ('key_agrif not available yet for tangent')
      
!      IF (.NOT.Agrif_Root())    CALL Agrif_Update_Dyn( kt )
#endif      


      !                                                ! ===============
      DO jk = jpkm1, 1, -1                             ! Horizontal slab
         !                                             ! ===============
         ! Next velocity
         ! -------------

         ! Time filter and swap of dynamics arrays
         ! ------------------------------------------
         IF( neuler == 0 .AND. kt == nit000 ) THEN
            IF( (lk_vvl .AND. .NOT. lk_dynspg_flt) ) THEN      ! Varying levels
               CALL ctl_stop ('vvl not available for tangent')
            ELSE                                               ! Fixed levels
               DO jj = 1, jpj
                  DO ji = 1, jpi
                     ! Euler (forward) time stepping
                     ua_ad(ji,jj,jk) = ua_ad(ji,jj,jk) + un_ad(ji,jj,jk)
                     va_ad(ji,jj,jk) = va_ad(ji,jj,jk) + vn_ad(ji,jj,jk)
                     un_ad(ji,jj,jk) = 0.0_wp
                     vn_ad(ji,jj,jk) = 0.0_wp
                     un_ad(ji,jj,jk) = un_ad(ji,jj,jk) + ub_ad(ji,jj,jk)
                     vn_ad(ji,jj,jk) = vn_ad(ji,jj,jk) + vb_ad(ji,jj,jk)
                     ub_ad(ji,jj,jk) = 0.0_wp
                     vb_ad(ji,jj,jk) = 0.0_wp
                  END DO
               END DO
            ENDIF
         ELSE
            IF( (lk_vvl .AND. .NOT. lk_dynspg_flt) ) THEN      ! Varying levels
               CALL ctl_stop ('vvl not available for tangent')
            ELSE                                               ! Fixed levels
               DO jj = 1, jpj
                  DO ji = 1, jpi
                     ! Leap-frog time stepping
                     ua_ad(ji,jj,jk) = ua_ad(ji,jj,jk) + un_ad(ji,jj,jk)
                     un_ad(ji,jj,jk) = 0.0_wp
                     va_ad(ji,jj,jk) = va_ad(ji,jj,jk) + vn_ad(ji,jj,jk)
                     vn_ad(ji,jj,jk) = 0.0_wp
                     ua_ad(ji,jj,jk) = ua_ad(ji,jj,jk) + atfp  * ub_ad(ji,jj,jk)
                     un_ad(ji,jj,jk) = un_ad(ji,jj,jk) + atfp1 * ub_ad(ji,jj,jk)
                     ub_ad(ji,jj,jk) =                   atfp  * ub_ad(ji,jj,jk)

                     va_ad(ji,jj,jk) = va_ad(ji,jj,jk) + atfp  * vb_ad(ji,jj,jk)
                     vn_ad(ji,jj,jk) = vn_ad(ji,jj,jk) + atfp1 * vb_ad(ji,jj,jk)
                     vb_ad(ji,jj,jk) =                   atfp  * vb_ad(ji,jj,jk)

                  END DO
               END DO
            ENDIF
         ENDIF

#if defined key_dynspg_flt
         ! Leap-frog time stepping already done in dynspg.F routine
#else
# if defined key_agrif
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============
      CALL ctl_stop ('key_agrif not available yet for adjoint')
      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
# endif
# if defined key_bdy 
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============
      CALL ctl_stop ('key_bdy not available yet for adjoint')

      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
# elif defined key_obc
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============
      CALL ctl_stop ('key_obc not available yet for adjoint')


      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
# endif
         IF( lk_vvl ) THEN
            CALL ctl_stop ('key_vvl not available yet for adjoint')
         ENDIF

         DO jj = 1, jpj                      ! caution: don't use (:,:) for this loop
            DO ji = 1, jpi                   ! it causes optimization problems on NEC in auto-tasking
               ! Leap-frog time stepping
               ub_ad(ji,jj,jk) = ub_ad(ji,jj,jk) + ua_ad(ji,jj,jk) * umask(ji,jj,jk)
               ua_ad(ji,jj,jk) =            z2dt * ua_ad(ji,jj,jk) * umask(ji,jj,jk)
               vb_ad(ji,jj,jk) = vb_ad(ji,jj,jk) + va_ad(ji,jj,jk) * vmask(ji,jj,jk)
               va_ad(ji,jj,jk) =            z2dt * va_ad(ji,jj,jk) * vmask(ji,jj,jk)
            END DO
         END DO  
#endif 
!                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============
      ! Lateral boundary conditions on ( ua, va )
      CALL lbc_lnk_adj( ua_ad, 'U', -1. )
      CALL lbc_lnk_adj( va_ad, 'V', -1. )

      !! Explicit physics with thickness weighted updates
      IF( lk_vvl .AND. .NOT. lk_dynspg_flt ) THEN
         CALL ctl_stop ('vvl not available for tangent')
      ENDIF

   END SUBROUTINE dyn_nxt_adj
   SUBROUTINE dyn_nxt_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE dyn_nxt_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !! ** Action  : Separate tests are applied for the following dx and dy:
      !!                
      !!              1) dx = ( SSH ) and dy = ( SSH )
      !!                   
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      INTEGER :: &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator

      !! * Local declarations
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & zun_tlin,     & ! Tangent input:   now   u-velocity
         & zvn_tlin,     & ! Tangent input:   now   v-velocity
         & zua_tlin,     & ! Tangent input:  after  u-velocity
         & zva_tlin,     & ! Tangent input:  after  u-velocity         
         & zub_tlin,     & ! Tangent input:  before u-velocity
         & zvb_tlin,     & ! Tangent input:  before u-velocity         
         & zun_adin,     & ! Adjoint input:   now   u-velocity
         & zvn_adin,     & ! Adjoint input:   now   v-velocity
         & zua_adin,     & ! Adjoint input:  after  u-velocity
         & zva_adin,     & ! Adjoint input:  after  u-velocity         
         & zub_adin,     & ! Adjoint input:  before u-velocity
         & zvb_adin,     & ! Adjoint input:  before u-velocity         
         & zun_tlout,    & ! Tangent output:  now   u-velocity
         & zvn_tlout,    & ! Tangent output:  now   v-velocity
         & zua_tlout,    & ! Tangent output: after  u-velocity
         & zva_tlout,    & ! Tangent output: after  u-velocity         
         & zub_tlout,    & ! Tangent output: before u-velocity
         & zvb_tlout,    & ! Tangent output: before u-velocity         
         & zun_adout,    & ! Adjoint output:  now   u-velocity
         & zvn_adout,    & ! Adjoint output:  now   v-velocity
         & zua_adout,    & ! Adjoint output: after  u-velocity
         & zva_adout,    & ! Adjoint output: after  u-velocity         
         & zub_adout,    & ! Adjoint output: before u-velocity
         & zvb_adout,    & ! Adjoint output: before u-velocity         
         & znu,          & ! 3D random field for u
         & znv,          & ! 3D random field for v
         & zbu,          & ! 3D random field for u
         & zbv,          & ! 3D random field for v
         & zau,          & ! 3D random field for u
         & zav             ! 3D random field for v
         
      REAL(KIND=wp) :: &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp1_1,       & !   scalar product components
         & zsp1_2,       & 
         & zsp1_3,       & 
         & zsp1_4,       & 
         & zsp1_5,       & 
         & zsp1_6,       & 
         & zsp2,         & ! scalar product involving the adjoint routine
         & zsp2_1,       & !   scalar product components
         & zsp2_2,       & 
         & zsp2_3,       & 
         & zsp2_4,       & 
         & zsp2_5,       & 
         & zsp2_6
      CHARACTER(LEN=14) :: cl_name

      ! Allocate memory

      ALLOCATE( &
         & zun_tlin(jpi,jpj,jpk),     & 
         & zvn_tlin(jpi,jpj,jpk),     & 
         & zua_tlin(jpi,jpj,jpk),     & 
         & zva_tlin(jpi,jpj,jpk),     & 
         & zub_tlin(jpi,jpj,jpk),     & 
         & zvb_tlin(jpi,jpj,jpk),     & 
         & zun_adin(jpi,jpj,jpk),     & 
         & zvn_adin(jpi,jpj,jpk),     & 
         & zua_adin(jpi,jpj,jpk),     & 
         & zva_adin(jpi,jpj,jpk),     & 
         & zub_adin(jpi,jpj,jpk),     & 
         & zvb_adin(jpi,jpj,jpk),     & 
         & zun_tlout(jpi,jpj,jpk),    & 
         & zvn_tlout(jpi,jpj,jpk),    & 
         & zua_tlout(jpi,jpj,jpk),    & 
         & zva_tlout(jpi,jpj,jpk),    & 
         & zub_tlout(jpi,jpj,jpk),    & 
         & zvb_tlout(jpi,jpj,jpk),    & 
         & zun_adout(jpi,jpj,jpk),    & 
         & zvn_adout(jpi,jpj,jpk),    & 
         & zua_adout(jpi,jpj,jpk),    & 
         & zva_adout(jpi,jpj,jpk),    & 
         & zub_adout(jpi,jpj,jpk),    & 
         & zvb_adout(jpi,jpj,jpk),    & 
         & znu(jpi,jpj,jpk),          & 
         & znv(jpi,jpj,jpk),          & 
         & zbu(jpi,jpj,jpk),          & 
         & zbv(jpi,jpj,jpk),          & 
         & zau(jpi,jpj,jpk),          & 
         & zav(jpi,jpj,jpk)           & 
         & )


      !==================================================================
      ! 1) dx = ( un_tl, vn_tl, hdivn_tl ) and 
      !    dy = ( hdivb_tl, hdivn_tl )
      !==================================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------

          zun_tlin(:,:,:) = 0.0_wp      
          zvn_tlin(:,:,:) = 0.0_wp      
          zua_tlin(:,:,:) = 0.0_wp      
          zva_tlin(:,:,:) = 0.0_wp      
          zub_tlin(:,:,:) = 0.0_wp      
          zvb_tlin(:,:,:) = 0.0_wp      
          zun_adin(:,:,:) = 0.0_wp      
          zvn_adin(:,:,:) = 0.0_wp      
          zua_adin(:,:,:) = 0.0_wp      
          zva_adin(:,:,:) = 0.0_wp      
          zub_adin(:,:,:) = 0.0_wp      
          zvb_adin(:,:,:) = 0.0_wp      
          zun_tlout(:,:,:) = 0.0_wp     
          zvn_tlout(:,:,:) = 0.0_wp     
          zua_tlout(:,:,:) = 0.0_wp     
          zva_tlout(:,:,:) = 0.0_wp     
          zub_tlout(:,:,:) = 0.0_wp     
          zvb_tlout(:,:,:) = 0.0_wp     
          zun_adout(:,:,:) = 0.0_wp     
          zvn_adout(:,:,:) = 0.0_wp     
          zua_adout(:,:,:) = 0.0_wp     
          zva_adout(:,:,:) = 0.0_wp     
          zub_adout(:,:,:) = 0.0_wp     
          zvb_adout(:,:,:) = 0.0_wp     
          znu(:,:,:) = 0.0_wp           
          znv(:,:,:) = 0.0_wp           
          zbu(:,:,:) = 0.0_wp           
          zbv(:,:,:) = 0.0_wp           
          zau(:,:,:) = 0.0_wp           
          zav(:,:,:) = 0.0_wp           

          un_tl(:,:,:) = 0.0_wp      
          vn_tl(:,:,:) = 0.0_wp      
          ua_tl(:,:,:) = 0.0_wp      
          va_tl(:,:,:) = 0.0_wp      
          ub_tl(:,:,:) = 0.0_wp      
          vb_tl(:,:,:) = 0.0_wp      
          un_ad(:,:,:) = 0.0_wp      
          vn_ad(:,:,:) = 0.0_wp      
          ua_ad(:,:,:) = 0.0_wp      
          va_ad(:,:,:) = 0.0_wp      
          ub_ad(:,:,:) = 0.0_wp      
          vb_ad(:,:,:) = 0.0_wp      


      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, znu, 'U', 0.0_wp, stdu )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 523432 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, znv, 'V', 0.0_wp, stdv )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 456953 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zbu, 'U', 0.0_wp, stdu )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 267406 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zbv, 'V', 0.0_wp, stdv )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 432545 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zau, 'U', 0.0_wp, stdu )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 287503 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zav, 'V', 0.0_wp, stdv )

      DO jk = 1, jpk
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zun_tlin(ji,jj,jk) = znu(ji,jj,jk) 
               zvn_tlin(ji,jj,jk) = znv(ji,jj,jk) 
               zub_tlin(ji,jj,jk) = zbu(ji,jj,jk)
               zvb_tlin(ji,jj,jk) = zbv(ji,jj,jk)
               zua_tlin(ji,jj,jk) = zau(ji,jj,jk)
               zva_tlin(ji,jj,jk) = zav(ji,jj,jk)
            END DO
         END DO
      END DO 

      un_tl(:,:,:) = zun_tlin(:,:,:)
      vn_tl(:,:,:) = zvn_tlin(:,:,:)
      ub_tl(:,:,:) = zub_tlin(:,:,:)
      vb_tl(:,:,:) = zvb_tlin(:,:,:)
      ua_tl(:,:,:) = zua_tlin(:,:,:)
      va_tl(:,:,:) = zva_tlin(:,:,:)

      call dyn_nxt_tan ( nit000 )

      zun_tlout(:,:,:) = un_tl(:,:,:)
      zvn_tlout(:,:,:) = vn_tl(:,:,:)
      zub_tlout(:,:,:) = ub_tl(:,:,:)
      zvb_tlout(:,:,:) = vb_tl(:,:,:)
      zua_tlout(:,:,:) = ua_tl(:,:,:)
      zva_tlout(:,:,:) = va_tl(:,:,:)

      !--------------------------------------------------------------------
      ! Initialize the adjoint variables: dy^* = W dy
      !--------------------------------------------------------------------

      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zun_adin(ji,jj,jk) = zun_tlout(ji,jj,jk) &
                 &               * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,jk) &
                 &               * umask(ji,jj,jk)
              zvn_adin(ji,jj,jk) = zvn_tlout(ji,jj,jk) &
                 &               * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,jk) &
                 &               * vmask(ji,jj,jk)
              zub_adin(ji,jj,jk) = zub_tlout(ji,jj,jk) &
                 &               * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,jk) &
                 &               * umask(ji,jj,jk)
              zvb_adin(ji,jj,jk) = zvb_tlout(ji,jj,jk) &
                 &               * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,jk) &
                 &               * vmask(ji,jj,jk)
              zua_adin(ji,jj,jk) = zua_tlout(ji,jj,jk) &
                 &               * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,jk) &
                 &               * umask(ji,jj,jk)
              zva_adin(ji,jj,jk) = zva_tlout(ji,jj,jk) &
                 &               * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,jk) &
                 &               * vmask(ji,jj,jk)
            END DO
         END DO
      END DO
      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1_1 = DOT_PRODUCT( zun_tlout, zun_adin )
      zsp1_2 = DOT_PRODUCT( zvn_tlout, zvn_adin )
      zsp1_3 = DOT_PRODUCT( zub_tlout, zub_adin )
      zsp1_4 = DOT_PRODUCT( zvb_tlout, zvb_adin )
      zsp1_5 = DOT_PRODUCT( zua_tlout, zua_adin )
      zsp1_6 = DOT_PRODUCT( zva_tlout, zva_adin )
      zsp1   = zsp1_1 + zsp1_2 + zsp1_3 + zsp1_4 + zsp1_5 + zsp1_6 

      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------

      un_ad(:,:,:) = zun_adin(:,:,:)
      vn_ad(:,:,:) = zvn_adin(:,:,:)
      ub_ad(:,:,:) = zub_adin(:,:,:)
      vb_ad(:,:,:) = zvb_adin(:,:,:)
      ua_ad(:,:,:) = zua_adin(:,:,:)
      va_ad(:,:,:) = zva_adin(:,:,:)

      CALL dyn_nxt_adj ( nit000 )

      zun_adout(:,:,:) = un_ad(:,:,:)
      zvn_adout(:,:,:) = vn_ad(:,:,:)
      zub_adout(:,:,:) = ub_ad(:,:,:)
      zvb_adout(:,:,:) = vb_ad(:,:,:)
      zua_adout(:,:,:) = ua_ad(:,:,:)
      zva_adout(:,:,:) = va_ad(:,:,:)

      zsp2_1 = DOT_PRODUCT( zun_tlin, zun_adout )
      zsp2_2 = DOT_PRODUCT( zvn_tlin, zvn_adout )
      zsp2_3 = DOT_PRODUCT( zub_tlin, zub_adout )
      zsp2_4 = DOT_PRODUCT( zvb_tlin, zvb_adout )
      zsp2_5 = DOT_PRODUCT( zua_tlin, zua_adout )
      zsp2_6 = DOT_PRODUCT( zva_tlin, zva_adout )
      zsp2   = zsp2_1 + zsp2_2 + zsp2_3 + zsp2_4 + zsp2_5 + zsp2_6

      ! Compare the scalar products
      ! 14 char:'12345678901234'
      cl_name = 'dyn_nxt_adj   '
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      DEALLOCATE( &
         & zun_tlin,     & 
         & zvn_tlin,     & 
         & zua_tlin,     & 
         & zva_tlin,     & 
         & zub_tlin,     & 
         & zvb_tlin,     & 
         & zun_adin,     & 
         & zvn_adin,     & 
         & zua_adin,     & 
         & zva_adin,     & 
         & zub_adin,     & 
         & zvb_adin,     & 
         & zun_tlout,    & 
         & zvn_tlout,    & 
         & zua_tlout,    & 
         & zva_tlout,    & 
         & zub_tlout,    & 
         & zvb_tlout,    & 
         & zun_adout,    & 
         & zvn_adout,    & 
         & zua_adout,    & 
         & zva_adout,    & 
         & zub_adout,    & 
         & zvb_adout,    & 
         & znu,          & 
         & znv,          & 
         & zbu,          & 
         & zbv,          & 
         & zau,          & 
         & zav           & 
         & )


   END SUBROUTINE dyn_nxt_adj_tst
   !!======================================================================
#endif
END MODULE dynnxt_tam
