MODULE dynzdf_exp_tam
#ifdef key_tam
   !!==============================================================================
   !!                     ***  MODULE  dynzdf_exp_tam  ***
   !! Ocean dynamics:  vertical component(s) of the momentum mixing trend
   !!                 Tangent and Adjoint Module
   !!==============================================================================
   !! History of the direct module:
   !!                !  90-10  (B. Blanke)  Original code
   !!                !  97-05  (G. Madec)  vertical component of isopycnal
   !!           8.5  !  02-08  (G. Madec)  F90: Free form and module
   !! History of the TAM module:
   !!           9.0  !  08-0!  (A. Vidard) Skeleton
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   dyn_zdf_exp  : update the momentum trend with the vertical diffu-
   !!                  sion using an explicit time-stepping scheme.
   !!----------------------------------------------------------------------
   !! * Modules used

   USE in_out_manager  ! I/O manager
   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC dyn_zdf_exp_tan     ! called by dynzdf_tam.F90
   PUBLIC dyn_zdf_exp_adj     ! called by dynzdf_tam.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE dyn_zdf_exp_tan( kt, p2dt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_zdf_exp_tan  ***
      !!                   
      !! ** Purpose of the direct routine:
      !!      Compute the trend due to the vert. momentum diffusion
      !!
      !! ** Method of the direct routine:
      !!      Explicit forward time stepping with a time splitting
      !!      technique. The vertical diffusion of momentum is given by:
      !!         diffu = dz( avmu dz(u) ) = 1/e3u dk+1( avmu/e3uw dk(ub) )
      !!      Surface boundary conditions: wind stress input
      !!      Bottom boundary conditions : bottom stress (cf zdfbfr.F90)
      !!      Add this trend to the general trend ua :
      !!         ua = ua + dz( avmu dz(u) )
      !!
      !! ** Action : - Update (ua,va) with the vertical diffusive trend
      !!---------------------------------------------------------------------
      !! * Arguments
      INTEGER , INTENT( in ) ::   kt                           ! ocean time-step index
      REAL(wp), INTENT( in ) ::   p2dt                         ! time-step 

      !! * Local declarations
      INTEGER ::   ji, jj, jk, jl                              ! dummy loop indices
      REAL(wp) ::   zrau0r, zlavmr, zua, zva                   ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpk) ::   zwx, zwy, zwz, zww     ! temporary workspace arrays
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn_zdf_exp_tan : vertical momentum diffusion explicit operator'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~ '
      ENDIF

      CALL ctl_stop ('dyn_zdf_exp_tan not available yet')
   END SUBROUTINE dyn_zdf_exp_tan
   SUBROUTINE dyn_zdf_exp_adj( kt, p2dt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_zdf_exp_adj  ***
      !!                   
      !! ** Purpose of the direct routine:
      !!      Compute the trend due to the vert. momentum diffusion
      !!
      !! ** Method of the direct routine:
      !!      Explicit forward time stepping with a time splitting
      !!      technique. The vertical diffusion of momentum is given by:
      !!         diffu = dz( avmu dz(u) ) = 1/e3u dk+1( avmu/e3uw dk(ub) )
      !!      Surface boundary conditions: wind stress input
      !!      Bottom boundary conditions : bottom stress (cf zdfbfr.F90)
      !!      Add this trend to the general trend ua :
      !!         ua = ua + dz( avmu dz(u) )
      !!
      !! ** Action : - Update (ua,va) with the vertical diffusive trend
      !!---------------------------------------------------------------------
      !! * Arguments
      INTEGER , INTENT( in ) ::   kt                           ! ocean time-step index
      REAL(wp), INTENT( in ) ::   p2dt                         ! time-step 

      !! * Local declarations
      INTEGER ::   ji, jj, jk, jl                              ! dummy loop indices
      REAL(wp) ::   zrau0r, zlavmr, zua, zva                   ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpk) ::   zwx, zwy, zwz, zww     ! temporary workspace arrays
      !!----------------------------------------------------------------------

      IF( kt == nitend ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn_zdf_exp_adj : vertical momentum diffusion explicit operator'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~ '
      ENDIF

      CALL ctl_stop ('dyn_zdf_exp_adj not available yet')
   END SUBROUTINE dyn_zdf_exp_adj
#endif
   !!==============================================================================
END MODULE dynzdf_exp_tam
