MODULE obctra_tam
   !!=================================================================================
   !!                       ***  MODULE  obctra  ***
   !! Ocean tracers:   Radiation of tracers on each open boundary
   !!=================================================================================
#if defined key_obc
   !!---------------------------------------------------------------------------------
   !!   'key_obc'      :                                      Open Boundary Conditions
   !!---------------------------------------------------------------------------------
   !!   obc_tra_tan        : call the subroutine for each open boundary
   !!   obc_tra_east_tan   : 
   !!   obc_tra_west_tan   : 
   !!   obc_tra_north_tan  : 
   !!   obc_tra_south_tan  : 
   !!   obc_tra_adj        : call the subroutine for each open boundary
   !!   obc_tra_east_adj   : 
   !!   obc_tra_west_adj   : 
   !!   obc_tra_north_adj  : 
   !!   obc_tra_south_adj  : 
   !!----------------------------------------------------------------------------------
   !! * Modules used
   USE oce_tam,         ONLY : tb_tl, sb_tl, ta_tl, sa_tl, & ! ocean dynamics and tracers
        &                      tb_ad, sb_ad, ta_ad, sa_ad
   USE dom_oce         ! ocean space and time domain variables 
   USE phycst          ! physical constants
   USE obc_oce         ! ocean open boundary conditions
   USE lib_mpp         ! ???
   USE lbclnk          ! ???
   USE lbclnk_tam,      ONLY : lbc_lnk_adj ! TAM lateral boundary conditions
   USE in_out_manager  ! I/O manager

   IMPLICIT NONE
   PRIVATE

   !! * Accessibility
   PUBLIC obc_tra_tan     ! routine called in tranxt.F90 
   PUBLIC obc_tra_adj     ! routine called in tranxt.F90 

   !! * Module variables
   LOGICAL  ::   ll_fbc

   
   !! * Substitutions
#  include "obc_vectopt_loop_substitute.h90"
   !!---------------------------------------------------------------------------------
   !!   OPA 9.0 , LOCEAN-IPSL (2005) 
   !! $Header: /home/opalod/NEMOCVSROOT/NEMO/OPA_SRC/OBC/obctra.F90,v 1.4 2005/03/27 18:35:10 opalod Exp $ 
   !! This software is governed by the CeCILL licence see modipsl/doc/NEMO_CeCILL.txt 
   !!---------------------------------------------------------------------------------

CONTAINS

   SUBROUTINE obc_tra_tan( kt )
      !!-------------------------------------------------------------------------------
      !!                 ***  SUBROUTINE obc_tra  ***
      !!                    
      !! ** Purpose :   Compute tracer fields (t,s) along the open boundaries.
      !!      This routine is called by the tranxt.F routine and updates ta,sa
      !!      which are the actual temperature and salinity fields.
      !!        The logical variable lp_obc_east, and/or lp_obc_west, and/or lp_obc_north,
      !!      and/or lp_obc_south allow the user to determine which boundary is an
      !!      open one (must be done in the param_obc.h90 file).
      !!
      !! Reference : 
      !!   Marchesiello P., 1995, these de l'universite J. Fourier, Grenoble, France.
      !!
      !!  History :
      !!        !  95-03 (J.-M. Molines) Original, SPEM
      !!        !  97-07 (G. Madec, J.-M. Molines) addition
      !!   8.5  !  02-10 (C. Talandier, A-M. Treguier) F90
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt
      !!----------------------------------------------------------------------

      ! 0. Local constant initialization

#if defined key_pomme_r025
      ! Warning : TAM is available only for fixed boundary conditions (FBC)
      ll_fbc = .true.
#else
      Error, OBC not ready.
#endif

!    Remove the dependance on kt to avoid compilation warnings
      IF( lp_obc_east  )   CALL obc_tra_east_tan     ! East open boundary
      IF( lp_obc_west  )   CALL obc_tra_west_tan     ! West open boundary
      IF( lp_obc_north )   CALL obc_tra_north_tan    ! North open boundary
      IF( lp_obc_south )   CALL obc_tra_south_tan    ! South open boundary

      IF( lk_mpp ) THEN
         CALL lbc_lnk( ta_tl, 'T', 1. )
         CALL lbc_lnk( sa_tl, 'T', 1. )
      ENDIF

   END SUBROUTINE obc_tra_tan

   SUBROUTINE obc_tra_adj( kt )
      !!                 ***  SUBROUTINE obc_tra  ***
      INTEGER, INTENT( in ) ::   kt
#if defined key_pomme_r025
      ! Warning : TAM is available only for fixed boundary conditions (FBC)
      ll_fbc = .true.
#else
      Error, OBC not ready.
#endif

      IF( lk_mpp ) THEN
         CALL lbc_lnk_adj( sa_ad, 'T', 1. )
         CALL lbc_lnk_adj( ta_ad, 'T', 1. )
      ENDIF

      IF( lp_obc_south )   CALL obc_tra_south_adj    ! South open boundary
      IF( lp_obc_north )   CALL obc_tra_north_adj    ! North open boundary
      IF( lp_obc_west  )   CALL obc_tra_west_adj     ! West open boundary
      IF( lp_obc_east  )   CALL obc_tra_east_adj     ! East open boundary

   END SUBROUTINE obc_tra_adj

   SUBROUTINE obc_tra_east_tan
      !!------------------------------------------------------------------------------
      !!                ***  SUBROUTINE obc_tra_east  ***
      !!                  
      !! ** Purpose :
      !!      Apply the radiation algorithm on east OBC tracers ta, sa using the 
      !!      phase velocities calculated in obc_rad_east subroutine in obcrad.F90 module
      !!      If the logical lfbceast is .TRUE., there is no radiation but only fixed OBC
      !!
      !!  History :
      !!         ! 95-03 (J.-M. Molines) Original from SPEM
      !!         ! 97-07 (G. Madec, J.-M. Molines) additions
      !!         ! 97-12 (M. Imbard) Mpp adaptation
      !!         ! 00-06 (J.-M. Molines) 
      !!    8.5  ! 02-10 (C. Talandier, A-M. Treguier) F90
      !!------------------------------------------------------------------------------
      !! * Arguments
      !! * Local declaration
      INTEGER ::   ji, jj, jk      ! dummy loop indices
      !!------------------------------------------------------------------------------
      ! 1. First three time steps and more if lfbceast is .TRUE.
      !    In that case open boundary conditions are FIXED.
      ! --------------------------------------------------------

      IF ( ll_fbc .OR. lfbceast ) THEN

        DO ji = fs_nie0+1, fs_nie1+1 ! Vector opt.
            DO jk = 1, jpkm1
               DO jj = 1, jpj
                  ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) * (1. - temsk(jj,jk))
                  sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk) * (1. - temsk(jj,jk))
               END DO
            END DO
         END DO

      ELSE

         CALL ctl_stop( 'Error in obctra_tam : TAM is available only for fixed boundary conditions' )

      END IF

   END SUBROUTINE obc_tra_east_tan

   SUBROUTINE obc_tra_east_adj
      !!                ***  SUBROUTINE obc_tra_east  ***
      INTEGER ::   ji, jj, jk      ! dummy loop indices
      IF ( ll_fbc .OR. lfbceast ) THEN

        DO ji = fs_nie0+1, fs_nie1+1 ! Vector opt.
            DO jk = 1, jpkm1
               DO jj = 1, jpj
                  ta_ad(ji,jj,jk) = ta_ad(ji,jj,jk) * (1. - temsk(jj,jk))
                  sa_ad(ji,jj,jk) = sa_ad(ji,jj,jk) * (1. - temsk(jj,jk))
               END DO
            END DO
         END DO

      ELSE

         CALL ctl_stop( 'Error in obctra_tam : TAM is available only for fixed boundary conditions' )

      END IF

   END SUBROUTINE obc_tra_east_adj

   SUBROUTINE obc_tra_west_tan
      !!------------------------------------------------------------------------------
      !!                 ***  SUBROUTINE obc_tra_west  ***
      !!           
      !! ** Purpose :
      !!      Apply the radiation algorithm on west OBC tracers ta, sa using the 
      !!      phase velocities calculated in obc_rad_west subroutine in obcrad.F90 module
      !!      If the logical lfbcwest is .TRUE., there is no radiation but only fixed OBC
      !!
      !!  History :
      !!         ! 95-03 (J.-M. Molines) Original from SPEM
      !!         ! 97-07 (G. Madec, J.-M. Molines) additions
      !!         ! 97-12 (M. Imbard) Mpp adaptation
      !!         ! 00-06 (J.-M. Molines) 
      !!    8.5  ! 02-10 (C. Talandier, A-M. Treguier) F90
      !!------------------------------------------------------------------------------
      !! * Arguments
      !! * Local declaration
      INTEGER ::   ji, jj, jk      ! dummy loop indices
      !!------------------------------------------------------------------------------

      ! 1. First three time steps and more if lfbcwest is .TRUE.
      !    In that case open boundary conditions are FIXED.
      ! --------------------------------------------------------

      IF ( ll_fbc .OR. lfbcwest ) THEN

         DO ji = fs_niw0, fs_niw1 ! Vector opt.
            DO jk = 1, jpkm1
               DO jj = 1, jpj
                  ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) * (1. - twmsk(jj,jk))
                  sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk) * (1. - twmsk(jj,jk))
               END DO
            END DO
         END DO

      ELSE

         CALL ctl_stop( 'Error in obctra_tam : TAM is available only for fixed boundary conditions' )

      END IF

   END SUBROUTINE obc_tra_west_tan

   SUBROUTINE obc_tra_west_adj
      !!                 ***  SUBROUTINE obc_tra_west  ***
      INTEGER ::   ji, jj, jk      ! dummy loop indices
      IF ( ll_fbc .OR. lfbcwest ) THEN

         DO ji = fs_niw0, fs_niw1 ! Vector opt.
            DO jk = 1, jpkm1
               DO jj = 1, jpj
                  ta_ad(ji,jj,jk) = ta_ad(ji,jj,jk) * (1. - twmsk(jj,jk))
                  sa_ad(ji,jj,jk) = sa_ad(ji,jj,jk) * (1. - twmsk(jj,jk))
               END DO
            END DO
         END DO

      ELSE

         CALL ctl_stop( 'Error in obctra_tam : TAM is available only for fixed boundary conditions' )

      END IF

   END SUBROUTINE obc_tra_west_adj

   SUBROUTINE obc_tra_north_tan
      !!------------------------------------------------------------------------------
      !!                 ***  SUBROUTINE obc_tra_north  ***
      !!
      !! ** Purpose :
      !!      Apply the radiation algorithm on north OBC tracers ta, sa using the 
      !!      phase velocities calculated in obc_rad_north subroutine in obcrad.F90 module
      !!      If the logical lfbcnorth is .TRUE., there is no radiation but only fixed OBC
      !!
      !!  History :
      !!         ! 95-03 (J.-M. Molines) Original from SPEM
      !!         ! 97-07 (G. Madec, J.-M. Molines) additions
      !!         ! 97-12 (M. Imbard) Mpp adaptation
      !!         ! 00-06 (J.-M. Molines) 
      !!    8.5  ! 02-10 (C. Talandier, A-M. Treguier) F90
      !!------------------------------------------------------------------------------
      !! * Arguments
      !! * Local declaration
      INTEGER ::   ji, jj, jk      ! dummy loop indices
      !!------------------------------------------------------------------------------

      ! 1. First three time steps and more if lfbcnorth is .TRUE.
      !    In that case open boundary conditions are FIXED.
      ! --------------------------------------------------------

      IF ( ll_fbc .OR. lfbcnorth ) THEN

         DO jj = fs_njn0+1, fs_njn1+1  ! Vector opt.
            DO jk = 1, jpkm1
               DO ji = 1, jpi
                  ta_tl(ji,jj,jk)= ta_tl(ji,jj,jk) * (1.-tnmsk(ji,jk)) 
                  sa_tl(ji,jj,jk)= sa_tl(ji,jj,jk) * (1.-tnmsk(ji,jk)) 
               END DO
            END DO
         END DO

      ELSE

         CALL ctl_stop( 'Error in obctra_tam : TAM is available only for fixed boundary conditions' )

      END IF

   END SUBROUTINE obc_tra_north_tan

   SUBROUTINE obc_tra_north_adj
      !!                 ***  SUBROUTINE obc_tra_north  ***
      INTEGER ::   ji, jj, jk      ! dummy loop indices

      IF ( ll_fbc .OR. lfbcnorth ) THEN

         DO jj = fs_njn0+1, fs_njn1+1  ! Vector opt.
            DO jk = 1, jpkm1
               DO ji = 1, jpi
                  ta_ad(ji,jj,jk)= ta_ad(ji,jj,jk) * (1.-tnmsk(ji,jk)) 
                  sa_ad(ji,jj,jk)= sa_ad(ji,jj,jk) * (1.-tnmsk(ji,jk)) 
               END DO
            END DO
         END DO

      ELSE

         CALL ctl_stop( 'Error in obctra_tam : TAM is available only for fixed boundary conditions' )

      END IF

   END SUBROUTINE obc_tra_north_adj

   SUBROUTINE obc_tra_south_tan
      !!------------------------------------------------------------------------------
      !!                ***  SUBROUTINE obc_tra_south  ***
      !!     
      !! ** Purpose :
      !!      Apply the radiation algorithm on south OBC tracers ta, sa using the 
      !!      phase velocities calculated in obc_rad_south subroutine in obcrad.F90 module
      !!      If the logical lfbcsouth is .TRUE., there is no radiation but only fixed OBC
      !!
      !!  History :
      !!         ! 95-03 (J.-M. Molines) Original from SPEM
      !!         ! 97-07 (G. Madec, J.-M. Molines) additions
      !!         ! 97-12 (M. Imbard) Mpp adaptation
      !!         ! 00-06 (J.-M. Molines) 
      !!    8.5  ! 02-10 (C. Talandier, A-M Treguier) F90
      !!------------------------------------------------------------------------------
      !! * Arguments
      !! * Local declaration
      INTEGER ::   ji, jj, jk      ! dummy loop indices
      !!------------------------------------------------------------------------------

      ! 1. First three time steps and more if lfbcsouth is .TRUE.
      !    In that case open boundary conditions are FIXED.
      ! --------------------------------------------------------

      IF ( ll_fbc .OR. lfbcsouth ) THEN

         DO jj = fs_njs0, fs_njs1  ! Vector opt.
            DO jk = 1, jpkm1
               DO ji = 1, jpi
                  ta_tl(ji,jj,jk)= ta_tl(ji,jj,jk) * (1.-tsmsk(ji,jk))
                  sa_tl(ji,jj,jk)= sa_tl(ji,jj,jk) * (1.-tsmsk(ji,jk))
               END DO
            END DO
         END DO

      ELSE

         CALL ctl_stop( 'Error in obctra_tam : TAM is available only for fixed boundary conditions' )

      END IF   

   END SUBROUTINE obc_tra_south_tan

   SUBROUTINE obc_tra_south_adj
      !!                ***  SUBROUTINE obc_tra_south  ***
      INTEGER ::   ji, jj, jk      ! dummy loop indices

      IF ( ll_fbc .OR. lfbcsouth ) THEN

         DO jj = fs_njs0, fs_njs1  ! Vector opt.
            DO jk = 1, jpkm1
               DO ji = 1, jpi
                  ta_ad(ji,jj,jk)= ta_ad(ji,jj,jk) * (1.-tsmsk(ji,jk))
                  sa_ad(ji,jj,jk)= sa_ad(ji,jj,jk) * (1.-tsmsk(ji,jk))
               END DO
            END DO
         END DO

      ELSE

         CALL ctl_stop( 'Error in obctra_tam : TAM is available only for fixed boundary conditions' )

      END IF   

   END SUBROUTINE obc_tra_south_adj

#else
   !!---------------------------------------------------------------------------------
   !!   Default option                                                    Empty module
   !!---------------------------------------------------------------------------------
CONTAINS
   SUBROUTINE obc_tra_tan      ! Empty routine
   END SUBROUTINE obc_tra_tan
   SUBROUTINE obc_tra_adj
   END SUBROUTINE obc_tra_adj

#endif

   !!=================================================================================
END MODULE obctra_tam
