MODULE sbcssm_tam
#ifdef key_tam
   !!======================================================================
   !!                       ***  MODULE  sbcssm_tam  ***
   !! Surface module :  provide time-mean ocean surface variables
   !!                   Tangent and adjoint module
   !!======================================================================
   !! History of the direct module:
   !!            9.0   !  06-07  (G. Madec)  Original code
   !! History of the TAM module:
   !!            9.0   !  08-11  (A. Vidard) Original code
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   sbc_ssm_[tan adj]: calculate sea surface mean currents, temperature,  
   !!                    and salinity over nn_fsbc time-step
   !!----------------------------------------------------------------------
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 &
      & jpiglo
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE oce_tam       , ONLY: & ! ocean dynamics and tracers
      & un_tl,               &
      & vn_tl,               &
      & tn_tl,               &
      & sn_tl,               &
      & un_ad,               &
      & vn_ad,               &
      & tn_ad,               &
      & sn_ad
   USE dom_oce       , ONLY: & ! Ocean space and time domain
      & e2u,                 &
      & e1u,                 &
      & e2v,                 &
      & e1v,                 &
      & e1t,                 &
      & e2t,                 &
# if defined key_vvl
      & e3t_1,               &
# else
#  if defined key_zco
      & e3t_0,               &
#  else
      & e3t,                 &
#  endif
# endif
# if defined key_zco
      & e3t_0,               &
# else
      & e3u,                 &
      & e3v,                 &
# endif
      & tmask,               &
      & umask,               &
      & mig,                 &
      & mjg,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej
   USE sbc_oce       , ONLY: & !
      & nn_fsbc                ! Surface boundary condition: frequency of sbc computation (as well as sea-ice model)
   USE sbc_oce_tam   , ONLY: & ! Surface boundary condition: ocean fields
      & ssu_m_tl,            & ! mean (nn_fsbc time-step) surface sea i-current (U-point) [m/s]
      & ssv_m_tl,            & ! mean (nn_fsbc time-step) surface sea j-current (U-point) [m/s]
      & sst_m_tl,            & ! mean (nn_fsbc time-step) surface sea temperature     [Celsius]
      & sss_m_tl,            & ! mean (nn_fsbc time-step) surface sea salinity            [psu]
      & ssu_m_ad,            & ! mean (nn_fsbc time-step) surface sea i-current (U-point) [m/s]
      & ssv_m_ad,            & ! mean (nn_fsbc time-step) surface sea j-current (U-point) [m/s]
      & sst_m_ad,            & ! mean (nn_fsbc time-step) surface sea temperature     [Celsius]
      & sss_m_ad               ! mean (nn_fsbc time-step) surface sea salinity            [psu]
   USE in_out_manager, ONLY: & ! I/O manager 
      & lwp,                 &
      & numout,              & 
      & nit000,              & 
      & nitend,              & 
      & ln_rstart
   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld,     ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE paresp        , ONLY: & ! Weights for an energy-type scalar product
      & wesp_t,              &
      & wesp_s
   USE tstool_tam    , ONLY: &
      & stdu,                &
      & stdv,                &
      & stdt,                &
      & stds,                &
      & prntst_adj

   IMPLICIT NONE
   PRIVATE

   PUBLIC   sbc_ssm_tan     ! routine called by step_tam.F90
   PUBLIC   sbc_ssm_adj     ! routine called by step_tam.F90
   PUBLIC   sbc_ssm_adj_tst ! routine called by tst.F90
   
   !! * Substitutions
#  include "domzgr_substitute.h90"
   !!----------------------------------------------------------------------
   !!   OPA 9.0 , LOCEAN-IPSL (2006) 
   !! $Id: sbcssm.F90 1196 2008-09-19 07:07:00Z ctlod $
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE sbc_ssm_tan( kt )
      !!---------------------------------------------------------------------
      !!                     ***  ROUTINE sbc_ssm_tan  ***
      !!                     
      !! ** Purpose of the direct routine:
      !!                provide ocean surface variable to sea-surface boundary
      !!                condition computation 
      !!                
      !! ** Method of the direct routine:
      !!      compute mean surface velocity (2 components at U and 
      !!      V-points) [m/s], temperature [Celcius] and salinity [psu] over
      !!      the periode (kt - nn_fsbc) to kt
      !!---------------------------------------------------------------------
      INTEGER, INTENT(in) ::   kt        ! ocean time step
      !
      REAL(wp) ::   zcoef       ! temporary scalar
      REAL(wp) ::   zf_sbc      ! read sbc frequency 
      !!---------------------------------------------------------------------
      !                                                   ! ---------------------------------------- !
      IF( nn_fsbc == 1 ) THEN                             !      Instantaneous surface fields        !
         !                                                ! ---------------------------------------- !
         IF( kt == nit000 ) THEN
            IF(lwp) WRITE(numout,*)
            IF(lwp) WRITE(numout,*) 'sbc_ssm_tan: sea surface mean fields, nn_fsbc=1 : instantaneous values'
            IF(lwp) WRITE(numout,*) '~~~~~~~~~~~ '
         ENDIF
         !
         ssu_m_tl(:,:) = un_tl(:,:,1)
         ssv_m_tl(:,:) = vn_tl(:,:,1)
         sst_m_tl(:,:) = tn_tl(:,:,1)
         sss_m_tl(:,:) = sn_tl(:,:,1)
         !
      ELSE
         !                                                ! ---------------------------------------- !
         IF( kt == nit000) THEN                           !       Initialisation: 1st time-step      !
            !                                             ! ---------------------------------------- !

            IF( ln_rstart ) THEN 
                  ssu_m_tl(:,:) = 0.0_wp
                  ssv_m_tl(:,:) = 0.0_wp
                  sst_m_tl(:,:) = 0.0_wp
                  sss_m_tl(:,:) = 0.0_wp
             ELSE
               IF(lwp) WRITE(numout,*) '~~~~~~~   mean fields initialised to instantaneous values'
               zcoef = REAL( nn_fsbc - 1, wp )
               ssu_m_tl(:,:) = zcoef * un_tl(:,:,1)
               ssv_m_tl(:,:) = zcoef * vn_tl(:,:,1)
               sst_m_tl(:,:) = zcoef * tn_tl(:,:,1)
               sss_m_tl(:,:) = zcoef * sn_tl(:,:,1)
            ENDIF
            !                                             ! ---------------------------------------- !
         ELSEIF( MOD( kt - 2 , nn_fsbc ) == 0 ) THEN      !   Initialisation: New mean computation   !
            !                                             ! ---------------------------------------- !
            ssu_m_tl(:,:) = 0.0_wp      ! reset to zero ocean mean sbc fields
            ssv_m_tl(:,:) = 0.0_wp
            sst_m_tl(:,:) = 0.0_wp
            sss_m_tl(:,:) = 0.0_wp
         ENDIF
         !                                                ! ---------------------------------------- !
         !                                                !        Cumulate at each time step        !
         !                                                ! ---------------------------------------- !
         ssu_m_tl(:,:) = ssu_m_tl(:,:) + un_tl(:,:,1)
         ssv_m_tl(:,:) = ssv_m_tl(:,:) + vn_tl(:,:,1)
         sst_m_tl(:,:) = sst_m_tl(:,:) + tn_tl(:,:,1)
         sss_m_tl(:,:) = sss_m_tl(:,:) + sn_tl(:,:,1)

         !                                                ! ---------------------------------------- !
         IF( MOD( kt - 1 , nn_fsbc ) == 0 ) THEN          !   Mean value at each nn_fsbc time-step   !
            !                                             ! ---------------------------------------- !
            zcoef = 1. / REAL( nn_fsbc, wp )
            sst_m_tl(:,:) = sst_m_tl(:,:) * zcoef           ! mean SST             [Celcius]
            sss_m_tl(:,:) = sss_m_tl(:,:) * zcoef           ! mean SSS             [psu]
            ssu_m_tl(:,:) = ssu_m_tl(:,:) * zcoef           ! mean suface current  [m/s]
            ssv_m_tl(:,:) = ssv_m_tl(:,:) * zcoef           !
            !
         ENDIF
         !
      ENDIF
      !
   END SUBROUTINE sbc_ssm_tan

   SUBROUTINE sbc_ssm_adj( kt )
      !!---------------------------------------------------------------------
      !!                     ***  ROUTINE sbc_ssm_adj  ***
      !!                     
      !! ** Purpose of the direct routine:
      !!                provide ocean surface variable to sea-surface boundary
      !!                condition computation 
      !!                
      !! ** Method of the direct routine:
      !!      compute mean surface velocity (2 components at U and 
      !!      V-points) [m/s], temperature [Celcius] and salinity [psu] over
      !!      the periode (kt - nn_fsbc) to kt
      !!---------------------------------------------------------------------
      INTEGER, INTENT(in) ::   kt        ! ocean time step
      !
      REAL(wp) ::   zcoef       ! temporary scalar
      REAL(wp) ::   zf_sbc      ! read sbc frequency 
      !!---------------------------------------------------------------------
      !                                                   ! ---------------------------------------- !
      IF( nn_fsbc == 1 ) THEN                             !      Instantaneous surface fields        !
         !                                                ! ---------------------------------------- !
         IF( kt == nitend) THEN
            IF(lwp) WRITE(numout,*)
            IF(lwp) WRITE(numout,*) 'sbc_ssm_adj: sea surface mean fields, nn_fsbc=1 : instantaneous values'
            IF(lwp) WRITE(numout,*) '~~~~~~~~~~~ '
         ENDIF
         !
         un_ad(:,:,1)  = un_ad(:,:,1) + ssu_m_ad(:,:)
         vn_ad(:,:,1)  = vn_ad(:,:,1) + ssv_m_ad(:,:)
         tn_ad(:,:,1)  = tn_ad(:,:,1) + sst_m_ad(:,:)
         sn_ad(:,:,1)  = sn_ad(:,:,1) + sss_m_ad(:,:)
         ssu_m_ad(:,:) = 0.0_wp
         ssv_m_ad(:,:) = 0.0_wp
         sst_m_ad(:,:) = 0.0_wp
         sss_m_ad(:,:) = 0.0_wp
         !
      ELSE
         !                                                ! ---------------------------------------- !
         IF( MOD( kt - 1 , nn_fsbc ) == 0 ) THEN          !   Mean value at each nn_fsbc time-step   !
            !                                             ! ---------------------------------------- !
            zcoef = 1. / REAL( nn_fsbc, wp )
            sst_m_ad(:,:) = sst_m_ad(:,:) * zcoef           ! mean SST             [Celcius]
            sss_m_ad(:,:) = sss_m_ad(:,:) * zcoef           ! mean SSS             [psu]
            ssu_m_ad(:,:) = ssu_m_ad(:,:) * zcoef           ! mean suface current  [m/s]
            ssv_m_ad(:,:) = ssv_m_ad(:,:) * zcoef           !
            !
         ENDIF
         !                                                ! ---------------------------------------- !
         !                                                !        Cumulate at each time step        !
         !                                                ! ---------------------------------------- !
         un_ad(:,:,1) = ssu_m_ad(:,:) + un_ad(:,:,1)
         vn_ad(:,:,1) = ssv_m_ad(:,:) + vn_ad(:,:,1)
         tn_ad(:,:,1) = sst_m_ad(:,:) + tn_ad(:,:,1)
         sn_ad(:,:,1) = sss_m_ad(:,:) + sn_ad(:,:,1)
         !                                                ! ---------------------------------------- !
         IF( kt == nit000) THEN                           !       Initialisation: 1st time-step      !
            !                                             ! ---------------------------------------- !

            IF( ln_rstart ) THEN 
                  ssu_m_ad(:,:) = 0.0_wp
                  ssv_m_ad(:,:) = 0.0_wp
                  sst_m_ad(:,:) = 0.0_wp
                  sss_m_ad(:,:) = 0.0_wp
             ELSE
               IF(lwp) WRITE(numout,*) '~~~~~~~   mean fields initialised to instantaneous values'
               zcoef = REAL( nn_fsbc - 1, wp )
               un_ad(:,:,1)  = un_ad(:,:,1) + zcoef * ssu_m_ad(:,:)
               vn_ad(:,:,1)  = vn_ad(:,:,1) + zcoef * ssv_m_ad(:,:)
               tn_ad(:,:,1)  = tn_ad(:,:,1) + zcoef * sst_m_ad(:,:)
               sn_ad(:,:,1)  = sn_ad(:,:,1) + zcoef * sss_m_ad(:,:)
               ssu_m_ad(:,:) = 0.0_wp
               ssv_m_ad(:,:) = 0.0_wp
               sst_m_ad(:,:) = 0.0_wp
               sss_m_ad(:,:) = 0.0_wp
            ENDIF
            !                                             ! ---------------------------------------- !
         ELSEIF( MOD( kt - 2 , nn_fsbc ) == 0 ) THEN      !   Initialisation: New mean computation   !
            !                                             ! ---------------------------------------- !
            ssu_m_ad(:,:) = 0.0_wp      ! reset to zero ocean mean sbc fields
            ssv_m_ad(:,:) = 0.0_wp
            sst_m_ad(:,:) = 0.0_wp
            sss_m_ad(:,:) = 0.0_wp
         ENDIF

         !
      ENDIF
      !
   END SUBROUTINE sbc_ssm_adj

   SUBROUTINE sbc_ssm_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE sbc_ssm_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !! ** Action
      !!              dx = ( un_tl, vn_tl, tn_tl, sn_tl )
      !!              dy = ( ssu_m_tl, ssv_m_tl, sst_m_tl, sss_m_tl )
      !!
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!        ! 09-01 (A. Weaver) cleaning
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      INTEGER ::  &
         & ji,    &        ! dummy loop indices
         & jj
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator
      REAL(KIND=wp) :: &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp2            ! scalar product involving the adjoint routine
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zun_tlin   ,     & ! Tangent input
         & zvn_tlin   ,     & ! Tangent input
         & ztn_tlin   ,     & ! Tangent input
         & zsn_tlin   ,     & ! Tangent input
         & zssum_tlin ,     & ! Tangent input
         & zssvm_tlin ,     & ! Tangent input
         & zsstm_tlin ,     & ! Tangent input
         & zsssm_tlin ,     & ! Tangent input
         & zssum_tlout,     & ! Tangent output
         & zssvm_tlout,     & ! Tangent output
         & zsstm_tlout,     & ! Tangent output
         & zsssm_tlout,     & ! Tangent output
         & zun_adout  ,     & ! Adjoint input
         & zvn_adout  ,     & ! Adjoint input
         & ztn_adout  ,     & ! Adjoint input
         & zsn_adout  ,     & ! Adjoint input
         & zssum_adout,     & ! Adjoint input
         & zssvm_adout,     & ! Adjoint input
         & zsstm_adout,     & ! Adjoint input
         & zsssm_adout,     & ! Adjoint input
         & zssum_adin ,     & ! Adjoint output
         & zssvm_adin ,     & ! Adjoint output
         & zsstm_adin ,     & ! Adjoint output
         & zsssm_adin ,     & ! Adjoint output
         & zr                 ! 2D random field 
      CHARACTER(LEN=14) :: cl_name
      ! Allocate memory

      ALLOCATE(                      &
         & zun_tlin   (jpi,jpj),     &
         & zvn_tlin   (jpi,jpj),     &
         & ztn_tlin   (jpi,jpj),     &
         & zsn_tlin   (jpi,jpj),     &
         & zssum_tlin (jpi,jpj),     &
         & zssvm_tlin (jpi,jpj),     &
         & zsstm_tlin (jpi,jpj),     &
         & zsssm_tlin (jpi,jpj),     &
         & zssum_tlout(jpi,jpj),     &
         & zssvm_tlout(jpi,jpj),     &
         & zsstm_tlout(jpi,jpj),     &
         & zsssm_tlout(jpi,jpj),     &
         & zun_adout  (jpi,jpj),     &
         & zvn_adout  (jpi,jpj),     &
         & ztn_adout  (jpi,jpj),     &
         & zsn_adout  (jpi,jpj),     &
         & zssum_adout(jpi,jpj),     &
         & zssvm_adout(jpi,jpj),     &
         & zsstm_adout(jpi,jpj),     &
         & zsssm_adout(jpi,jpj),     &
         & zssum_adin (jpi,jpj),     &
         & zssvm_adin (jpi,jpj),     &
         & zsstm_adin (jpi,jpj),     &
         & zsssm_adin (jpi,jpj),     &
         & zr         (jpi,jpj)      &
         & )

      !==================================================================
      ! 1) dx = ( un_tl, vn_tl, tn_tl, sn_tl ) and 
      !    dy = ( ssu_m_tl, ssv_m_tl, sst_m_tl, sss_m_tl )
      !==================================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
      zun_tlin   (:,:) = 0.0_wp     
      zvn_tlin   (:,:) = 0.0_wp     
      ztn_tlin   (:,:) = 0.0_wp     
      zsn_tlin   (:,:) = 0.0_wp     
      zssum_tlin (:,:) = 0.0_wp     
      zssvm_tlin (:,:) = 0.0_wp     
      zsstm_tlin (:,:) = 0.0_wp     
      zsssm_tlin (:,:) = 0.0_wp     
      zssum_tlout(:,:) = 0.0_wp     
      zssvm_tlout(:,:) = 0.0_wp     
      zsstm_tlout(:,:) = 0.0_wp     
      zsssm_tlout(:,:) = 0.0_wp     
      zun_adout  (:,:) = 0.0_wp     
      zvn_adout  (:,:) = 0.0_wp     
      ztn_adout  (:,:) = 0.0_wp     
      zsn_adout  (:,:) = 0.0_wp     
      zssum_adout(:,:) = 0.0_wp     
      zssvm_adout(:,:) = 0.0_wp     
      zsstm_adout(:,:) = 0.0_wp     
      zsssm_adout(:,:) = 0.0_wp     
      zssum_adin (:,:) = 0.0_wp     
      zssvm_adin (:,:) = 0.0_wp     
      zsstm_adin (:,:) = 0.0_wp     
      zsssm_adin (:,:) = 0.0_wp     
      zr         (:,:) = 0.0_wp      

      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'U', 0.0_wp, stdu )
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zun_tlin(ji,jj) = zr(ji,jj) 
         END DO
      END DO 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 234789 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'V', 0.0_wp, stdv ) 
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zvn_tlin(ji,jj) = zr(ji,jj) 
         END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 471426 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stds )
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zsn_tlin(ji,jj) = zr(ji,jj) 
         END DO
      END DO  

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 153859 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stdt )
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            ztn_tlin(ji,jj) = zr(ji,jj) 
         END DO
      END DO 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 572840 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'U', 0.0_wp, stdu )
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zssum_tlin(ji,jj) = zr(ji,jj) 
         END DO
      END DO 
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 225179 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'V', 0.0_wp, stdv )
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zssvm_tlin(ji,jj) = zr(ji,jj) 
         END DO
      END DO 
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 264801 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stdt ) 
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zsstm_tlin(ji,jj) = zr(ji,jj) 
         END DO
      END DO 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 142859 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stds ) 
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zsssm_tlin(ji,jj) = zr(ji,jj) 
         END DO
      END DO 

      un_tl (:,:,1) = zun_tlin  (:,:) 
      vn_tl (:,:,1) = zvn_tlin  (:,:) 
      tn_tl (:,:,1) = ztn_tlin  (:,:) 
      sn_tl (:,:,1) = zsn_tlin  (:,:) 
      ssu_m_tl(:,:) = zssum_tlin(:,:) 
      ssv_m_tl(:,:) = zssvm_tlin(:,:)  
      sst_m_tl(:,:) = zsstm_tlin(:,:) 
      sss_m_tl(:,:) = zsssm_tlin(:,:)
      
      CALL sbc_ssm_tan( nit000 + 1 )

      zssum_tlout (:,:) = ssu_m_tl(:,:)
      zssvm_tlout (:,:) = ssv_m_tl(:,:)
      zsstm_tlout (:,:) = sst_m_tl(:,:)
      zsssm_tlout (:,:) = sss_m_tl(:,:)

      !--------------------------------------------------------------------
      ! Initialize the adjoint variables: dy^* = W dy
      !--------------------------------------------------------------------

      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zssum_adin(ji,jj) = zssum_tlout(ji,jj) &
               &               * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,1) &
               &               * umask(ji,jj,1)
            zssvm_adin(ji,jj) = zssvm_tlout(ji,jj) &
               &               * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,1) &
               &               * umask(ji,jj,1)
            zsstm_adin(ji,jj) = zsstm_tlout(ji,jj) &
               &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,1) &
               &               * tmask(ji,jj,1) * wesp_t(1)
            zsssm_adin(ji,jj) = zsssm_tlout(ji,jj) &
               &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,1) &
               &               * tmask(ji,jj,1) * wesp_s(1)
         END DO
      END DO

      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1 = DOT_PRODUCT( zssum_tlout, zssum_adin ) &
         & + DOT_PRODUCT( zssvm_tlout, zssvm_adin ) &
         & + DOT_PRODUCT( zsstm_tlout, zsstm_adin ) &
         & + DOT_PRODUCT( zsssm_tlout, zsssm_adin )
 
      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------

      ssu_m_ad(:,:) = zssum_adin(:,:) 
      ssv_m_ad(:,:) = zssvm_adin(:,:)  
      sst_m_ad(:,:) = zsstm_adin(:,:) 
      sss_m_ad(:,:) = zsssm_adin(:,:)
      un_ad(:,:,1)  = 0.0_wp     
      vn_ad(:,:,1)  = 0.0_wp     
      tn_ad(:,:,1)  = 0.0_wp     
      sn_ad(:,:,1)  = 0.0_wp     

      CALL sbc_ssm_adj( nit000 + 1 )

      zun_adout  (:,:) = un_ad(:,:,1) 
      zvn_adout  (:,:) = vn_ad(:,:,1) 
      ztn_adout  (:,:) = tn_ad(:,:,1) 
      zsn_adout  (:,:) = sn_ad(:,:,1) 
      zssum_adout(:,:) = ssu_m_ad(:,:) 
      zssvm_adout(:,:) = ssv_m_ad(:,:)
      zsstm_adout(:,:) = sst_m_ad(:,:)
      zsssm_adout(:,:) = sss_m_ad(:,:)

      !--------------------------------------------------------------------
      ! Compute the scalar product: dx^T dx^*
      !--------------------------------------------------------------------

      zsp2 = DOT_PRODUCT( zun_tlin  , zun_adout   ) &
         & + DOT_PRODUCT( zvn_tlin  , zvn_adout   ) &
         & + DOT_PRODUCT( ztn_tlin  , ztn_adout   ) &
         & + DOT_PRODUCT( zsn_tlin  , zsn_adout   ) &
         & + DOT_PRODUCT( zssum_tlin, zssum_adout ) &
         & + DOT_PRODUCT( zssvm_tlin, zssvm_adout ) &
         & + DOT_PRODUCT( zsstm_tlin, zsstm_adout ) &
         & + DOT_PRODUCT( zsssm_tlin, zsssm_adout )

      ! 14 char:'12345678901234'
      cl_name = 'sbc_ssm_adj   '
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      DEALLOCATE(           &
         & zun_tlin   ,     & ! Tangent input
         & zvn_tlin   ,     & ! Tangent input
         & ztn_tlin   ,     & ! Tangent input
         & zsn_tlin   ,     & ! Tangent input
         & zssum_tlin ,     & ! Tangent input
         & zssvm_tlin ,     & ! Tangent input
         & zsstm_tlin ,     & ! Tangent input
         & zsssm_tlin ,     & ! Tangent input
         & zssum_tlout,     & ! Tangent output
         & zssvm_tlout,     & ! Tangent output
         & zsstm_tlout,     & ! Tangent output
         & zsssm_tlout,     & ! Tangent output
         & zun_adout  ,     & ! Adjoint input
         & zvn_adout  ,     & ! Adjoint input
         & ztn_adout  ,     & ! Adjoint input
         & zsn_adout  ,     & ! Adjoint input
         & zssum_adout,     & ! Adjoint input
         & zssvm_adout,     & ! Adjoint input
         & zsstm_adout,     & ! Adjoint input
         & zsssm_adout,     & ! Adjoint input
         & zssum_adin ,     & ! Adjoint output
         & zssvm_adin ,     & ! Adjoint output
         & zsstm_adin ,     & ! Adjoint output
         & zsssm_adin ,     & ! Adjoint output
         & zr               &
         & )

   END SUBROUTINE sbc_ssm_adj_tst
#endif
   !!======================================================================
END MODULE sbcssm_tam
