MODULE sol_oce_tam
   !!----------------------------------------------------------------------
   !!    This software is governed by the CeCILL licence (Version 2)
   !!----------------------------------------------------------------------
   !!======================================================================
   !!                       ***  MODULE sol_oce_tam ***
   !! NEMOVAR : variables controlling the solver for tangent and adjoint models
   !!======================================================================
   !! History of the direct module:
   !!            9.0  !  02-11  (G. Madec)  F90: Free form and module
   !! History of the TAM module:
   !!            9.0  !  09-03  (A. Weaver) original version
   !!----------------------------------------------------------------------
   !! * Modules used
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE in_out_manager, ONLY: & ! I/O manager 
      & nit000,              &
      & nitend
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 & 
      & jpr2di,              &
      & jpr2dj
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! AW: This is the maximum number of iterations used in the balance operator
!!      for a different elliptic equation than the one used in TAM.
!!      Not a good idea to use the same parameter for both.
!!
   USE tamctl        , ONLY: & ! NEMOVAR variables for controlling the run
      & nmax_fs
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   USE solver        , ONLY: & ! Solver
      & solver_init

   !! * Routine accessibility
   
   IMPLICIT NONE

   !! * Accessibility
   PUBLIC &
      & sol_oce_tam_init, & !: routine called by nemovar.F90
                    !
      & gcx_tl,   & !: Tangent of now solution of the elliptic equation
      & gcxb_tl,  & !: Tangent of before solution of the elliptic equation
      & gcb_tl,   & !: Tangent of 2nd member of barotropic linear system
      & gcr_tl,   & !: Tangent of residual =b-a.x
      & gcx_ad,   & !: Adjoint of solution of the elliptic equation
      & gcxb_ad,  & !: Adjoint of before solution of the elliptic equation
      & gcb_ad      !: Adjoint of 2nd member of barotropic linear system

   !! * Module variables
   INTEGER, DIMENSION(:), ALLOCATABLE :: &
      & nitsor      !: Number of SOR iterations

   REAL(wp), DIMENSION(:,:), ALLOCATABLE :: &  
      & gcx_tl,   & !: Tangent of now solution of the elliptic equation
      & gcxb_tl,  & !: Tangent of before solution of the elliptic equation
      & gcb_tl,   & !: Tangent of 2nd member of barotropic linear system
      & gcr_tl,   & !: Tangent of residual =b-a.x
      & gcx_ad,   & !: Adjoint of solution of the elliptic equation
      & gcxb_ad,  & !: Adjoint of before solution of the elliptic equation
      & gcb_ad      !: Adjoint of 2nd member of barotropic linear system

CONTAINS

   SUBROUTINE sol_oce_tam_init( kindic )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE sol_oce_tam_init  ***
      !!
      !! ** Purpose : 
      !!
      !! ** Method  : kindic = 0  allocate/initialize both tl and ad variables
      !!                          and allocate/initialize solver
      !!              kindic = 1  allocate/initialize only tl variables
      !!              kindic = 2  allocate/initialize only ad variables
      !!
      !! ** Action  :
      !!                   
      !! References : 
      !!
      !! History :
      !!        ! 07-09 (K. Mogensen) Initial version
      !!        ! 09-03 (A. Weaver) Allocate and intialization added
      !!-----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kindic        ! indicate which variables to allocate/initialize

      IF ( kindic == 0 .OR. kindic == 1 ) THEN

         IF ( .NOT. ALLOCATED(nitsor) ) THEN

            ALLOCATE( nitsor( nitend - nit000 + 1 ) )

         ENDIF

         nitsor(:) = nmax_fs

      IF ( kindic == 0 ) CALL solver_init( nit000 )

      ENDIF
      
      IF ( kindic == 0 .OR. kindic == 1 ) THEN

         IF (.NOT. ALLOCATED(gcx_tl) ) THEN
         
            ALLOCATE( gcx_tl (1-jpr2di:jpi+jpr2di,1-jpr2dj:jpj+jpr2dj) )

         ENDIF
         IF (.NOT. ALLOCATED(gcxb_tl) ) THEN

            ALLOCATE( gcxb_tl(1-jpr2di:jpi+jpr2di,1-jpr2dj:jpj+jpr2dj) )

         ENDIF
         IF (.NOT. ALLOCATED(gcb_tl) ) THEN

            ALLOCATE( gcb_tl (1-jpr2di:jpi+jpr2di,1-jpr2dj:jpj+jpr2dj) )

         ENDIF
         IF (.NOT. ALLOCATED(gcr_tl) ) THEN

            ALLOCATE( gcr_tl (1-jpr2di:jpi+jpr2di,1-jpr2dj:jpj+jpr2dj) )

         ENDIF

         ! Initialize tangent linear variables arrays to zero
         ! --------------------------------------------------

         gcx_tl (:,:) = 0.0_wp
         gcxb_tl(:,:) = 0.0_wp
         gcb_tl (:,:) = 0.0_wp
         gcr_tl (:,:) = 0.0_wp

      ENDIF

      IF ( kindic == 0 .OR. kindic == 2 ) THEN

         IF (.NOT. ALLOCATED(gcx_ad) ) THEN
         
            ALLOCATE( gcx_ad (1-jpr2di:jpi+jpr2di,1-jpr2dj:jpj+jpr2dj) )

         ENDIF
         IF (.NOT. ALLOCATED(gcxb_ad) ) THEN

            ALLOCATE( gcxb_ad(1-jpr2di:jpi+jpr2di,1-jpr2dj:jpj+jpr2dj) )

         ENDIF
         IF (.NOT. ALLOCATED(gcb_ad) ) THEN
 
            ALLOCATE( gcb_ad (1-jpr2di:jpi+jpr2di,1-jpr2dj:jpj+jpr2dj) )

         ENDIF

         ! Initialize adjoint variables arrays to zero
         ! -------------------------------------------

         gcx_ad (:,:) = 0.0_wp
         gcxb_ad(:,:) = 0.0_wp
         gcb_ad (:,:) = 0.0_wp

      ENDIF

   END SUBROUTINE sol_oce_tam_init

END MODULE sol_oce_tam
