MODULE traldf_iso_tam
#if defined key_tam
   !!======================================================================
   !!                   ***  MODULE  traldf_iso_tam  ***
   !! Ocean active tracers:  horizontal component of the lateral tracer mixing trend
   !!                        Tangent and adjoint module
   !!======================================================================
   !! History of the direct module:
   !!                  !  94-08  (G. Madec, M. Imbard)
   !!                  !  97-05  (G. Madec)  split into traldf and trazdf
   !!             8.5  !  02-08  (G. Madec)  Free form, F90
   !!             9.0  !  05-11  (G. Madec)  merge traldf and trazdf :-)
   !! History of the T&A module:
   !!             9.0  !  08-12  (A. Vidard) original version
   !!              -   !  09-01  (A. Weaver) misc. bug fixes
   !!----------------------------------------------------------------------
# if   defined key_ldfslp   ||   defined key_esopa
   !!----------------------------------------------------------------------
   !!   'key_ldfslp'               slope of the lateral diffusive direction
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   tra_ldf_iso  : update the tracer trend with the horizontal 
   !!                  component of a iso-neutral laplacian operator
   !!                  and with the vertical part of
   !!                  the isopycnal or geopotential s-coord. operator 
   !!----------------------------------------------------------------------
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpiglo,              &
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 &
      & jpim1,               &
      & jpjm1,               &
      & jpkm1
   USE oce_tam       , ONLY: & ! ocean dynamics and active tracers
      & tb_tl,               &
      & sb_tl,               &
      & ta_tl,               &
      & sa_tl,               &
      & gtu_tl,              &
      & gsu_tl,              &
      & gtv_tl,              &
      & gsv_tl,              &
      & tb_ad,               &
      & sb_ad,               &
      & ta_ad,               &
      & sa_ad,               &
      & gtu_ad,              &
      & gsu_ad,              &
      & gtv_ad,              &
      & gsv_ad
   USE dom_oce       , ONLY: & ! ocean space and time domain
      & e1u,                 &
      & e2u,                 &
      & e1v,                 &
      & e2v,                 &
      & e1t,                 &
      & e2t,                 &
#if defined key_zco
      & e3t_0,               &
#else
      & e3u,                 &
      & e3v,                 &
      & e3t,                 &
#endif
      & tmask,               &
      & umask,               &
      & vmask,               &
      & mbathy,              &
      & ln_zps,              &
      & mig,                 &
      & mjg,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej
   USE ldftra_oce    , ONLY: & ! ocean active tracers: lateral physics
      & ahtv,                &
      & ahtu,                &
      & ahtw,                &
      & ahtb0,               &
      & aht0
!   USE zdf_oce         ! ocean vertical physics
   USE in_out_manager, ONLY: & ! I/O manager
      & lwp,                 &
      & numout,              &
      & nitend,              &
      & nit000
   USE ldfslp        , ONLY: & ! iso-neutral slopes
      & uslp,                &
      & vslp,                &
      & wslpi,               &
      & wslpj
   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld    , ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE tstool_tam    , ONLY: &
      & prntst_adj,          & !
      & stdt,                & ! stdev for u-velocity
      & stds                   !           v-velocity
   USE paresp        , ONLY: & ! Weights for an energy-type scalar product
      & wesp_t,              &
      & wesp_s

   IMPLICIT NONE
   PRIVATE

   PUBLIC   tra_ldf_iso_tan     ! routine called by tralfd_tam.F90
   PUBLIC   tra_ldf_iso_adj     ! routine called by traldf_tam.F90
   PUBLIC   tra_ldf_iso_adj_tst ! routine called by traldf_tam.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "ldftra_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE tra_ldf_iso_tan( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_ldf_iso_tan  ***
      !!
      !! ** Purpose of the direct routine:
      !!      Compute the before horizontal tracer (t & s) diffusive 
      !!      trend for a laplacian tensor (ezxcept the dz[ dz[.] ] term) and 
      !!      add it to the general trend of tracer equation.
      !!
      !! ** Method of the direct routine:
      !!      The horizontal component of the lateral diffusive trends 
      !!      is provided by a 2nd order operator rotated along neural or geopo-
      !!      tential surfaces to which an eddy induced advection can be added
      !!      It is computed using before fields (forward in time) and isopyc-
      !!      nal or geopotential slopes computed in routine ldfslp.
      !!
      !!      1st part :  masked horizontal derivative of T & S ( di[ t ] )
      !!      ========    with partial cell update if ln_zps=T.
      !!
      !!      2nd part :  horizontal fluxes of the lateral mixing operator
      !!      ========    
      !!         zftu = (aht+ahtb0) e2u*e3u/e1u di[ tb ]
      !!               - aht       e2u*uslp    dk[ mi(mk(tb)) ]
      !!         zftv = (aht+ahtb0) e1v*e3v/e2v dj[ tb ]
      !!               - aht       e2u*vslp    dk[ mj(mk(tb)) ]
      !!      take the horizontal divergence of the fluxes:
      !!         difft = 1/(e1t*e2t*e3t) {  di-1[ zftu ] +  dj-1[ zftv ]  }
      !!      Add this trend to the general trend (ta,sa):
      !!         ta = ta + difft
      !!
      !!      3rd part: vertical trends of the lateral mixing operator
      !!      ========  (excluding the vertical flux proportional to dk[t] )
      !!      vertical fluxes associated with the rotated lateral mixing:
      !!         zftw =-aht {  e2t*wslpi di[ mi(mk(tb)) ]
      !!                     + e1t*wslpj dj[ mj(mk(tb)) ]  }
      !!      take the horizontal divergence of the fluxes:
      !!         difft = 1/(e1t*e2t*e3t) dk[ zftw ]
      !!      Add this trend to the general trend (ta,sa):
      !!         ta = ta + difft
      !!
      !! ** Action :   Update (ta,sa) arrays with the before rotated diffusion
      !!            trend (except the dk[ dk[.] ] term)
      !!----------------------------------------------------------------------

      INTEGER, INTENT( in ) ::   kt    ! ocean time-step index
      !!
      INTEGER  ::   ji, jj, jk   ! dummy loop indices
      INTEGER  ::   iku, ikv     ! temporary integer
      REAL(wp) ::   zmsku, zabe1, zcof1, zcoef3, ztatl   ! temporary scalars
      REAL(wp) ::   zmskv, zabe2, zcof2, zcoef4, zsatl   !    "         "
      REAL(wp) ::   zcoef0, zbtr                       !    "         "
      REAL(wp), DIMENSION(jpi,jpj)     ::   zdkttl , zdk1ttl, zftutl, zftvtl   ! 2D workspace
      REAL(wp), DIMENSION(jpi,jpj)     ::   zdkstl , zdk1stl, zfsutl, zfsvtl   !  "     "
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zdittl, zdjttl, ztfwtl     ! 3D workspace
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zdistl, zdjstl, zsfwtl     !  "      "
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_ldf_iso_tan : rotated laplacian diffusion operator'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~~'
      ENDIF

      !!----------------------------------------------------------------------
      !!   I - masked horizontal derivative of T & S
      !!----------------------------------------------------------------------
!!bug ajout.... why?   ( 1,jpj,:) and (jpi,1,:) should be sufficient....
      zdittl (1,:,:) = 0.e0     ;     zdittl (jpi,:,:) = 0.e0
      zdistl (1,:,:) = 0.e0     ;     zdistl (jpi,:,:) = 0.e0
      zdjttl (1,:,:) = 0.e0     ;     zdjttl (jpi,:,:) = 0.e0
      zdjstl (1,:,:) = 0.e0     ;     zdjstl (jpi,:,:) = 0.e0
!!end

      ! Horizontal temperature and salinity gradient 
      DO jk = 1, jpkm1
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               zdittl(ji,jj,jk) = ( tb_tl(ji+1,jj  ,jk) - tb_tl(ji,jj,jk) ) * umask(ji,jj,jk)
               zdistl(ji,jj,jk) = ( sb_tl(ji+1,jj  ,jk) - sb_tl(ji,jj,jk) ) * umask(ji,jj,jk)
               zdjttl(ji,jj,jk) = ( tb_tl(ji  ,jj+1,jk) - tb_tl(ji,jj,jk) ) * vmask(ji,jj,jk)
               zdjstl(ji,jj,jk) = ( sb_tl(ji  ,jj+1,jk) - sb_tl(ji,jj,jk) ) * vmask(ji,jj,jk)
            END DO
         END DO
      END DO
      IF( ln_zps ) THEN      ! partial steps correction at the last level 
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               ! last level
               iku = MIN( mbathy(ji,jj), mbathy(ji+1,jj  ) ) - 1
               ikv = MIN( mbathy(ji,jj), mbathy(ji  ,jj+1) ) - 1
               zdittl(ji,jj,iku) = gtu_tl(ji,jj) 
               zdistl(ji,jj,iku) = gsu_tl(ji,jj)               
               zdjttl(ji,jj,ikv) = gtv_tl(ji,jj) 
               zdjstl(ji,jj,ikv) = gsv_tl(ji,jj)               
            END DO
         END DO
      ENDIF

      !!----------------------------------------------------------------------
      !!   II - horizontal trend of T & S (full)
      !!----------------------------------------------------------------------
      
      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
         ! 1. Vertical tracer gradient at level jk and jk+1
         ! ------------------------------------------------
         ! surface boundary condition: zdkt(jk=1)=zdkt(jk=2)

         zdk1ttl(:,:) = ( tb_tl(:,:,jk) - tb_tl(:,:,jk+1) ) * tmask(:,:,jk+1)
         zdk1stl(:,:) = ( sb_tl(:,:,jk) - sb_tl(:,:,jk+1) ) * tmask(:,:,jk+1)

         IF( jk == 1 ) THEN
            zdkttl(:,:) = zdk1ttl(:,:)
            zdkstl(:,:) = zdk1stl(:,:)
         ELSE
            zdkttl(:,:) = ( tb_tl(:,:,jk-1) - tb_tl(:,:,jk) ) * tmask(:,:,jk)
            zdkstl(:,:) = ( sb_tl(:,:,jk-1) - sb_tl(:,:,jk) ) * tmask(:,:,jk)
         ENDIF

         ! 2. Horizontal fluxes
         ! --------------------

         DO jj = 1 , jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.

               zabe1 = ( fsahtu(ji,jj,jk) + ahtb0 ) * e2u(ji,jj) * fse3u(ji,jj,jk) / e1u(ji,jj)
               zabe2 = ( fsahtv(ji,jj,jk) + ahtb0 ) * e1v(ji,jj) * fse3v(ji,jj,jk) / e2v(ji,jj)

               zmsku = 1.0_wp / MAX(  tmask(ji+1,jj,jk  ) + tmask(ji,jj,jk+1)   &
                  &                 + tmask(ji+1,jj,jk+1) + tmask(ji,jj,jk  ), 1.0_wp )

               zmskv = 1.0_wp / MAX(  tmask(ji,jj+1,jk  ) + tmask(ji,jj,jk+1)   &
                  &                 + tmask(ji,jj+1,jk+1) + tmask(ji,jj,jk  ), 1.0_wp )

               ! *** NOTE ***  uslp() and vslp() are not linearized.

               zcof1 = -fsahtu(ji,jj,jk) * e2u(ji,jj) * uslp(ji,jj,jk) * zmsku
               zcof2 = -fsahtv(ji,jj,jk) * e1v(ji,jj) * vslp(ji,jj,jk) * zmskv

               zftutl(ji,jj) = (  zabe1 * zdittl(ji,jj,jk)   &
                  &              + zcof1 * (  zdkttl (ji+1,jj) + zdk1ttl(ji,jj)      &
                  &                         + zdk1ttl(ji+1,jj) + zdkttl (ji,jj)  )  ) * umask(ji,jj,jk)
               zftvtl(ji,jj) = (  zabe2 * zdjttl(ji,jj,jk)   &
                  &              + zcof2 * (  zdkttl (ji,jj+1) + zdk1ttl(ji,jj)      &
                  &                         + zdk1ttl(ji,jj+1) + zdkttl (ji,jj)  )  ) * vmask(ji,jj,jk)
               zfsutl(ji,jj) = (  zabe1 * zdistl(ji,jj,jk)   &
                  &              + zcof1 * (  zdkstl (ji+1,jj) + zdk1stl(ji,jj)      &
                  &                         + zdk1stl(ji+1,jj) + zdkstl (ji,jj)  )  ) * umask(ji,jj,jk)
               zfsvtl(ji,jj) = (  zabe2 * zdjstl(ji,jj,jk)   &
                  &              + zcof2 * (  zdkstl (ji,jj+1) + zdk1stl(ji,jj)      &
                  &                         + zdk1stl(ji,jj+1) + zdkstl (ji,jj)  )  ) * vmask(ji,jj,jk)

            END DO
         END DO


         ! II.4 Second derivative (divergence) and add to the general trend
         ! ----------------------------------------------------------------
         DO jj = 2 , jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zbtr= 1.0_wp / ( e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) )
               ztatl = zbtr * (   zftutl(ji,jj) - zftutl(ji-1,jj  ) &
                  &             + zftvtl(ji,jj) - zftvtl(ji  ,jj-1)  )
               zsatl = zbtr * (   zfsutl(ji,jj) - zfsutl(ji-1,jj  ) &
                  &             + zfsvtl(ji,jj) - zfsvtl(ji  ,jj-1)  )

               ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) + ztatl
               sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk) + zsatl
            END DO
         END DO
         !                                          ! ===============
      END DO                                        !   End of slab  
      !                                             ! ===============

      !!----------------------------------------------------------------------
      !!   III - vertical trend of T & S (extra diagonal terms only)
      !!----------------------------------------------------------------------

      ! Local constant initialization
      ! -----------------------------
      ztfwtl(1,:,:) = 0.0_wp     ;     ztfwtl(jpi,:,:) = 0.0_wp
      zsfwtl(1,:,:) = 0.0_wp     ;     zsfwtl(jpi,:,:) = 0.0_wp


      ! Vertical fluxes
      ! ---------------

      ! Surface and bottom vertical fluxes set to zero
      ztfwtl(:,:, 1 ) = 0.0_wp   ;     ztfwtl(:,:,jpk) = 0.0_wp
      zsfwtl(:,:, 1 ) = 0.0_wp   ;     zsfwtl(:,:,jpk) = 0.0_wp

      ! interior (2=<jk=<jpk-1)
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zcoef0 = - fsahtw(ji,jj,jk) * tmask(ji,jj,jk)

               zmsku = 1.0_wp / MAX(   umask(ji  ,jj,jk-1) + umask(ji-1,jj,jk)      &
                  &                  + umask(ji-1,jj,jk-1) + umask(ji  ,jj,jk), 1.0_wp  )

               zmskv = 1.0_wp / MAX(   vmask(ji,jj  ,jk-1) + vmask(ji,jj-1,jk)      &
                  &                  + vmask(ji,jj-1,jk-1) + vmask(ji,jj  ,jk), 1.0_wp  )

               ! *** NOTE ***  wslpi() and wslpj() are not linearized.

               zcoef3 = zcoef0 * e2t(ji,jj) * zmsku * wslpi(ji,jj,jk)
               zcoef4 = zcoef0 * e1t(ji,jj) * zmskv * wslpj(ji,jj,jk)

               ztfwtl(ji,jj,jk) = zcoef3 * ( zdittl(ji  ,jj  ,jk-1) + zdittl(ji-1,jj  ,jk)      &
                  &                        + zdittl(ji-1,jj  ,jk-1) + zdittl(ji  ,jj  ,jk)  )   &
                  &             + zcoef4 * ( zdjttl(ji  ,jj  ,jk-1) + zdjttl(ji  ,jj-1,jk)      &
                  &                        + zdjttl(ji  ,jj-1,jk-1) + zdjttl(ji  ,jj  ,jk)  )

               zsfwtl(ji,jj,jk) = zcoef3 * ( zdistl(ji  ,jj  ,jk-1) + zdistl(ji-1,jj  ,jk)      &
                  &                        + zdistl(ji-1,jj  ,jk-1) + zdistl(ji  ,jj  ,jk)  )   &
                  &             + zcoef4 * ( zdjstl(ji  ,jj  ,jk-1) + zdjstl(ji  ,jj-1,jk)      &
                  &                        + zdjstl(ji  ,jj-1,jk-1) + zdjstl(ji  ,jj  ,jk)  )
            END DO
         END DO
      END DO


      ! I.5 Divergence of vertical fluxes added to the general tracer trend
      ! -------------------------------------------------------------------

      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zbtr =  1.0_wp / ( e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) )

               ztatl = ( ztfwtl(ji,jj,jk) - ztfwtl(ji,jj,jk+1) ) * zbtr
               zsatl = ( zsfwtl(ji,jj,jk) - zsfwtl(ji,jj,jk+1) ) * zbtr

               ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) + ztatl
               sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk) + zsatl
            END DO
         END DO
      END DO
      !
   END SUBROUTINE tra_ldf_iso_tan

   SUBROUTINE tra_ldf_iso_adj( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_ldf_iso_adj  ***
      !!
      !! ** Purpose of the direct routine:
      !!      Compute the before horizontal tracer (t & s) diffusive 
      !!      trend for a laplacian tensor (ezxcept the dz[ dz[.] ] term) and 
      !!      add it to the general trend of tracer equation.
      !!
      !! ** Method of the direct routine:
      !!      The horizontal component of the lateral diffusive trends 
      !!      is provided by a 2nd order operator rotated along neural or geopo-
      !!      tential surfaces to which an eddy induced advection can be added
      !!      It is computed using before fields (forward in time) and isopyc-
      !!      nal or geopotential slopes computed in routine ldfslp.
      !!
      !!      1st part :  masked horizontal derivative of T & S ( di[ t ] )
      !!      ========    with partial cell update if ln_zps=T.
      !!
      !!      2nd part :  horizontal fluxes of the lateral mixing operator
      !!      ========    
      !!         zftu = (aht+ahtb0) e2u*e3u/e1u di[ tb ]
      !!               - aht       e2u*uslp    dk[ mi(mk(tb)) ]
      !!         zftv = (aht+ahtb0) e1v*e3v/e2v dj[ tb ]
      !!               - aht       e2u*vslp    dk[ mj(mk(tb)) ]
      !!      take the horizontal divergence of the fluxes:
      !!         difft = 1/(e1t*e2t*e3t) {  di-1[ zftu ] +  dj-1[ zftv ]  }
      !!      Add this trend to the general trend (ta,sa):
      !!         ta = ta + difft
      !!
      !!      3rd part: vertical trends of the lateral mixing operator
      !!      ========  (excluding the vertical flux proportional to dk[t] )
      !!      vertical fluxes associated with the rotated lateral mixing:
      !!         zftw =-aht {  e2t*wslpi di[ mi(mk(tb)) ]
      !!                     + e1t*wslpj dj[ mj(mk(tb)) ]  }
      !!      take the horizontal divergence of the fluxes:
      !!         difft = 1/(e1t*e2t*e3t) dk[ zftw ]
      !!      Add this trend to the general trend (ta,sa):
      !!         ta = ta + difft
      !!
      !! ** Action :   Update (ta,sa) arrays with the before rotated diffusion
      !!            trend (except the dk[ dk[.] ] term)
      !!----------------------------------------------------------------------

      INTEGER, INTENT( in ) ::   kt    ! ocean time-step index
      !!
      INTEGER  ::   ji, jj, jk   ! dummy loop indices
      INTEGER  ::   iku, ikv     ! temporary integer
      REAL(wp) ::   zmsku, zabe1, zcof1, zcoef3, ztaad   ! temporary scalars
      REAL(wp) ::   zmskv, zabe2, zcof2, zcoef4, zsaad   !    "         "
      REAL(wp) ::   ztf3, ztf4, zsf3, zsf4               !
      REAL(wp) ::   zcoef0, zbtr                       !    "         "
      REAL(wp), DIMENSION(jpi,jpj)     ::   zdktad , zdk1tad, zftuad, zftvad   ! 2D workspace
      REAL(wp), DIMENSION(jpi,jpj)     ::   zdksad , zdk1sad, zfsuad, zfsvad   !  "     "
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zditad, zdjtad, ztfwad     ! 3D workspace
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zdisad, zdjsad, zsfwad     !  "      "
      !!----------------------------------------------------------------------

      zditad(:,:,:) = 0.0_wp ;  zdjtad(:,:,:) = 0.0_wp ;  ztfwad(:,:,:) = 0.0_wp
      zdisad(:,:,:) = 0.0_wp ;  zdjsad(:,:,:) = 0.0_wp ;  zsfwad(:,:,:) = 0.0_wp

      zdktad(:,:) = 0.0_wp ;  zdk1tad(:,:) = 0.0_wp   
      zdksad(:,:) = 0.0_wp ;  zdk1sad(:,:) = 0.0_wp 

      zftuad(:,:) = 0.0_wp ;  zftvad (:,:) = 0.0_wp 
      zfsuad(:,:) = 0.0_wp ;  zfsvad (:,:) = 0.0_wp

      IF( kt == nitend ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_ldf_iso_adj : rotated laplacian diffusion operator'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~~'
      ENDIF

      !!----------------------------------------------------------------------
      !!   III - vertical trend of T & S (extra diagonal terms only)
      !!----------------------------------------------------------------------
      ! I.5 Divergence of vertical fluxes added to the general tracer trend
      ! -------------------------------------------------------------------

      DO jk = jpkm1, 1, -1
         DO jj = jpjm1, 2, -1
            DO ji = fs_jpim1, fs_2, -1   ! vector opt.
               zbtr =  1.0_wp / ( e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) )
               ztaad = ta_ad(ji,jj,jk) * zbtr
               zsaad = sa_ad(ji,jj,jk) * zbtr

               ztfwad(ji,jj,jk  ) = ztfwad(ji,jj,jk  ) + ztaad
               ztfwad(ji,jj,jk+1) = ztfwad(ji,jj,jk+1) - ztaad
               zsfwad(ji,jj,jk  ) = zsfwad(ji,jj,jk  ) + zsaad
               zsfwad(ji,jj,jk+1) = zsfwad(ji,jj,jk+1) - zsaad
            END DO
         END DO
      END DO
      ! interior (2=<jk=<jpk-1)
      DO jk = jpkm1, 2, -1
         DO jj = jpjm1, 2, -1
            DO ji = fs_jpim1, fs_2, -1   ! vector opt.
               zcoef0 = - fsahtw(ji,jj,jk) * tmask(ji,jj,jk)

               zmsku = 1.0_wp / MAX(   umask(ji  ,jj,jk-1) + umask(ji-1,jj,jk)      &
                  &                  + umask(ji-1,jj,jk-1) + umask(ji  ,jj,jk), 1.0_wp  )

               zmskv = 1.0_wp / MAX(   vmask(ji,jj  ,jk-1) + vmask(ji,jj-1,jk)      &
                  &                  + vmask(ji,jj-1,jk-1) + vmask(ji,jj  ,jk), 1.0_wp  )

               ! *** NOTE ***  wslpi() and wslpj() are not linearized.

               zcoef3 = zcoef0 * e2t(ji,jj) * zmsku * wslpi(ji,jj,jk)
               zcoef4 = zcoef0 * e1t(ji,jj) * zmskv * wslpj(ji,jj,jk)

               ztf3 = ztfwad(ji,jj,jk) * zcoef3
               ztf4 = ztfwad(ji,jj,jk) * zcoef4
               zsf3 = zsfwad(ji,jj,jk) * zcoef3
               zsf4 = zsfwad(ji,jj,jk) * zcoef4

               zditad(ji  ,jj  ,jk-1) = zditad(ji  ,jj  ,jk-1) + ztf3
               zditad(ji-1,jj  ,jk  ) = zditad(ji-1,jj  ,jk  ) + ztf3
               zditad(ji-1,jj  ,jk-1) = zditad(ji-1,jj  ,jk-1) + ztf3
               zditad(ji  ,jj  ,jk  ) = zditad(ji  ,jj  ,jk  ) + ztf3

               zdjtad(ji  ,jj  ,jk-1) = zdjtad(ji  ,jj  ,jk-1) + ztf4
               zdjtad(ji  ,jj-1,jk  ) = zdjtad(ji  ,jj-1,jk  ) + ztf4
               zdjtad(ji  ,jj-1,jk-1) = zdjtad(ji  ,jj-1,jk-1) + ztf4
               zdjtad(ji  ,jj  ,jk  ) = zdjtad(ji  ,jj  ,jk  ) + ztf4

               zdisad(ji  ,jj  ,jk-1) = zdisad(ji  ,jj  ,jk-1) + zsf3
               zdisad(ji-1,jj  ,jk  ) = zdisad(ji-1,jj  ,jk  ) + zsf3
               zdisad(ji-1,jj  ,jk-1) = zdisad(ji-1,jj  ,jk-1) + zsf3
               zdisad(ji  ,jj  ,jk  ) = zdisad(ji  ,jj  ,jk  ) + zsf3

               zdjsad(ji  ,jj  ,jk-1) = zdjsad(ji  ,jj  ,jk-1) + zsf4
               zdjsad(ji  ,jj-1,jk  ) = zdjsad(ji  ,jj-1,jk  ) + zsf4
               zdjsad(ji  ,jj-1,jk-1) = zdjsad(ji  ,jj-1,jk-1) + zsf4
               zdjsad(ji  ,jj  ,jk  ) = zdjsad(ji  ,jj  ,jk  ) + zsf4

               ztfwad(ji,jj,jk) = 0.0_wp
               zsfwad(ji,jj,jk) = 0.0_wp
            END DO
         END DO
      END DO
 
      ! Local constant initialization
      ! -----------------------------
      ztfwad(1,:,:) = 0.0_wp     ;     ztfwad(jpi,:,:) = 0.0_wp
      zsfwad(1,:,:) = 0.0_wp     ;     zsfwad(jpi,:,:) = 0.0_wp

      ! Vertical fluxes
      ! ---------------

      ! Surface and bottom vertical fluxes set to zero
      ztfwad(:,:, 1 ) = 0.0_wp   ;     ztfwad(:,:,jpk) = 0.0_wp
      zsfwad(:,:, 1 ) = 0.0_wp   ;     zsfwad(:,:,jpk) = 0.0_wp

      !!----------------------------------------------------------------------
      !!   II - horizontal trend of T & S (full)
      !!----------------------------------------------------------------------
      
      !                                                ! ===============
      DO jk = jpkm1, 1, -1                             ! Horizontal slab
         !                                             ! ===============
         ! II.4 Second derivative (divergence) and add to the general trend
         ! ----------------------------------------------------------------
         DO jj = jpjm1, 2, -1
            DO ji = fs_jpim1, fs_2, -1   ! vector opt.

               zbtr= 1.0_wp / ( e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) )
               ztaad = ta_ad(ji,jj,jk) * zbtr
               zsaad = sa_ad(ji,jj,jk) * zbtr

               zftuad(ji  ,jj  ) = zftuad(ji  ,jj  ) + ztaad
               zftuad(ji-1,jj  ) = zftuad(ji-1,jj  ) - ztaad
               zftvad(ji  ,jj  ) = zftvad(ji  ,jj  ) + ztaad
               zftvad(ji  ,jj-1) = zftvad(ji  ,jj-1) - ztaad

               zfsuad(ji  ,jj  ) = zfsuad(ji  ,jj  ) + zsaad
               zfsuad(ji-1,jj  ) = zfsuad(ji-1,jj  ) - zsaad
               zfsvad(ji  ,jj  ) = zfsvad(ji  ,jj  ) + zsaad
               zfsvad(ji  ,jj-1) = zfsvad(ji  ,jj-1) - zsaad

            END DO
         END DO

         ! 2. Horizontal fluxes
         ! --------------------
         DO jj = jpjm1, 1, -1
            DO ji = fs_jpim1, 1, -1   ! vector opt.
               zabe1 = umask(ji,jj,jk) * ( fsahtu(ji,jj,jk) + ahtb0 ) &
                  &                    * e2u(ji,jj) * fse3u(ji,jj,jk) / e1u(ji,jj)
               zabe2 = vmask(ji,jj,jk) * ( fsahtv(ji,jj,jk) + ahtb0 ) &
                  &                    * e1v(ji,jj) * fse3v(ji,jj,jk) / e2v(ji,jj)

               zmsku = 1.0_wp / MAX(  tmask(ji+1,jj,jk  ) + tmask(ji,jj,jk+1)   &
                  &                 + tmask(ji+1,jj,jk+1) + tmask(ji,jj,jk  ), 1.0_wp )

               zmskv = 1.0_wp / MAX(  tmask(ji,jj+1,jk  ) + tmask(ji,jj,jk+1)   &
                  &                 + tmask(ji,jj+1,jk+1) + tmask(ji,jj,jk  ), 1.0_wp )

               ! *** NOTE ***  uslp() and vslp() are not linearized.

               zcof1 = -fsahtu(ji,jj,jk) * e2u(ji,jj) * uslp(ji,jj,jk) * zmsku * umask(ji,jj,jk)
               zcof2 = -fsahtv(ji,jj,jk) * e1v(ji,jj) * vslp(ji,jj,jk) * zmskv * vmask(ji,jj,jk)

               zditad(ji,jj,jk) = zditad(ji,jj,jk) + zftuad(ji,jj) * zabe1

               zdktad (ji+1,jj) = zdktad (ji+1,jj) + zftuad(ji,jj) * zcof1
               zdktad (ji  ,jj) = zdktad (ji  ,jj) + zftuad(ji,jj) * zcof1
               zdk1tad(ji  ,jj) = zdk1tad(ji  ,jj) + zftuad(ji,jj) * zcof1
               zdk1tad(ji+1,jj) = zdk1tad(ji+1,jj) + zftuad(ji,jj) * zcof1
               zftuad (ji  ,jj) = 0.0_wp
               !
               zdjtad(ji,jj,jk) = zdjtad(ji,jj,jk) + zftvad(ji,jj) * zabe2

               zdktad (ji,jj+1) = zdktad (ji,jj+1) + zftvad(ji,jj) * zcof2
               zdktad (ji,jj  ) = zdktad (ji,jj  ) + zftvad(ji,jj) * zcof2
               zdk1tad(ji,jj  ) = zdk1tad(ji,jj  ) + zftvad(ji,jj) * zcof2
               zdk1tad(ji,jj+1) = zdk1tad(ji,jj+1) + zftvad(ji,jj) * zcof2
               zftvad (ji,jj  ) = 0.0_wp
               !
               zdisad(ji,jj,jk) = zdisad(ji,jj,jk) + zfsuad(ji,jj) * zabe1

               zdksad (ji+1,jj) = zdksad (ji+1,jj) + zfsuad(ji,jj) * zcof1 
               zdksad (ji  ,jj) = zdksad (ji  ,jj) + zfsuad(ji,jj) * zcof1 
               zdk1sad(ji  ,jj) = zdk1sad(ji  ,jj) + zfsuad(ji,jj) * zcof1
               zdk1sad(ji+1,jj) = zdk1sad(ji+1,jj) + zfsuad(ji,jj) * zcof1 
               zfsuad (ji  ,jj) = 0.0_wp
               !
               zdjsad(ji,jj,jk) = zdjsad(ji,jj,jk) + zfsvad(ji,jj) * zabe2

               zdksad (ji,jj+1) = zdksad (ji,jj+1) + zfsvad(ji,jj) * zcof2
               zdksad (ji,jj  ) = zdksad (ji,jj  ) + zfsvad(ji,jj) * zcof2
               zdk1sad(ji,jj  ) = zdk1sad(ji,jj  ) + zfsvad(ji,jj) * zcof2
               zdk1sad(ji,jj+1) = zdk1sad(ji,jj+1) + zfsvad(ji,jj) * zcof2
               zfsvad (ji,jj  ) = 0.0_wp

            END DO
         END DO

         ! 1. Vertical tracer gradient at level jk and jk+1
         ! ------------------------------------------------
         ! surface boundary condition: zdkt(jk=1)=zdkt(jk=2)

         IF( jk == 1 ) THEN

            zdk1tad(:,:) = zdk1tad(:,:) + zdktad(:,:)
            zdk1sad(:,:) = zdk1sad(:,:) + zdksad(:,:)

            zdktad(:,:)  = 0.0_wp
            zdksad(:,:)  = 0.0_wp

         ELSE

            tb_ad(:,:,jk-1) = tb_ad(:,:,jk-1) + zdktad(:,:) * tmask(:,:,jk)
            tb_ad(:,:,jk  ) = tb_ad(:,:,jk  ) - zdktad(:,:) * tmask(:,:,jk)

            sb_ad(:,:,jk-1) = sb_ad(:,:,jk-1) + zdksad(:,:) * tmask(:,:,jk)
            sb_ad(:,:,jk  ) = sb_ad(:,:,jk  ) - zdksad(:,:) * tmask(:,:,jk)

            zdktad(:,:) = 0.0_wp
            zdksad(:,:) = 0.0_wp

         ENDIF

         tb_ad(:,:,jk  ) = tb_ad(:,:,jk  ) + zdk1tad(:,:) * tmask(:,:,jk+1)
         tb_ad(:,:,jk+1) = tb_ad(:,:,jk+1) - zdk1tad(:,:) * tmask(:,:,jk+1)
         
         sb_ad(:,:,jk  ) = sb_ad(:,:,jk  ) + zdk1sad(:,:) * tmask(:,:,jk+1)
         sb_ad(:,:,jk+1) = sb_ad(:,:,jk+1) - zdk1sad(:,:) * tmask(:,:,jk+1)
         
         zdk1tad(:,:) = 0.0_wp
         zdk1sad(:,:) = 0.0_wp
         !                                          ! ===============
      END DO                                        !   End of slab  
      !                                             ! ===============
      !!----------------------------------------------------------------------
      !!   I - masked horizontal derivative of T & S
      !!----------------------------------------------------------------------
      IF( ln_zps ) THEN      ! partial steps correction at the last level 
         DO jj = jpjm1, 1, -1
            DO ji = fs_jpim1, 1, -1   ! vector opt.

               ! last level
               iku = MIN( mbathy(ji,jj), mbathy(ji+1,jj  ) ) - 1
               ikv = MIN( mbathy(ji,jj), mbathy(ji  ,jj+1) ) - 1

               gtu_ad(ji,jj) = gtu_ad(ji,jj) + zditad(ji,jj,iku)
               gtv_ad(ji,jj) = gtv_ad(ji,jj) + zdjtad(ji,jj,ikv)

               gsu_ad(ji,jj) = gsu_ad(ji,jj) + zdisad(ji,jj,iku)
               gsv_ad(ji,jj) = gsv_ad(ji,jj) + zdjsad(ji,jj,ikv)

               zditad(ji,jj,iku) = 0.0_wp
               zdjtad(ji,jj,ikv) = 0.0_wp

               zdisad(ji,jj,iku) = 0.0_wp
               zdjsad(ji,jj,ikv) = 0.0_wp
 
           END DO
         END DO
      ENDIF

      ! Horizontal temperature and salinity gradient 
      DO jk = jpkm1, 1, -1
         DO jj = jpjm1, 1, -1
            DO ji = fs_jpim1, 1, -1   ! vector opt.

               zditad(ji,jj,jk) = zditad(ji,jj,jk) * umask(ji,jj,jk)
               zdjtad(ji,jj,jk) = zdjtad(ji,jj,jk) * vmask(ji,jj,jk)
               zdisad(ji,jj,jk) = zdisad(ji,jj,jk) * umask(ji,jj,jk)
               zdjsad(ji,jj,jk) = zdjsad(ji,jj,jk) * vmask(ji,jj,jk)

               tb_ad(ji+1,jj  ,jk) = tb_ad(ji+1,jj  ,jk) + zditad(ji,jj,jk)
               tb_ad(ji  ,jj  ,jk) = tb_ad(ji  ,jj  ,jk) - zditad(ji,jj,jk)
               tb_ad(ji  ,jj+1,jk) = tb_ad(ji  ,jj+1,jk) + zdjtad(ji,jj,jk)
               tb_ad(ji  ,jj  ,jk) = tb_ad(ji  ,jj  ,jk) - zdjtad(ji,jj,jk)

               sb_ad(ji+1,jj  ,jk) = sb_ad(ji+1,jj  ,jk) + zdisad(ji,jj,jk)
               sb_ad(ji  ,jj  ,jk) = sb_ad(ji  ,jj  ,jk) - zdisad(ji,jj,jk)
               sb_ad(ji  ,jj+1,jk) = sb_ad(ji  ,jj+1,jk) + zdjsad(ji,jj,jk)
               sb_ad(ji  ,jj  ,jk) = sb_ad(ji  ,jj  ,jk) - zdjsad(ji,jj,jk)

            END DO
         END DO
      END DO
      !
   END SUBROUTINE tra_ldf_iso_adj

   SUBROUTINE tra_ldf_iso_adj_tst ( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE example_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!                   
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      INTEGER ::  &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator
      REAL(KIND=wp) :: &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp1_T,       &
         & zsp1_S,       &
         & zsp2,         & ! scalar product involving the adjoint routine
         & zsp2_1,       &
         & zsp2_2,       &
         & zsp2_3,       &
         & zsp2_4,       &
         & zsp2_5,       &
         & zsp2_6,       &
         & zsp2_7,       &
         & zsp2_8,       &
         & zsp2_T,       &
         & zsp2_S
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & ztb_tlin ,      & ! Tangent input
         & zsb_tlin ,      & ! Tangent input
         & zta_tlin ,      & ! Tangent input
         & zsa_tlin ,      & ! Tangent input
         & zta_tlout,      & ! Tangent output
         & zsa_tlout,      & ! Tangent output
         & zta_adin,       & ! Adjoint input
         & zsa_adin,       & ! Adjoint input
         & ztb_adout ,     & ! Adjoint output
         & zsb_adout ,     & ! Adjoint output
         & zta_adout ,     & ! Adjoint output
         & zsa_adout ,     & ! Adjoint output
         & z3r               ! 3D random field 
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zgtu_tlin ,     & ! Tangent input
         & zgsu_tlin ,     & ! Tangent input
         & zgtv_tlin ,     & ! Tangent input
         & zgsv_tlin ,     & ! Tangent input
         & zgtu_adout ,    & ! Adjoint output
         & zgsu_adout ,    & ! Adjoint output
         & zgtv_adout ,    & ! Adjoint output
         & zgsv_adout ,    & ! Adjoint output
         & z2r               ! 2D random field 
      CHARACTER(LEN=14) :: cl_name
      ! Allocate memory

      ALLOCATE( &
         & ztb_tlin(jpi,jpj,jpk),      & 
         & zsb_tlin(jpi,jpj,jpk),      & 
         & zta_tlin(jpi,jpj,jpk),      & 
         & zsa_tlin(jpi,jpj,jpk),      & 
         & zgtu_tlin(jpi,jpj),         & 
         & zgsu_tlin(jpi,jpj),         & 
         & zgtv_tlin(jpi,jpj),         & 
         & zgsv_tlin(jpi,jpj),         & 
         & zta_tlout(jpi,jpj,jpk),     & 
         & zsa_tlout(jpi,jpj,jpk),     & 
         & zta_adin(jpi,jpj,jpk),      & 
         & zsa_adin(jpi,jpj,jpk),      & 
         & ztb_adout(jpi,jpj,jpk),     & 
         & zsb_adout(jpi,jpj,jpk),     & 
         & zta_adout(jpi,jpj,jpk),     & 
         & zsa_adout(jpi,jpj,jpk),     & 
         & zgtu_adout(jpi,jpj),        & 
         & zgsu_adout(jpi,jpj),        & 
         & zgtv_adout(jpi,jpj),        & 
         & zgsv_adout(jpi,jpj),        & 
         & z3r(jpi,jpj,jpk),           &
         & z2r(jpi,jpj)                &
         & )

      ! Initialize the reference state
      uslp (:,:,:) = 2.0_wp
      vslp (:,:,:) = 3.0_wp
      wslpi(:,:,:) = 4.0_wp
      wslpj(:,:,:) = 5.0_wp

      !=======================================================================
      ! 1) dx = ( tb_tl, ta_tl, sb_tl, sa_tl, gtu_tl, gtv_tl, gsu_tl, gsv_tl )
      !    dy = ( ta_tl, sa_tl )
      !=======================================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------

      ztb_tlin(:,:,:)  = 0.0_wp
      zsb_tlin(:,:,:)  = 0.0_wp
      zta_tlin(:,:,:)  = 0.0_wp
      zsa_tlin(:,:,:)  = 0.0_wp
      zgtu_tlin(:,:)   = 0.0_wp
      zgsu_tlin(:,:)   = 0.0_wp
      zgtv_tlin(:,:)   = 0.0_wp
      zgsv_tlin(:,:)   = 0.0_wp
      zta_tlout(:,:,:) = 0.0_wp
      zsa_tlout(:,:,:) = 0.0_wp
      zta_adin(:,:,:)  = 0.0_wp
      zsa_adin(:,:,:)  = 0.0_wp
      ztb_adout(:,:,:) = 0.0_wp
      zsb_adout(:,:,:) = 0.0_wp
      zta_adout(:,:,:) = 0.0_wp
      zsa_adout(:,:,:) = 0.0_wp
      zgtu_adout(:,:)  = 0.0_wp
      zgsu_adout(:,:)  = 0.0_wp
      zgtv_adout(:,:)  = 0.0_wp
      zgsv_adout(:,:)  = 0.0_wp

      tb_tl(:,:,:) = 0.0_wp
      sb_tl(:,:,:) = 0.0_wp
      ta_tl(:,:,:) = 0.0_wp
      sa_tl(:,:,:) = 0.0_wp
      gtu_tl(:,:)  = 0.0_wp
      gsu_tl(:,:)  = 0.0_wp
      gtv_tl(:,:)  = 0.0_wp
      gsv_tl(:,:)  = 0.0_wp
      tb_ad(:,:,:) = 0.0_wp
      sb_ad(:,:,:) = 0.0_wp
      ta_ad(:,:,:) = 0.0_wp
      sa_ad(:,:,:) = 0.0_wp
      gtu_ad(:,:)  = 0.0_wp
      gsu_ad(:,:)  = 0.0_wp
      gtv_ad(:,:)  = 0.0_wp
      gsv_ad(:,:)  = 0.0_wp

      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stdt )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              ztb_tlin(ji,jj,jk) = z3r(ji,jj,jk) 
            END DO
         END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 727391 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stds ) 
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zsb_tlin(ji,jj,jk) = z3r(ji,jj,jk) 
            END DO
         END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 249741 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stdt )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zta_tlin(ji,jj,jk) = z3r(ji,jj,jk) 
            END DO
         END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 182029 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stds )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zsa_tlin(ji,jj,jk) = z3r(ji,jj,jk) 
            END DO
         END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z2r, 'U', 0.0_wp, stds )
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zgtu_tlin(ji,jj) = z2r(ji,jj) 
         END DO
      END DO 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 727391 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z2r, 'U', 0.0_wp, stds ) 
      DO jj = nldj, nlej
        DO ji = nldi, nlei
           zgsu_tlin(ji,jj) = z2r(ji,jj) 
        END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 249741 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z2r, 'V', 0.0_wp, stds )
      DO jj = nldj, nlej
        DO ji = nldi, nlei
           zgtv_tlin(ji,jj) = z2r(ji,jj) 
        END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 182029 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z2r, 'V', 0.0_wp, stds ) 
      DO jj = nldj, nlej
        DO ji = nldi, nlei
           zgsv_tlin(ji,jj) = z2r(ji,jj) 
        END DO
      END DO

      tb_tl(:,:,:) = ztb_tlin(:,:,:) 
      sb_tl(:,:,:) = zsb_tlin(:,:,:) 
      ta_tl(:,:,:) = zta_tlin(:,:,:) 
      sa_tl(:,:,:) = zsa_tlin(:,:,:) 
      gtu_tl(:,:)  = zgtu_tlin(:,:)
      gsu_tl(:,:)  = zgsu_tlin(:,:)
      gtv_tl(:,:)  = zgtv_tlin(:,:)
      gsv_tl(:,:)  = zgsv_tlin(:,:)

      CALL tra_ldf_iso_tan( nit000 ) 
      
      zta_tlout(:,:,:) = ta_tl(:,:,:)
      zsa_tlout(:,:,:) = sa_tl(:,:,:)

      !--------------------------------------------------------------------
      ! Initialize the adjoint variables: dy^* = W dy
      !--------------------------------------------------------------------

      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zsa_adin(ji,jj,jk) = zsa_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_s(jk)
              zta_adin(ji,jj,jk) = zta_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_t(jk)
            END DO
         END DO
      END DO

      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1_T = DOT_PRODUCT( zta_tlout, zta_adin )
      zsp1_S = DOT_PRODUCT( zsa_tlout, zsa_adin )
      zsp1 = zsp1_T + zsp1_S

      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------

      ta_ad(:,:,:) = zta_adin(:,:,:)
      sa_ad(:,:,:) = zsa_adin(:,:,:)

      CALL tra_ldf_iso_adj( nit000 ) 

      ztb_adout(:,:,:) = tb_ad(:,:,:)
      zsb_adout(:,:,:) = sb_ad(:,:,:)
      zta_adout(:,:,:) = ta_ad(:,:,:)
      zsa_adout(:,:,:) = sa_ad(:,:,:)
      zgtu_adout(:,:)  = gtu_ad(:,:)
      zgsu_adout(:,:)  = gsu_ad(:,:)
      zgtv_adout(:,:)  = gtv_ad(:,:)
      zgsv_adout(:,:)  = gsv_ad(:,:)
          
      zsp2_1 = DOT_PRODUCT( ztb_tlin , ztb_adout  )
      zsp2_2 = DOT_PRODUCT( zta_tlin , zta_adout  )
      zsp2_3 = DOT_PRODUCT( zgtu_tlin, zgtu_adout )
      zsp2_4 = DOT_PRODUCT( zgtv_tlin, zgtv_adout )
      zsp2_5 = DOT_PRODUCT( zsb_tlin , zsb_adout  )
      zsp2_6 = DOT_PRODUCT( zsa_tlin , zsa_adout  )
      zsp2_7 = DOT_PRODUCT( zgsu_tlin, zgsu_adout )
      zsp2_8 = DOT_PRODUCT( zgsv_tlin, zgsv_adout ) 

      zsp2_T = zsp2_1 + zsp2_2 + zsp2_3 + zsp2_4
      zsp2_S = zsp2_5 + zsp2_6 + zsp2_7 + zsp2_8
      zsp2   = zsp2_T + zsp2_S

      cl_name = 'tra_ldf_iso_ad'
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      DEALLOCATE(         &
         & ztb_tlin,      & ! Tangent input
         & zsb_tlin,      & ! Tangent input
         & zta_tlin,      & ! Tangent input
         & zsa_tlin,      & ! Tangent input
         & zgtu_tlin,     & ! Tangent input
         & zgsu_tlin,     & ! Tangent input
         & zgtv_tlin,     & ! Tangent input
         & zgsv_tlin,     & ! Tangent input
         & zta_tlout,     & ! Tangent output
         & zsa_tlout,     & ! Tangent output
         & zta_adin,      & ! Adjoint input
         & zsa_adin,      & ! Adjoint input
         & ztb_adout,     & ! Adjoint output
         & zsb_adout,     & ! Adjoint output
         & zta_adout,     & ! Adjoint output
         & zsa_adout,     & ! Adjoint output
         & zgtu_adout,    & ! Adjoint output
         & zgsu_adout,    & ! Adjoint output
         & zgtv_adout,    & ! Adjoint output
         & zgsv_adout,    & ! Adjoint output
         & z3r,           & ! 3D random field 
         & z2r            &
         & )
      
   END SUBROUTINE tra_ldf_iso_adj_tst
# else
   !!----------------------------------------------------------------------
   !!   default option :   Dummy code   NO rotation of the diffusive tensor
   !!----------------------------------------------------------------------
CONTAINS
   SUBROUTINE tra_ldf_iso_tan( kt )               ! Empty routine
      WRITE(*,*) 'tra_ldf_iso_tan: You should not have seen this print! error?', kt
   END SUBROUTINE tra_ldf_iso_tan
   SUBROUTINE tra_ldf_iso_adj( kt )               ! Empty routine
      WRITE(*,*) 'tra_ldf_iso_adj: You should not have seen this print! error?', kt
   END SUBROUTINE tra_ldf_iso_adj
   SUBROUTINE tra_ldf_iso_adj_tst ( kumadt )
      WRITE(*,*) 'tra_ldf_iso_adj_tst: You should not have seen this print! error?', kt
   END SUBROUTINE tra_ldf_iso_adj_tst
# endif
#endif

   !!==============================================================================
END MODULE traldf_iso_tam
