MODULE trazdf_exp_tam
#ifdef key_tam
   !!==============================================================================
   !!                    ***  MODULE  trazdf_exp_tam  ***
   !! Ocean active tracers:  vertical component of the tracer mixing trend using
   !!                        a split-explicit time-stepping
   !!                        Tangent and Adjoint module 
   !!==============================================================================
   !! History of the direct module :
   !!   OPA           !  1990-10  (B. Blanke)  Original code
   !!            7.0  !  1991-11  (G. Madec)
   !!                 !  1992-06  (M. Imbard)  correction on tracer trend loops
   !!                 !  1996-01  (G. Madec)  statement function for e3
   !!                 !  1997-05  (G. Madec)  vertical component of isopycnal
   !!                 !  1997-07  (G. Madec)  geopotential diffusion in s-coord
   !!                 !  2000-08  (G. Madec)  double diffusive mixing
   !!   NEMO     1.0  !  2002-08  (G. Madec)  F90: Free form and module
   !!             -   !  2004-08  (C. Talandier) New trends organisation
   !!             -   !  2005-11  (G. Madec)  New organisation
   !!            3.0  !  2008-04  (G. Madec)  leap-frog time stepping done in trazdf
   !! History of the T&A module :
   !!                 !  2009-01  (A. Vidard) tam of the 2008-04 version
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   tra_zdf_exp_tan  : compute the tracer the vertical diffusion trend using a
   !!                  split-explicit time stepping and provide the after tracer (tangent)
   !!   tra_zdf_exp_adj  : compute the tracer the vertical diffusion trend using a
   !!                  split-explicit time stepping and provide the after tracer (adjoint)
   !!----------------------------------------------------------------------
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 &
      & jpim1,               &
      & jpjm1,               &
      & jpkm1,               &
      & jpiglo
   USE oce_tam       , ONLY: & ! ocean dynamics and active tracers
      & tb_tl,               &
      & sb_tl,               &
      & ta_tl,               &
      & sa_tl,               &
      & tb_ad,               &
      & sb_ad,               &
      & ta_ad,               &
      & sa_ad
   USE dom_oce       , ONLY: & ! ocean space and time domain
      & e1t,                 &
      & e2t,                 &
# if defined key_vvl
      & e3t_1,               &
# else
#  if defined key_zco
      & e3t_0,               &
      & e3w_0,               &
#  else
      & e3t,                 &
      & e3w,                 &
#  endif
# endif
      & tmask,               &
      & lk_vvl,              &
      & mig,                 &
      & mjg,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej,                &
      & rdttra
   USE zdf_oce       , ONLY: & ! ocean vertical physics
      & avt,                 &
      & n_zdfexp
#if defined key_zdfddm
   USE zdfddm        , ONLY: &
      & avs
#endif
   USE in_out_manager, ONLY: & ! I/O manager 
      & lwp,                 &
      & numout,              & 
      & nitend,              & 
      & nit000
   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld    , ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE paresp        , ONLY: & ! Weights for an energy-type scalar product
      & wesp_t,              &
      & wesp_s
   USE tstool_tam    , ONLY: &
      & prntst_adj,          & !
      & stdt,                & ! stdev for s-velocity
      & stds                   !           t-velocity

#if defined key_obc
   USE obc_oce
#endif

   IMPLICIT NONE
   PRIVATE

   PUBLIC   tra_zdf_exp_tan       ! routine called by tra_zdf_tan.F90
   PUBLIC   tra_zdf_exp_adj       ! routine called by tra_zdf_adj.F90
   PUBLIC   tra_zdf_exp_adj_tst   ! routine called by tst.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "zdfddm_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !! NEMO/OPA  3.0 , LOCEAN-IPSL (2008) 
   !! $Id: trazdf_exp.F90 1146 2008-06-25 11:42:56Z rblod $ 
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE tra_zdf_exp_tan( kt, p2dt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_zdf_exp_tan  ***
      !!                   
      !! ** Purpose of the direct routine:
      !!      Compute the after tracer fields due to the vertical
      !!      tracer mixing alone, and then due to the whole tracer trend.
      !!
      !! ** Method of the direct routine :
      !!               - The after tracer fields due to the vertical diffusion
      !!      of tracers alone is given by:
      !!                zwx = tb + p2dt difft
      !!      where difft = dz( avt dz(tb) ) = 1/e3t dk+1( avt/e3w dk(tb) )
      !!           (if lk_zdfddm=T use avs on salinity instead of avt)
      !!      difft is evaluated with an Euler split-explit scheme using a
      !!      no flux boundary condition at both surface and bottomi boundaries.
      !!      (N.B. bottom condition is applied through the masked field avt).
      !!              - the after tracer fields due to the whole trend is 
      !!      obtained in leap-frog environment by :
      !!          ta = zwx + p2dt ta
      !!              - in case of variable level thickness (lk_vvl=T) the 
      !!     the leap-frog is applied on thickness weighted tracer. That is:
      !!          ta = [ tb*e3tb + e3tn*( zwx - tb + p2dt ta ) ] / e3tn
      !!
      !! ** Action : - after tracer fields (ta,sa) 
      !!---------------------------------------------------------------------
      INTEGER , INTENT(in)                 ::   kt     ! ocean time-step index
      REAL(wp), INTENT(in), DIMENSION(jpk) ::   p2dt   ! vertical profile of tracer time-step
      !!
      INTEGER  ::   ji, jj, jk, jl            ! dummy loop indices
      REAL(wp) ::   zlavmr, zave3r, ze3tr     ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zwxtl, zwytl, zwztl, zwwtl   ! 3D workspace
      !!---------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_zdf_exp_tan : explicit vertical mixing'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~~'
      ENDIF

      ! Initializations
      ! ---------------
      zlavmr = 1. / float( n_zdfexp )                           ! Local constant
      !
      zwytl(:,:, 1 ) = 0.e0        ;   zwwtl(:,:, 1 ) = 0.e0        ! surface boundary conditions: no flux
      zwytl(:,:,jpk) = 0.e0        ;   zwwtl(:,:,jpk) = 0.e0        ! bottom  boundary conditions: no flux
      !
      zwxtl(:,:,:)   = tb_tl(:,:,:)   ;   zwztl(:,:,:)   = sb_tl(:,:,:)   ! zwx and zwz arrays set to before tracer values

      ! Split-explicit loop  (after tracer due to the vertical diffusion alone)
      ! -------------------
      !
      DO jl = 1, n_zdfexp
         !                     ! first vertical derivative
         DO jk = 2, jpk
            DO jj = 2, jpjm1 
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zave3r = 1.e0 / fse3w(ji,jj,jk) 
                  zwytl(ji,jj,jk) =   avt(ji,jj,jk) * ( zwxtl(ji,jj,jk-1) - zwxtl(ji,jj,jk) ) * zave3r
                  zwwtl(ji,jj,jk) = fsavs(ji,jj,jk) * ( zwztl(ji,jj,jk-1) - zwztl(ji,jj,jk) ) * zave3r
               END DO
            END DO
         END DO
         !
         DO jk = 1, jpkm1      ! second vertical derivative   ==> tracer at kt+l*2*rdt/n_zdfexp
            DO jj = 2, jpjm1 
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  ze3tr = zlavmr / fse3t(ji,jj,jk)
                  zwxtl(ji,jj,jk) = zwxtl(ji,jj,jk) + p2dt(jk) * ( zwytl(ji,jj,jk) - zwytl(ji,jj,jk+1) ) * ze3tr
                  zwztl(ji,jj,jk) = zwztl(ji,jj,jk) + p2dt(jk) * ( zwwtl(ji,jj,jk) - zwwtl(ji,jj,jk+1) ) * ze3tr
               END DO
            END DO
         END DO
         !
      END DO

      ! After tracer due to all trends
      ! ------------------------------
      IF( lk_vvl ) THEN          ! variable level thickness : leap-frog on tracer*e3t
         IF(lwp) WRITE(numout,*) "key_vvl net available in tangent yet"
         CALL abort
      ELSE                       ! fixed level thickness : leap-frog on tracers
         DO jk = 1, jpkm1
            DO jj = 2, jpjm1 
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  ta_tl(ji,jj,jk) = ( zwxtl(ji,jj,jk) + p2dt(jk) * ta_tl(ji,jj,jk) ) *tmask(ji,jj,jk)
                  sa_tl(ji,jj,jk) = ( zwztl(ji,jj,jk) + p2dt(jk) * sa_tl(ji,jj,jk) ) *tmask(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF
      !
   END SUBROUTINE tra_zdf_exp_tan

   SUBROUTINE tra_zdf_exp_adj( kt, p2dt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_zdf_exp_adj  ***
      !!                   
      !! ** Purpose of the direct routine:
      !!      Compute the after tracer fields due to the vertical
      !!      tracer mixing alone, and then due to the whole tracer trend.
      !!
      !! ** Method of the direct routine :
      !!               - The after tracer fields due to the vertical diffusion
      !!      of tracers alone is given by:
      !!                zwx = tb + p2dt difft
      !!      where difft = dz( avt dz(tb) ) = 1/e3t dk+1( avt/e3w dk(tb) )
      !!           (if lk_zdfddm=T use avs on salinity instead of avt)
      !!      difft is evaluated with an Euler split-explit scheme using a
      !!      no flux boundary condition at both surface and bottomi boundaries.
      !!      (N.B. bottom condition is applied through the masked field avt).
      !!              - the after tracer fields due to the whole trend is 
      !!      obtained in leap-frog environment by :
      !!          ta = zwx + p2dt ta
      !!              - in case of variable level thickness (lk_vvl=T) the 
      !!     the leap-frog is applied on thickness weighted tracer. That is:
      !!          ta = [ tb*e3tb + e3tn*( zwx - tb + p2dt ta ) ] / e3tn
      !!
      !! ** Action : - after tracer fields (ta,sa) 
      !!---------------------------------------------------------------------
      INTEGER , INTENT(in)                 ::   kt     ! ocean time-step index
      REAL(wp), INTENT(in), DIMENSION(jpk) ::   p2dt   ! vertical profile of tracer time-step
      !!
      INTEGER  ::   ji, jj, jk, jl            ! dummy loop indices
      REAL(wp) ::   zlavmr, zave3r, ze3tr     ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zwxad, zwyad, zwzad, zwwad   ! 3D workspace
      !!---------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_zdf_exp_adj : explicit vertical mixing'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~~'
      ENDIF

      ! Initializations
      ! ---------------
      zlavmr = 1. / float( n_zdfexp )                           ! Local constant
      !
      zwwad(:,:,:) = 0.0_wp       ;   zwxad(:,:,:) = 0.0_wp 
      zwyad(:,:,:) = 0.0_wp       ;   zwzad(:,:,:) = 0.0_wp
      ! After tracer due to all trends
      ! ------------------------------
      IF( lk_vvl ) THEN          ! variable level thickness : leap-frog on tracer*e3t
         IF(lwp) WRITE(numout,*) "key_vvl net available in adjoint yet"
         CALL abort
      ELSE                       ! fixed level thickness : leap-frog on tracers
         DO jk = 1, jpkm1
            DO jj = 2, jpjm1 
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zwxad(ji,jj,jk) = zwxad(ji,jj,jk) + ta_ad(ji,jj,jk) * tmask(ji,jj,jk)
                  ta_ad(ji,jj,jk) = p2dt(jk) * ta_ad(ji,jj,jk) * tmask(ji,jj,jk)
                  zwzad(ji,jj,jk) = zwzad(ji,jj,jk) + sa_ad(ji,jj,jk) * tmask(ji,jj,jk)
                  sa_ad(ji,jj,jk) = p2dt(jk) * sa_ad(ji,jj,jk) * tmask(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF
      !

      ! Split-explicit loop  (after tracer due to the vertical diffusion alone)
      ! -------------------
      !
      DO jl = 1, n_zdfexp
         DO jk =  jpkm1, 1, -1      ! second vertical derivative   ==> tracer at kt+l*2*rdt/n_zdfexp
            DO jj = 2, jpjm1 
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  ze3tr = zlavmr / fse3t(ji,jj,jk)
                  zwyad(ji,jj,jk  ) = zwyad(ji,jj,jk  ) + p2dt(jk) * zwxad(ji,jj,jk) * ze3tr
                  zwyad(ji,jj,jk+1) = zwyad(ji,jj,jk+1) - p2dt(jk) * zwxad(ji,jj,jk) * ze3tr
                  zwwad(ji,jj,jk  ) = zwwad(ji,jj,jk  ) + p2dt(jk) * zwzad(ji,jj,jk) * ze3tr
                  zwwad(ji,jj,jk+1) = zwwad(ji,jj,jk+1) - p2dt(jk) * zwzad(ji,jj,jk) * ze3tr
               END DO
            END DO
         END DO
         !                     ! first vertical derivative
         DO jk = jpk, 2, -1
            DO jj = 2, jpjm1 
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zave3r = 1.e0 / fse3w(ji,jj,jk) 
                  zwxad(ji,jj,jk-1) = zwxad(ji,jj,jk-1) + avt(ji,jj,jk) * zwyad(ji,jj,jk) * zave3r
                  zwxad(ji,jj,jk  ) = zwxad(ji,jj,jk  ) - avt(ji,jj,jk) * zwyad(ji,jj,jk) * zave3r
                  zwyad(ji,jj,jk  ) = 0.0_wp
                  zwzad(ji,jj,jk-1) = zwzad(ji,jj,jk-1) + fsavs(ji,jj,jk) * zwwad(ji,jj,jk) * zave3r
                  zwzad(ji,jj,jk  ) = zwzad(ji,jj,jk  ) - fsavs(ji,jj,jk) * zwwad(ji,jj,jk) * zave3r
                  zwwad(ji,jj,jk  ) = 0.0_wp
               END DO
            END DO
         END DO
         !
         !
      END DO

      tb_ad(:,:,:) = tb_ad(:,:,:) + zwxad(:,:,:) 
      sb_ad(:,:,:) = sb_ad(:,:,:) + zwzad(:,:,:)

   END SUBROUTINE tra_zdf_exp_adj

   SUBROUTINE tra_zdf_exp_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE tra_zdf_exp_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !!                   
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      INTEGER ::  &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator
      REAL(KIND=wp) ::   &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp2            ! scalar product involving the adjoint routine
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & zta_tlin ,     & ! Tangent input
         & ztb_tlin ,     & ! Tangent input
#if defined key_obc
         & ztb_tlout, zsb_tlout, ztb_adin, zsb_adin,  & 
#endif
         & zsa_tlin ,     & ! Tangent input
         & zsb_tlin ,     & ! Tangent input
         & zta_tlout,     & ! Tangent output
         & zsa_tlout,     & ! Tangent output
         & zta_adin ,     & ! Adjoint input
         & zsa_adin ,     & ! Adjoint input
         & zta_adout,     & ! Adjoint output
         & ztb_adout,     & ! Adjoint output
         & zsa_adout,     & ! Adjoint output
         & zsb_adout,     & ! Adjoint output
         & zr             ! 3D random field 
      CHARACTER(LEN=14) :: cl_name
      ! Allocate memory

      ALLOCATE( &
         & zta_tlin( jpi,jpj,jpk),     &
         & zsa_tlin( jpi,jpj,jpk),     &
         & ztb_tlin( jpi,jpj,jpk),     &
         & zsb_tlin( jpi,jpj,jpk),     &
         & zta_tlout(jpi,jpj,jpk),     &
         & zsa_tlout(jpi,jpj,jpk),     &
         & zta_adin( jpi,jpj,jpk),     &
         & zsa_adin( jpi,jpj,jpk),     &
         & zta_adout(jpi,jpj,jpk),     &
         & zsa_adout(jpi,jpj,jpk),     &
         & ztb_adout(jpi,jpj,jpk),     &
         & zsb_adout(jpi,jpj,jpk),     &
         & zr(       jpi,jpj,jpk)      &
         & )

#if defined key_obc
      ALLOCATE( ztb_tlout(jpi,jpj,jpk),  zsb_tlout(jpi,jpj,jpk),     &
           &    ztb_adin (jpi,jpj,jpk),  zsb_adin (jpi,jpj,jpk) )
#endif

      !==================================================================
      ! 1) dx = ( un_tl, vn_tl, hdivn_tl ) and 
      !    dy = ( hdivb_tl, hdivn_tl )
      !==================================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
          zta_tlin( :,:,:) = 0.0_wp
          ztb_tlin( :,:,:) = 0.0_wp
          zsa_tlin( :,:,:) = 0.0_wp
          zsb_tlin( :,:,:) = 0.0_wp
          zta_tlout(:,:,:) = 0.0_wp
          zsa_tlout(:,:,:) = 0.0_wp
          zta_adin( :,:,:) = 0.0_wp
          zsa_adin( :,:,:) = 0.0_wp
          zta_adout(:,:,:) = 0.0_wp
          zsa_adout(:,:,:) = 0.0_wp
          ztb_adout(:,:,:) = 0.0_wp
          zsb_adout(:,:,:) = 0.0_wp
          zr(       :,:,:) = 0.0_wp
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stdt )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zta_tlin(ji,jj,jk) = zr(ji,jj,jk) 
           END DO
        END DO
     END DO

     DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 352791 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stdt )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              ztb_tlin(ji,jj,jk) = zr(ji,jj,jk) 
           END DO
        END DO
     END DO

     DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 142746 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stds )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zsa_tlin(ji,jj,jk) = zr(ji,jj,jk) 
           END DO
        END DO
     END DO
     DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 214934 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stds )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zsb_tlin(ji,jj,jk) = zr(ji,jj,jk) 
           END DO
        END DO
     END DO
      ta_tl(:,:,:) = zta_tlin(:,:,:)
      sa_tl(:,:,:) = zsa_tlin(:,:,:)
      tb_tl(:,:,:) = ztb_tlin(:,:,:)
      sb_tl(:,:,:) = zsb_tlin(:,:,:)
      CALL tra_zdf_exp_tan ( nit000, rdttra )
      zta_tlout(:,:,:) = ta_tl(:,:,:)
      zsa_tlout(:,:,:) = sa_tl(:,:,:)
#if defined key_obc
      ztb_tlout(:,:,:) = tb_tl(:,:,:)
      zsb_tlout(:,:,:) = sb_tl(:,:,:)     
#endif

      !--------------------------------------------------------------------
      ! Initialize the adjoint variables: dy^* = W dy
      !--------------------------------------------------------------------

      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zta_adin(ji,jj,jk) = zta_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_t(jk)
              zsa_adin(ji,jj,jk) = zsa_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_s(jk)
#if defined key_obc
              ztb_adin(ji,jj,jk) = ztb_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_t(jk)
              zsb_adin(ji,jj,jk) = zsb_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_s(jk)
            
#endif
            END DO
         END DO
      END DO
      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1 = DOT_PRODUCT( zta_tlout, zta_adin ) &
         & + DOT_PRODUCT( zsa_tlout, zsa_adin )

#if defined key_obc
      zsp1 = zsp1 + DOT_PRODUCT( ztb_tlout, ztb_adin ) &
           &      + DOT_PRODUCT( zsb_tlout, zsb_adin )
#endif

      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------

      ta_ad(:,:,:) = zta_adin(:,:,:)
      sa_ad(:,:,:) = zsa_adin(:,:,:)

#if defined key_obc
      tb_ad(:,:,:) = ztb_adin(:,:,:)
      sb_ad(:,:,:) = zsb_adin(:,:,:)
#endif
      
      CALL tra_zdf_exp_adj ( nit000, rdttra )
      
      zta_adout(:,:,:) = ta_ad(:,:,:)
      zsa_adout(:,:,:) = sa_ad(:,:,:)
      ztb_adout(:,:,:) = tb_ad(:,:,:)
      zsb_adout(:,:,:) = sb_ad(:,:,:)

      zsp2 = DOT_PRODUCT( zta_tlin, zta_adout ) &
         & + DOT_PRODUCT( zsa_tlin, zsa_adout ) &
         & + DOT_PRODUCT( ztb_tlin, ztb_adout ) &
         & + DOT_PRODUCT( zsb_tlin, zsb_adout ) 

      ! 14 char:'12345678901234'
      cl_name = 'trazdf_exp_adj'
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      DEALLOCATE(   &
         & zta_tlin,  &
         & ztb_tlin,  &
         & zsa_tlin,  &
         & zsb_tlin,  &
         & zta_tlout, &
         & zsa_tlout, &
         & zta_adin,  &
         & zsa_adin,  &
         & zta_adout, &
         & ztb_adout, &
         & zsa_adout, &
         & zsb_adout, &
         & zr       &
         & )



   END SUBROUTINE tra_zdf_exp_adj_tst

   !!==============================================================================
#endif
END MODULE trazdf_exp_tam
