MODULE zdfkpp_tam
#ifdef key_tam
   !!======================================================================
   !!                       ***  MODULE  zdfkpp_tam  ***
   !! Ocean physics:  vertical mixing coefficient compute from the KPP 
   !!                 turbulent closure parameterization
   !!                 Tangent and Adjoint Module
   !!=====================================================================
   !! History of the direct module:
   !!            8.1  ! 00-03 (W.G. Large, J. Chanut) Original code
   !!            8.1  ! 02-06 (J.M. Molines) for real case CLIPPER  
   !!            8.2  ! 03-10 (Chanut J.) re-writting
   !!            9.0  ! 05-01 (C. Ethe) Free form, F90
   !!                 ! 08-07 (M. Balmaseda). Add parameteres to namelist
   !! History of the TAM:
   !!                 ! 08-06 (A. Vidard) skeleton TAM of the 05-01 version
   !!                 ! 08-11 (A. Vidard) skeleton TAM of the 08-07 version
   !!----------------------------------------------------------------------
#if defined key_zdfkpp   ||   defined key_esopa
   !!----------------------------------------------------------------------
   !!   'key_zdfkpp'                                             KPP scheme
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   zdf_kpp      : update momentum and tracer Kz from a kpp scheme
   !!   zdf_kpp_init : initialization, namelist read, and parameters control
   !!----------------------------------------------------------------------
   USE oce             ! ocean dynamics and active tracers 
   USE dom_oce         ! ocean space and time domain
   USE zdf_oce         ! ocean vertical physics
   USE phycst          ! physical constants
   USE eosbn2          ! equation of state
   USE sbc_oce         ! thermohaline fluxes
   USE zdfddm          ! double diffusion mixing
   USE in_out_manager  ! I/O manager
   USE lbclnk          ! ocean lateral boundary conditions (or mpp link)
   USE prtctl          ! Print control

   IMPLICIT NONE
   PRIVATE

   PUBLIC   zdf_kpp_tan   ! routine called by steptan.F90
   PUBLIC   tra_kpp_tan   ! routine called by steptan.F90
   PUBLIC   zdf_kpp_adj   ! routine called by stepadj.F90
   PUBLIC   tra_kpp_adj   ! routine called by stepadj.F90

   LOGICAL, PUBLIC, PARAMETER ::   lk_zdfkpp = .TRUE.    !: KPP vertical mixing flag
   REAL(wp), PUBLIC, DIMENSION(jpi,jpj,jpk) ::   ghats   !: non-local scalar mixing term (gamma/<ws>o)
   REAL(wp), PUBLIC, DIMENSION(jpi,jpj) ::   & 
      wt0                   ,  &  !: surface temperature flux for non local flux
      ws0                   ,  &  !: surface salinity flux for non local flux
      hkpp                        !: boundary layer depht

   INTEGER ::                  & !!! ** kpp namelist (namkpp) **
      nave  =  1            ,  &  ! = 0/1 flag for horizontal average on avt, avmu, avmv
      navb  =  0                  ! = 0/1 flag for constant or profile background avt

   REAL(wp) ::                 & !!! ** Interior Mixing 
      difmiw  = 1.2e-04_wp  ,  &  ! constant internal wave viscosity (m2/s)
      difsiw  = 1.2e-05_wp  ,  &  ! constant internal wave diffusivity (m2/s)
      Riinfty = 0.8_wp      ,  &  ! local Richardson Number limit for shear instability
      difri   = 5.e-03_wp   ,  &  ! maximum shear mixing at Rig = 0    (m2/s)
      bvsqcon = -1.e-09_wp  ,  &  ! Brunt-Vaisala squared (1/s**2) for maximum convection
      difcon  = 1._wp             ! maximum mixing in interior convection (m2/s)
#if defined key_zdfddm
   REAL(wp) ::                 & !!! ** Double diffusion Mixing
      difssf  = 1.e-03_wp   ,  &  ! maximum salt fingering mixing  (nmlst)
      Rrho0   = 1.9_wp      ,  &  ! limit for salt  fingering mixing 
      difsdc  = 1.5e-06_wp  ,  &  ! maximum diffusive convection mixing (nmlst)
      rdifst  = 0.7_wp            ! ration salt/Temp diff if salt fingering 
#endif
   LOGICAL  ::                 &
      ln_kpprimix  = .TRUE.       ! Shear instability mixing 

   REAL(wp) ::                 & !!! ** General constants  **
      epsln   = 1.0e-20_wp   , &  ! a small positive number    
      pthird  = 1._wp/3._wp  , &  ! 1/3
      pfourth = 1._wp/4._wp       ! 1/4

   REAL(wp) ::                 & !!! ** Boundary Layer Turbulence Parameters  **
      vonk     = 0.4_wp     ,  &  ! von Karman's constant
      epsilon  = 0.1_wp     ,  &  ! nondimensional extent of the surface layer
      rconc1   = 5.0_wp     ,  &  ! standard flux profile function parmaeters
      rconc2   = 16.0_wp    ,  &  !         "        "
      rconcm   = 8.38_wp    ,  &  ! momentum flux profile fit
      rconam   = 1.26_wp    ,  &  !         "       "
      rzetam   = -.20_wp    ,  &  !         "       "       
      rconcs   = 98.96_wp   ,  &  !  scalar  flux profile fit
      rconas   = -28.86_wp  ,  &  !         "       "
      rzetas   = -1.0_wp          !         "       "  
   REAL(wp) ::                 & !!! ** Boundary Layer Depth Diagnostic  **
      Ricr     = 0.3_wp     ,  &  ! critical bulk Richardson Number
      rcekman  = 0.7_wp     ,  &  ! coefficient for ekman depth  
      rcmonob  = 1.0_wp     ,  &  ! coefficient for Monin-Obukhov depth 
      rconcv   = 1.7_wp     ,  &  ! ratio of interior buoyancy frequency to buoyancy frequency at entrainment depth (namelist)
      hbf      = 1.0_wp     ,  &  ! fraction of bound. layer depth to which absorbed solar 
      !                           ! rad. and contributes to surf. buo. forcing
      Vtc                         ! function of rconcv,rconcs,epsilon,vonk,Ricr
   REAL(wp) ::                 & !!! ** Nonlocal Boundary Layer Mixing **
      rcstar   = 5.0_wp     ,  &  ! coefficient for convective nonlocal transport
      rcs      = 1.0e-3_wp  ,  &  ! conversion: mm/s ==> m/s   
      rcg                         ! non-dimensional coefficient for nonlocal transport

#if ! defined key_kppcustom
   REAL(wp), DIMENSION(jpk,jpk) ::   & 
      del                         ! array for reference mean values of vertical integration 
#endif

#if defined key_kpplktb
   INTEGER, PARAMETER ::       & !!! ** Parameters for lookup table for turbulent velocity scales ** 
      nilktb   = 892        ,  &  ! number of values for zehat in KPP lookup table
      njlktb   = 482        ,  &  ! number of values for ustar in KPP lookup table
      nilktbm1 = nilktb - 1 ,  &  !
      njlktbm1 = njlktb - 1       !

   REAL(wp), DIMENSION(nilktb,njlktb) ::  &
      wmlktb                ,  &  ! lookup table for the turbulent vertical velocity scale for momentum
      wslktb                       ! lookup table for the turbulent vertical velocity scale for tracers

   REAL(wp) ::                 &
      dehatmin = -4.e-7_wp  ,  &  ! minimum limit for zhat in lookup table (m3/s3) 
      dehatmax = 0._wp      ,  &  ! maximum limit for zhat in lookup table (m3/s3)
      ustmin   = 0._wp      ,  &  ! minimum limit for ustar in lookup table (m/s)
      ustmax   = 0.04_wp    ,  &  ! maximum limit for ustar in lookup table (m/s)    
      dezehat               ,  &  ! delta zhat in lookup table
      deustar                     ! delta ustar in lookup table
#endif
   REAL(wp), DIMENSION(jpk) :: &  !!! attenuation coef  
      ratt        
   !! already defines in module traqsr, but only if the solar radiation penetration is considered
   REAL(wp) ::                 & !!! * penetrative solar radiation coefficient *
      rabs = 0.58_wp        ,  &  ! fraction associated with xsi1
      xsi1 = 0.35_wp        ,  &  ! first depth of extinction 
      xsi2 = 23.0_wp              ! second depth of extinction 
      !                           ! (default values: water type Ib) 

   REAL(wp), DIMENSION(jpi,jpj,jpk) ::   &
      etmean                ,  &  ! coefficient used for horizontal smoothing
      eumean                ,  &  ! at t-, u- and v-points
      evmean  
 
#if defined key_c1d
   REAL(wp), PUBLIC, DIMENSION(jpi,jpj,jpk) ::   &
      rig                   ,  &  ! gradient Richardson number
      rib                   ,  &  ! bulk Richardson number
      buof                  ,  &  ! buoyancy forcing
      mols                        ! moning-Obukhov length scale 
   REAL(wp), PUBLIC, DIMENSION(jpi,jpj) ::   &
      ekdp                        ! Ekman depth
#endif

   INTEGER  ::  &                 !
      jip = 62 , jjp = 111

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"

CONTAINS

   SUBROUTINE zdf_kpp_tan ( kt )
      !!----------------------------------------------------------------------
      !!                   ***  ROUTINE zdf_kpp_tan  ***
      !!
      !! ** Purpose of thedirect routine:
      !!      Compute the vertical eddy viscosity and diffusivity
      !!      coefficients and non local mixing using K-profile parameterization
      !!
      !! ** Method :   The boundary layer depth hkpp is diagnosed at tracer points
      !!      from profiles of buoyancy, and shear, and the surface forcing.
      !!      Above hbl (sigma=-z/hbl <1) the mixing coefficients are computed from 
      !!
      !!                      Kx =  hkpp  Wx(sigma) G(sigma)  
      !!
      !!             and the non local term ghat = Cs / Ws(sigma) / hkpp
      !!      Below hkpp  the coefficients are the sum of mixing due to internal waves
      !!      shear instability and double diffusion.
      !!
      !!      -1- Compute the now interior vertical mixing coefficients at all depths. 
      !!      -2- Diagnose the boundary layer depth. 
      !!      -3- Compute the now boundary layer vertical mixing coefficients. 
      !!      -4- Compute the now vertical eddy vicosity and diffusivity.
      !!      -5- Smoothing
      !!
      !!        N.B. The computation is done from jk=2 to jpkm1 
      !!             Surface value of avt avmu avmv are set once a time to zero
      !!             in routine zdf_kpp_init.
      !!
      !! ** Action  :   update the non-local terms ghats
      !!                update avt, avmu, avmv (before vertical eddy coef.)
      !!
      !! References : Large W.G., Mc Williams J.C. and Doney S.C.              
      !!         Reviews of Geophysics, 32, 4, November 1994
      !!         Comments in the code refer to this paper, particularly 
      !!         the equation number. (LMD94, here after)
      !!----------------------------------------------------------------------
#if defined  key_zdfddm
      USE oce     , zviscos => ua,      &  ! temp. array for viscosities use ua as workspace
         &          zdiffut => ta,      &  ! temp. array for diffusivities use sa as workspace
         &          zdiffus => sa          ! temp. array for diffusivities use sa as workspace
#else
      USE oce     , zviscos => ua,      &  ! temp. array for viscosities use ua as workspace
         &          zdiffut => ta          ! temp. array for diffusivities use sa as workspace
#endif
      !!
      INTEGER, INTENT( in  ) ::   kt   ! ocean time step
      !!

   END SUBROUTINE zdf_kpp_tan


   SUBROUTINE tra_kpp_tan
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_kpp_tan  ***
      !!
      !! ** Purpose of the direct routine:   
      !!      compute and add to the tracer trend the non-local
      !!      tracer flux
      !!
      !! ** Method  :   ???
      !!
      !! history or the direct routine:
      !!     9.0  ! 05-11 (G. Madec)  Original code
      !! history:
      !!          ! 08-06 (A. Vidard) 
      !!----------------------------------------------------------------------
      !! * Modules used
      USE oce, ONLY :    ztrdt => ua,       & ! use ua as 3D workspace
                         ztrds => va          ! use va as 3D workspace
      !!----------------------------------------------------------------------

 
   END SUBROUTINE tra_kpp_tan
   SUBROUTINE zdf_kpp_adj ( kt )
      !!----------------------------------------------------------------------
      !!                   ***  ROUTINE zdf_kpp_adj  ***
      !!
      !! ** Purpose of thedirect routine:
      !!      Compute the vertical eddy viscosity and diffusivity
      !!      coefficients and non local mixing using K-profile parameterization
      !!
      !! ** Method :   The boundary layer depth hkpp is diagnosed at tracer points
      !!      from profiles of buoyancy, and shear, and the surface forcing.
      !!      Above hbl (sigma=-z/hbl <1) the mixing coefficients are computed from 
      !!
      !!                      Kx =  hkpp  Wx(sigma) G(sigma)  
      !!
      !!             and the non local term ghat = Cs / Ws(sigma) / hkpp
      !!      Below hkpp  the coefficients are the sum of mixing due to internal waves
      !!      shear instability and double diffusion.
      !!
      !!      -1- Compute the now interior vertical mixing coefficients at all depths. 
      !!      -2- Diagnose the boundary layer depth. 
      !!      -3- Compute the now boundary layer vertical mixing coefficients. 
      !!      -4- Compute the now vertical eddy vicosity and diffusivity.
      !!      -5- Smoothing
      !!
      !!        N.B. The computation is done from jk=2 to jpkm1 
      !!             Surface value of avt avmu avmv are set once a time to zero
      !!             in routine zdf_kpp_init.
      !!
      !! ** Action  :   update the non-local terms ghats
      !!                update avt, avmu, avmv (before vertical eddy coef.)
      !!
      !! References : Large W.G., Mc Williams J.C. and Doney S.C.              
      !!         Reviews of Geophysics, 32, 4, November 1994
      !!         Comments in the code refer to this paper, particularly 
      !!         the equation number. (LMD94, here after)
      !!----------------------------------------------------------------------
#if defined  key_zdfddm
      USE oce     , zviscos => ua,      &  ! temp. array for viscosities use ua as workspace
         &          zdiffut => ta,      &  ! temp. array for diffusivities use sa as workspace
         &          zdiffus => sa          ! temp. array for diffusivities use sa as workspace
#else
      USE oce     , zviscos => ua,      &  ! temp. array for viscosities use ua as workspace
         &          zdiffut => ta          ! temp. array for diffusivities use sa as workspace
#endif
      !!
      INTEGER, INTENT( in  ) ::   kt   ! ocean time step
      !!

   END SUBROUTINE zdf_kpp_adj


   SUBROUTINE tra_kpp_adj
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_kpp_adj  ***
      !!
      !! ** Purpose :   compute and add to the tracer trend the non-local
      !!      tracer flux
      !!
      !! ** Method  :   ???
      !!
      !! history :
      !!     9.0  ! 05-11 (G. Madec)  Original code
      !!----------------------------------------------------------------------
      !! * Modules used
      USE oce, ONLY :    ztrdt => ua,       & ! use ua as 3D workspace
                         ztrds => va          ! use va as 3D workspace
      !!----------------------------------------------------------------------

 
   END SUBROUTINE tra_kpp_adj


 
#else
   !!----------------------------------------------------------------------
   !!   Dummy module :                                        NO KPP scheme
   !!----------------------------------------------------------------------
   LOGICAL, PUBLIC, PARAMETER ::   lk_zdfkpp = .FALSE.   !: KPP flag
CONTAINS
   SUBROUTINE zdf_kpp_tan( kt )          ! Empty routine
      INTEGER, INTENT( in  ) ::   kt   ! ocean time step
      WRITE(*,*) 'zdf_kpp_tan: You should not have seen this print! error?', kt
   END SUBROUTINE zdf_kpp_tan
   SUBROUTINE tra_kpp_tan( kt )          ! Empty routine
      INTEGER, INTENT( in  ) ::   kt   ! ocean time step
      WRITE(*,*) 'tra_kpp_tan: You should not have seen this print! error?', kt
   END SUBROUTINE tra_kpp_tan
   SUBROUTINE zdf_kpp_adj( kt )          ! Empty routine
      INTEGER, INTENT( in  ) ::   kt   ! ocean time step
      WRITE(*,*) 'zdf_kpp_adj: You should not have seen this print! error?', kt
   END SUBROUTINE zdf_kpp_adj
   SUBROUTINE tra_kpp_adj( kt )          ! Empty routine
      INTEGER, INTENT( in  ) ::   kt   ! ocean time step
      WRITE(*,*) 'tra_kpp_adj: You should not have seen this print! error?', kt
   END SUBROUTINE tra_kpp_adj
#endif

   !!======================================================================
#endif
END MODULE zdfkpp_tam

