MODULE cla_div_tam
#ifdef key_tam
   !!==============================================================================
   !!                    ***  MODULE  cla_div_tam  ***
   !! Ocean diagnostic variable : specific update of the horizontal divergence
   !!                             CAUTION: Specific to ORCA_R2
   !!                             Tangent and adjoint module
   !!==============================================================================
#  if defined key_orca_r2
   !!----------------------------------------------------------------------
   !!   'key_orca_r2'                                 global ocean model R2
   !!----------------------------------------------------------------------
   !!   div_cla      : 
   !!   div_bab_el_mandeb
   !!   div_gibraltar
   !!   div_hormuz
   !!   div_cla_init :
   !!----------------------------------------------------------------------
   !! * Modules used
   USE par_kind,       ONLY: & ! Precision variables
      & wp
   USE par_oce,        ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 & 
      & jpiglo
   USE oce_tam,        ONLY: & ! ocean dynamics and tracers
      & hdivn_tl,            &
      & hdivn_ad
   USE sbc_oce_tam,    ONLY: & ! surface variables
      & emp_tl,              &
      & emp_ad
   USE dom_oce,        ONLY: & ! ocean space and time domain
      & mi0,                 &
      & mi1,                 &
      & mj0,                 &
      & mj1,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej,                &
      & rdt,                 &
      & tmask,               &
      & mig,                 &
      & mjg,                 &
      & e2u,                 &
      & e1v,                 &
      & e1t,                 &
      & e2t,                 &
# if defined key_vvl
      & e3t_1,               &
# else
#  if defined key_zco
      & e3t_0,               &
#  else
      & e3t,                 &
#  endif
# endif
# if defined key_zco
      & e3t_0,               &
# else
      & e3u,                 &
      & e3v
# endif
   USE in_out_manager, ONLY: & ! I/O manager 
      & nit000
   USE phycst,         ONLY: & ! Physical constants
      & rauw
   USE lib_mpp,        ONLY: & ! distributed memory computing library
      & lk_mpp,              &
      & mpp_sum
   USE lbclnk,         ONLY: & ! ocean lateral boundary conditions (or mpp link)
      & lbc_lnk
   USE lbclnk_tam,     ONLY: & ! ocean lateral boundary conditions (or mpp link)
      & lbc_lnk_adj
   USE dotprodfld,     ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE gridrandom,     ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE tstool_tam,     ONLY: &
      & prntst_adj,          & !
      & stdemp,              & ! stdev for evaporation minus precip
      & stdh                   !           hdiv

   IMPLICIT NONE
   PRIVATE

   !! * Module variables
   REAL(wp) :: zempmed_tl, zempred_tl       ! EMP of Mediterranean and Red Sea 
   REAL(wp) :: zempmed_ad, zempred_ad       ! EMP of Mediterranean and Red Sea 
   !! * Routine accessibility
   PUBLIC div_cla_tan     ! routine called by step_tam.F90
   PUBLIC div_cla_adj     ! routine called by step_tam.F90
   PUBLIC div_cla_adj_tst ! routine called by tst.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"

CONTAINS

   SUBROUTINE div_cla_tan ( kt )
      !!----------------------------------------------------------------------
      !!                 ***  ROUTINE div_cla_tan  ***
      !!
      !! ** Purpose of the direct routine:
      !!      update the horizontal divergence of the velocity field
      !!      for at some straits ( Gibraltar, Bab el Mandeb and Hormuz ).
      !!
      !! ** Method  :   With imposed transport at each strait, we compute 
      !!      corresponding velocities and update horizontal divergence.
      !!        Apply lateral boundary conditions on hdivn through a call
      !!      to routine lbc_lnk.
      !!
      !! ** Action  :   update hdivn array : the now horizontal divergence
      !!
      !! History of the direct routine:
      !!   8.5  !  02-11 (A. Bozec)  Free form, F90
      !! History of the TAM routine:
      !!   9.0  !  08-06 (A. Vidard) Skeleton
      !!   9.0  !  08-09 (A. Vidard) tangent of the 02-11 version
      !!   9.0  !  09-02 (A. Vidard) cosmetic changes
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt      ! ocean time-step index
      !!----------------------------------------------------------------------
      ! Correction of the Divergence at some straits
            
      CALL div_bab_el_mandeb_tan                    ! New divergence at Bab el Mandeb 

      CALL div_gibraltar_tan                        ! New divergence at Gibraltar

      CALL div_hormuz_tan                           ! Hormuz Strait ( persian Gulf)

      ! Lateral boundary conditions on hdivn
      CALL lbc_lnk( hdivn_tl, 'T', 1. )


   END SUBROUTINE div_cla_tan

   SUBROUTINE div_cla_adj ( kt )
      !!----------------------------------------------------------------------
      !!                 ***  ROUTINE div_cla_adj  ***
      !!
      !! ** Purpose of the direct routine:
      !!      update the horizontal divergence of the velocity field
      !!      for at some straits ( Gibraltar, Bab el Mandeb and Hormuz ).
      !!
      !! ** Method  :   With imposed transport at each strait, we compute 
      !!      corresponding velocities and update horizontal divergence.
      !!        Apply lateral boundary conditions on hdivn through a call
      !!      to routine lbc_lnk.
      !!
      !! ** Action  :   update hdivn array : the now horizontal divergence
      !!
      !! History of the direct routine:
      !!   8.5  !  02-11 (A. Bozec)  Free form, F90
      !! History of the TAM routine:
      !!   9.0  !  08-06 (A. Vidard) Skeleton
      !!   9.0  !  08-09 (A. Vidard) adjoint of the 02-11 version
      !!   9.0  !  09-02 (A. Vidard) cosmetic changes
     !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt      ! ocean time-step index
      !!----------------------------------------------------------------------
      ! Correction of the Divergence at some straits

      ! Lateral boundary conditions on hdivn
      CALL lbc_lnk_adj( hdivn_ad, 'T', 1. )

      CALL div_hormuz_adj                           ! Hormuz Strait ( persian Gulf)

      CALL div_gibraltar_adj                        ! New divergence at Gibraltar

      CALL div_bab_el_mandeb_adj                    ! New divergence at Bab el Mandeb 

   END SUBROUTINE div_cla_adj

   SUBROUTINE div_bab_el_mandeb_tan
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE div_bab_el_mandeb_tan  ***
      !!       
      !! ** Purpose of the direct routine:
      !!     Update  the now horizontal divergence of the velocity 
      !!     field in Bab el Mandeb ( Red Sea strait ).
      !!
      !! ** Method of the direct routine:
      !!     Set the velocity field at each side of the strait :
      !!                                          |
      !!            |/ \|            N          |\ /|
      !!            |_|_|______      |          |___|______
      !!        88  |   |<-       W - - E    88 |   |<-
      !!        87  |___|______      |       87 |___|->____
      !!             160 161         S           160 161 
      !!       horizontal view                horizontal view
      !!          surface                        depth
      !!      The now divergence is given by :
      !!         hdivn = 1/(e1t*e2t) [ di(e2u  un) + dj(e1v  vn) ]
      !!
      !! ** History of the direct routine:
      !!           !         (A. Bozec) Original code 
      !!      8.5  !  02-11  (A. Bozec) F90: Free form and module
      !! ** History of the tangent routine:
      !!           !  08-09  (A. Vidard) Tangent linear of the 02-11 version
      !!      9.0  !  09-02 (A. Vidard) cosmetic changes
      !!----------------------------------------------------------------------
      !! * Local declarations
      INTEGER  :: ji, jj, jk   ! dummy loop indices
      REAL(wp) :: zsu, zvt, zwei   ! temporary scalar
      REAL(wp), DIMENSION (jpk) ::  zu1_rs_tl, zu2_rs_tl, zu3_rs_tl
      !!---------------------------------------------------------------------
      
      ! EMP on the Red Sea 
      ! ------------------

      zempred_tl = 0.e0
      zwei = 0.e0
      DO jj = mj0(87), mj1(96)
         DO ji = mi0(148), mi1(160) 
            zwei    = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            zempred_tl = zempred_tl + emp_tl(ji,jj) * zwei
         END DO
      END DO
      IF( lk_mpp )   CALL mpp_sum( zempred_tl )      ! sum with other processors value


      ! convert in m3
      zempred_tl = zempred_tl * 1.e-3         

      ! Velocity profile at each point
      ! ------------------------------

      zu1_rs_tl(:) = 0.0_wp
      zu2_rs_tl(:) = 0.0_wp
      zu3_rs_tl(:) = 0.0_wp

      ! velocity profile at 161,88 North point 
      ! we imposed zisw_rs + EMP above the Red Sea 
      DO jk = 1,  8                                      
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160) 
               zu1_rs_tl(jk) = zu1_rs_tl(jk) - ( zempred_tl / 8. ) / ( e2u(ji, jj) * fse3u(ji, jj,jk) )
            END DO
         END DO
      END DO

      ! velocity profile at 160,88 North  point
      ! we imposed zisw_rs + EMP above the Red Sea 
      DO jk = 1,  10                                     
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160) 
               zu3_rs_tl(jk) = zu3_rs_tl(jk) + ( zempred_tl / 10. ) / ( e1v(ji, jj) * fse3v(ji, jj,jk) )
            END DO
         END DO
      END DO
       
      ! Divergence at each point of the straits
      ! ---------------------------------------

      ! compute the new divergence at 161,88 
      DO jk = 1, 21
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(161), mi1(161) 
               zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
               zsu = e2u(ji-1, jj) * fse3u(ji-1, jj,jk)
               hdivn_tl(ji, jj  ,jk) = hdivn_tl(ji, jj  ,jk) - ( 1. / zvt ) * zsu * zu1_rs_tl(jk)
            END DO
         END DO
      END DO

      ! compute the new divergence at 161,87 
      DO jk = 1, 21
         DO jj = mj0(87), mj1(87) 
            DO ji = mi0(161), mi1(161) 
               zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
               zsu = e2u(ji-1, jj) * fse3u(ji-1, jj,jk)
               hdivn_tl(ji, jj,jk) = hdivn_tl(ji, jj,jk) - ( 1. / zvt ) * zsu * zu2_rs_tl(jk)
            END DO
         END DO
      END DO

      ! compute the divergence at 160,89 
      DO jk = 1, 18
         DO jj = mj0(89), mj1(89) 
            DO ji = mi0(160), mi1(160) 
               zvt = e1t(ji, jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu = e1v(ji, jj-1) * fse3v(ji, jj-1,jk)
               hdivn_tl(ji, jj,jk) = hdivn_tl(ji, jj,jk) - ( 1. / zvt ) * zsu * zu3_rs_tl(jk)
            END DO
         END DO
      END DO

   END SUBROUTINE div_bab_el_mandeb_tan


   SUBROUTINE div_bab_el_mandeb_adj
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE div_bab_el_mandeb_adj  ***
      !!       
      !! ** Purpose of the direct routine:
      !!     Update  the now horizontal divergence of the velocity 
      !!     field in Bab el Mandeb ( Red Sea strait ).
      !!
      !! ** Method of the direct routine:
      !!     Set the velocity field at each side of the strait :
      !!                                          |
      !!            |/ \|            N          |\ /|
      !!            |_|_|______      |          |___|______
      !!        88  |   |<-       W - - E    88 |   |<-
      !!        87  |___|______      |       87 |___|->____
      !!             160 161         S           160 161 
      !!       horizontal view                horizontal view
      !!          surface                        depth
      !!      The now divergence is given by :
      !!         hdivn = 1/(e1t*e2t) [ di(e2u  un) + dj(e1v  vn) ]
      !!
      !! ** History of the direct routine:
      !!           !         (A. Bozec) Original code 
      !!      8.5  !  02-11  (A. Bozec) F90: Free form and module
      !! ** History of the adjoint routine:
      !!           !  08-09  (A. Vidard) Adjoint of the 02-11 version
      !!      9.0  !  09-02 (A. Vidard) cosmetic changes
      !!----------------------------------------------------------------------
      !! * Local declarations
      INTEGER  :: ji, jj, jk   ! dummy loop indices
      REAL(wp) :: zsu, zvt, zwei   ! temporary scalar
      REAL(wp), DIMENSION (jpk) ::  zu1_rs_ad, zu2_rs_ad, zu3_rs_ad
      !!---------------------------------------------------------------------
      
      zempred_ad = 0.e0
      zu1_rs_ad(:) = 0.0_wp
      zu2_rs_ad(:) = 0.0_wp
      zu3_rs_ad(:) = 0.0_wp
      zwei = 0.e0
       
      ! Divergence at each point of the straits
      ! ---------------------------------------

      ! compute the divergence at 160,89 
      DO jk = 18, 1, -1
         DO jj = mj0(89), mj1(89) 
            DO ji = mi0(160), mi1(160) 
               zvt = e1t(ji, jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu = e1v(ji, jj-1) * fse3v(ji, jj-1,jk)
               zu3_rs_ad(jk) = zu3_rs_ad(jk) - hdivn_ad(ji, jj,jk) * ( 1. / zvt ) * zsu 
            END DO
         END DO
      END DO

      ! compute the new divergence at 161,87 
      DO jk = 21, 1, -1
         DO jj = mj0(87), mj1(87) 
            DO ji = mi0(161), mi1(161) 
               zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
               zsu = e2u(ji-1, jj) * fse3u(ji-1, jj,jk)
               zu2_rs_ad(jk) = zu2_rs_ad(jk) - hdivn_ad(ji, jj,jk) * ( 1. / zvt ) * zsu
            END DO
         END DO
      END DO

      ! compute the new divergence at 161,88 
      DO jk = 21, 1, -1
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(161), mi1(161) 
               zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
               zsu = e2u(ji-1, jj) * fse3u(ji-1, jj,jk)
               zu1_rs_ad(jk) = zu1_rs_ad(jk) - hdivn_ad(ji, jj  ,jk) * ( 1. / zvt ) * zsu
            END DO
         END DO
      END DO

      ! Velocity profile at each point
      ! ------------------------------

      ! velocity profile at 160,88 North  point
      ! we imposed zisw_rs + EMP above the Red Sea 
      DO jk = 10, 1, -1                                     
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160)
               zempred_ad = zempred_ad + ( zu3_rs_ad(jk) / 10. ) / ( e1v(ji, jj) * fse3v(ji, jj,jk) )
            END DO
         END DO
      END DO

      ! velocity profile at 161,88 North point 
      ! we imposed zisw_rs + EMP above the Red Sea 
      DO jk = 8, 1, -1     
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160) 
               zempred_ad = zempred_ad - ( zu1_rs_ad(jk) / 8. ) / ( e2u(ji, jj) * fse3u(ji, jj,jk) )
            END DO
         END DO
      END DO

      zu1_rs_ad(:) = 0.0_wp
      zu2_rs_ad(:) = 0.0_wp
      zu3_rs_ad(:) = 0.0_wp

      ! EMP on the Red Sea 
      ! ------------------

      ! convert in m3
      zempred_ad = zempred_ad * 1.e-3         

      IF( lk_mpp )   CALL mpp_sum( zempred_ad )      ! sum with other processors value

      DO jj = mj1(96), mj0(87), -1
         DO ji = mi1(160),  mi0(148), -1
            zwei    = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            emp_ad(ji,jj) = emp_ad(ji,jj) + zempred_ad * zwei
         END DO
      END DO
      zempred_ad = 0.e0

   END SUBROUTINE div_bab_el_mandeb_adj

   SUBROUTINE div_gibraltar_tan
      !! -------------------------------------------------------------------
      !!                 ***  ROUTINE div_gibraltar_tan  ***
      !!        
      !! ** Purpose of the direct routine:
      !!     update the now horizontal divergence of the velocity 
      !!     field in Gibraltar.
      !!
      !! ** Method of the direct routine:
      !!          ________________      N        ________________
      !! 102           |    |->         |           <-|    |<-
      !! 101      ___->|____|_____   W - - E     ___->|____|_____
      !!           139   140  141       |         139   140  141
      !!          horizontal view       S        horizontal view
      !!            surface                          depth
      !!      The now divergence is given by :
      !!         hdivn = 1/(e1t*e2t) [ di(e2u  un) + dj(e1v  vn) ]
      !!
      !! ** History of the direct routine:
      !!           !         (A. Bozec) Original code 
      !!      8.5  !  02-10  (A. Bozec) F90: Free form and module
      !! ** History of the tangent routine:
      !!           !  08-09  (A. Vidard) Tangent linear of the 02-10 version
      !!      9.0  !  09-02 (A. Vidard) cosmetic changes
      !!---------------------------------------------------------------------
      !! * Local declarations
      INTEGER  :: ji, jj, jk   ! dummy loop indices
      REAL(wp) :: zsu, zvt
      REAL(wp) :: zwei
      REAL(wp), DIMENSION (jpk) ::  zu1_ms_tl, zu2_ms_tl, zu3_ms_tl
      !!---------------------------------------------------------------------
      
      ! EMP on the Mediterranean Sea 
      ! ----------------------------

      zempmed_tl = 0.e0
      zwei = 0.e0
      DO jj = mj0(96), mj1(110)
         DO ji = mi0(141),mi1(181)
            zwei    = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            zempmed_tl = zempmed_tl + emp_tl(ji,jj) * zwei
         END DO
      END DO
      IF( lk_mpp )   CALL mpp_sum( zempmed_tl )      ! sum with other processors value

      ! minus 2 points in Red Sea and 3 in Atlantic 
      DO jj = mj0(96), mj1(96)
         DO ji = mi0(148),mi1(148)
            zempmed_tl = zempmed_tl -  emp_tl(ji  , jj) * tmask(ji  , jj,1) * e1t(ji  , jj) * e2t(ji  , jj)   &
                              -  emp_tl(ji+1, jj) * tmask(ji+1, jj,1) * e1t(ji+1, jj) * e2t(ji+1, jj)   
         END DO
      END DO

      ! convert in m3
      zempmed_tl = zempmed_tl * 1.e-3

      ! Velocity profile at each point
      ! ------------------------------

      zu1_ms_tl(:) = 0.0_wp
      zu2_ms_tl(:) = 0.0_wp
      zu3_ms_tl(:) = 0.0_wp

      ! velocity profile at 139,101 South point 
      ! we imposed zisw + EMP above the Mediterranean Sea 
      DO jk = 1, 14                      
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zu1_ms_tl(jk) =  zu1_ms_tl(jk) + ( zempmed_tl / 14. ) / ( e2u(ji-1, jj-1) * fse3u(ji-1, jj-1,jk) ) 
            END DO
         END DO
      END DO
      
      ! velocity profile at 141,102  East point
      ! flux in surface inflow of the Atlantic ocean + EMP   
      DO  jk = 1, 14                     
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zu3_ms_tl(jk) = zu3_ms_tl(jk) +  ( zempmed_tl / 14. ) / ( e2u(ji, jj) * fse3u(ji, jj,jk) ) 
            END DO
         END DO
      END DO
     
      ! Divergence at each point of the straits
      ! ---------------------------------------

      ! compute the new divergence at 139,101 South point  
      DO jk = 1, jpk
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
               zsu = e2u(ji, jj) * fse3u(ji, jj,jk)
               hdivn_tl(ji, jj,jk) = hdivn_tl(ji, jj,jk) + ( 1. / zvt ) * zsu * zu1_ms_tl(jk) 
            END DO
         END DO
      END DO

      ! compute the new divergence at 139,102 deep North point
      DO jk = 1, jpk
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(139), mi1(139) 
               zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
               zsu = e2u(ji, jj) * fse3u(ji, jj,jk)
               hdivn_tl(ji, jj,jk) = hdivn_tl(ji, jj,jk) + ( 1. / zvt ) * zsu * zu2_ms_tl(jk) 
            END DO
         END DO
      END DO

      ! compute the divergence at 141,102 East point
      DO jk = 1, jpk
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(141), mi1(141) 
               zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
               zsu = e2u(ji-1, jj) * fse3u(ji-1, jj,jk)
               hdivn_tl(ji, jj,jk) = hdivn_tl(ji, jj,jk) - ( 1. / zvt ) * zsu * zu3_ms_tl(jk)
            END DO
         END DO
      END DO

   END SUBROUTINE div_gibraltar_tan

   SUBROUTINE div_gibraltar_adj
      !! -------------------------------------------------------------------
      !!                 ***  ROUTINE div_gibraltar_adj  ***
      !!        
      !! ** Purpose of the direct routine:
      !!     update the now horizontal divergence of the velocity 
      !!     field in Gibraltar.
      !!
      !! ** Method of the direct routine:
      !!          ________________      N        ________________
      !! 102           |    |->         |           <-|    |<-
      !! 101      ___->|____|_____   W - - E     ___->|____|_____
      !!           139   140  141       |         139   140  141
      !!          horizontal view       S        horizontal view
      !!            surface                          depth
      !!      The now divergence is given by :
      !!         hdivn = 1/(e1t*e2t) [ di(e2u  un) + dj(e1v  vn) ]
      !!
      !! ** History of the direct routine:
      !!           !         (A. Bozec) Original code 
      !!      8.5  !  02-10  (A. Bozec) F90: Free form and module
      !! ** History of the adjoint routine:
      !!           !  08-09  (A. Vidard) Adjoint of the 02-10 version
      !!      9.0  !  09-02 (A. Vidard) cosmetic changes
      !!---------------------------------------------------------------------
      !! * Local declarations
      INTEGER  :: ji, jj, jk   ! dummy loop indices
      REAL(wp) :: zsu, zvt
      REAL(wp) :: zwei
      REAL(wp), DIMENSION (jpk) ::  zu1_ms_ad, zu2_ms_ad, zu3_ms_ad
      !!---------------------------------------------------------------------
      
      zu1_ms_ad = 0.0_wp
      zu2_ms_ad = 0.0_wp
      zu3_ms_ad = 0.0_wp
      zempmed_ad = 0.e0
      zwei = 0.e0

      ! Divergence at each point of the straits
      ! ---------------------------------------

      ! compute the divergence at 141,102 East point
      DO jk = jpk, 1, -1
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(141), mi1(141) 
               zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
               zsu = e2u(ji-1, jj) * fse3u(ji-1, jj,jk)
               zu3_ms_ad(jk) = zu3_ms_ad(jk) - hdivn_ad(ji, jj,jk) * ( 1. / zvt ) * zsu
            END DO
         END DO
      END DO

      ! compute the new divergence at 139,102 deep North point
      DO jk = jpk, 1, -1
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(139), mi1(139) 
               zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
               zsu = e2u(ji, jj) * fse3u(ji, jj,jk)
               zu2_ms_ad(jk) =  zu2_ms_ad(jk) + hdivn_ad(ji, jj,jk) * ( 1. / zvt ) * zsu 
            END DO
         END DO
      END DO

      ! compute the new divergence at 139,101 South point  
      DO jk = jpk, 1, -1
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
               zsu = e2u(ji, jj) * fse3u(ji, jj,jk)
               zu1_ms_ad(jk) = zu1_ms_ad(jk) + hdivn_ad(ji, jj,jk) * ( 1. / zvt ) * zsu 
            END DO
         END DO
      END DO

      ! Velocity profile at each point
      ! ------------------------------
      
      ! velocity profile at 141,102  East point
      ! flux in surface inflow of the Aadantic ocean + EMP   
      DO  jk = 14, 1, -1                     
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zempmed_ad = zempmed_ad + ( zu3_ms_ad(jk) / 14. ) / ( e2u(ji, jj) * fse3u(ji, jj,jk) ) 
            END DO
         END DO
      END DO

      ! velocity profile at 139,101 South point 
      ! we imposed zisw + EMP above the Mediterranean Sea 
      DO jk = 14, 1, -1                      
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zempmed_ad = zempmed_ad + ( zu1_ms_ad(jk) / 14. ) / ( e2u(ji-1, jj-1) * fse3u(ji-1, jj-1,jk) ) 
            END DO
         END DO
      END DO
      zu1_ms_ad(:) = 0.0_wp
      zu2_ms_ad(:) = 0.0_wp
      zu3_ms_ad(:) = 0.0_wp

      ! EMP on the Mediterranean Sea 
      ! ----------------------------

      ! convert in m3
      zempmed_ad = zempmed_ad * 1.e-3

      ! minus 2 points in Red Sea and 3 in Aadantic 
      DO jj = mj0(96), mj1(96)
         DO ji = mi0(148),mi1(148)
            emp_ad(ji  , jj) = emp_ad(ji  , jj) - zempmed_ad * tmask(ji  , jj,1) * e1t(ji  , jj) * e2t(ji  , jj)
            emp_ad(ji+1, jj) = emp_ad(ji+1, jj) - zempmed_ad * tmask(ji+1, jj,1) * e1t(ji+1, jj) * e2t(ji+1, jj)   
         END DO
      END DO
      IF( lk_mpp )   CALL mpp_sum( zempmed_ad )      ! sum with other processors value

      DO jj = mj1(110), mj0(96), -1
         DO ji = mi1(181), mi0(141), -1
            zwei    = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            emp_ad(ji,jj) = emp_ad(ji,jj) + zempmed_ad * zwei
         END DO
      END DO
      zempmed_ad = 0.e0

   END SUBROUTINE div_gibraltar_adj



   SUBROUTINE div_hormuz_tan
      !! -------------------------------------------------------------------
      !!                   ***  ROUTINE div_hormuz_tan  ***
      !!              
      !! ** Purpose of the direct routine:
      !!     update the now horizontal divergence of the velocity 
      !!     field in Hormuz ( Persic Gulf strait ) .
      !!
      !! ** Method  of the direct routine:
      !!      The now divergence is given by :
      !!         hdivn = 1/(e1t*e2t) [ di(e2u  un) + dj(e1v  vn) ]
      !!
      !! ** History of the direct routine  
      !!           !         (A. Bozec) Original code 
      !!      8.5  !  02-10  (A. Bozec) F90: Free form and module
      !! ** History of the tangent routine:
      !!           !  08-09  (A. Vidard) Tangent linear of the 02-10 version
      !!      9.0  !  09-02 (A. Vidard) cosmetic changes
      !!---------------------------------------------------------------------
      !! * Local declarations
      !... nothing

   END SUBROUTINE div_hormuz_tan
   SUBROUTINE div_hormuz_adj
      !! -------------------------------------------------------------------
      !!                   ***  ROUTINE div_hormuz_adj  ***
      !!              
      !! ** Purpose of the direct routine:
      !!     update the now horizontal divergence of the velocity 
      !!     field in Hormuz ( Persic Gulf strait ) .
      !!
      !! ** Method  of the direct routine:
      !!      The now divergence is given by :
      !!         hdivn = 1/(e1t*e2t) [ di(e2u  un) + dj(e1v  vn) ]
      !!
      !! ** History of the direct routine  
      !!           !         (A. Bozec) Original code 
      !!      8.5  !  02-10  (A. Bozec) F90: Free form and module
      !! ** History of the Adjoint routine:
      !!           !  08-09  (A. Vidard) Adjoint of the 02-10 version
      !!   9.0     !  09-02 (A. Vidard) cosmetic changes
      !!---------------------------------------------------------------------
      !! * Local declarations
      !... nothing

   END SUBROUTINE div_hormuz_adj

   SUBROUTINE div_cla_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE dyn_adv_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !! ** Action  : Separate tests are applied for the following dx and dy:
      !!                
      !!              1) dx = ( SSH ) and dy = ( SSH )
      !!                   
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      INTEGER :: &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator

      !! * Local declarations
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & zhdivn_tlin,   &
         & zhdivn_tlout,  &
         & zhdivn_adin,   &
         & zhdivn_adout,  &
         & zhdiv
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zemp_tlin,     &
         & zemp_adout,    &
         & zemp
     REAL(KIND=wp) :: &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp2,         & ! scalar product involving the adjoint routine
         & zsp2_1,       & !   scalar product components
         & zsp2_2,       & 
         & z2dt,         & ! temporary scalars
         & zraur
      CHARACTER(LEN=14) :: cl_name

      ! Allocate memory

      ALLOCATE( &
         & zhdivn_tlin(jpi,jpj,jpk),   &
         & zhdivn_tlout(jpi,jpj,jpk),  &
         & zhdivn_adin(jpi,jpj,jpk),   &
         & zhdivn_adout(jpi,jpj,jpk),  &
         & zhdiv(jpi,jpj,jpk),         &
         & zemp_tlin(jpi,jpj),         &
         & zemp_adout(jpi,jpj),        &
         & zemp (jpi,jpj)              &
         & )
      ! Initialize constants

      z2dt  = 2.0_wp * rdt       ! time step: leap-frog
      zraur = 1.0_wp / rauw      ! inverse density of pure water (m3/kg)
      !==================================================================
      ! 1) dx = ( un_tl, vn_tl, hdivn_tl ) and 
      !    dy = ( hdivb_tl, hdivn_tl )
      !==================================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
      zhdivn_tlin(:,:,:)  = 0.0_wp   
      zhdivn_tlout(:,:,:) = 0.0_wp
      zhdivn_adin(:,:,:)  = 0.0_wp 
      zhdivn_adout(:,:,:) = 0.0_wp
      zemp_tlin(:,:)      = 0.0_wp
      zemp_adout(:,:)     = 0.0_wp

      emp_tl(:,:)      = 0.0_wp
      emp_ad(:,:)      = 0.0_wp
      hdivn_tl(:,:,:)  = 0.0_wp
      hdivn_ad(:,:,:)  = 0.0_wp

          
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zemp, 'T', 0.0_wp, stdemp )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 523432 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zhdiv, 'T', 0.0_wp, stdh )

      DO jk = 1, jpk
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zhdivn_tlin(ji,jj,jk)  = zhdiv(ji,jj,jk)
            END DO
         END DO
      END DO 
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zemp_tlin(ji,jj)      = zemp(ji,jj) / ( z2dt * zraur )
         END DO
      END DO 
      hdivn_tl(:,:,:)  = zhdivn_tlin(:,:,:)
      emp_tl(:,:)      = zemp_tlin(:,:) 

      CALL div_cla_tan( nit000 )

      zhdivn_tlout(:,:,:) = hdivn_tl(:,:,:)

      !--------------------------------------------------------------------
      ! Initialize the adjoint variables: dy^* = W dy
      !--------------------------------------------------------------------

      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zhdivn_adin(ji,jj,jk) = zhdivn_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk)
              
            END DO
         END DO
      END DO
      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1 = DOT_PRODUCT( zhdivn_tlout, zhdivn_adin )

      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------
      hdivn_ad(:,:,:) = zhdivn_adin(:,:,:)

      CALL div_cla_adj( nit000 )

      zhdivn_adout(:,:,:) = hdivn_ad(:,:,:) 
      zemp_adout(:,:)     = emp_ad(:,:)

      zsp2_1 = DOT_PRODUCT( zhdivn_tlin, zhdivn_adout )
      zsp2_2 = DOT_PRODUCT( zemp_tlin,   zemp_adout   )
      zsp2   = zsp2_1 + zsp2_2 

      ! Compare the scalar products

      ! 14 char:'12345678901234'
      cl_name = 'div_cla_adj   '
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      DEALLOCATE( &
         & zhdivn_tlin,   &
         & zhdivn_tlout,  &
         & zhdivn_adin,   &
         & zhdivn_adout,  &
         & zhdiv,         &
         & zemp_tlin,     &
         & zemp_adout,    &
         & zemp           &
         & )


   END SUBROUTINE div_cla_adj_tst
   !!======================================================================
#  else
   !!----------------------------------------------------------------------
   !!   Default key                                            Dummy module
   !!----------------------------------------------------------------------
CONTAINS
   SUBROUTINE div_cla_tan( kt )
      WRITE(*,*) 'div_cla: You should have not see this print! error?', kt
   END SUBROUTINE div_cla_tan
   SUBROUTINE div_cla_adj( kt )
      WRITE(*,*) 'div_cla_adj: You should have not see this print! error?', kt
   END SUBROUTINE div_cla_adj
   SUBROUTINE div_cla_adj_tst( kt )
      WRITE(*,*) 'div_cla_adj_tst: You should have not see this print! error?', kt
   END SUBROUTINE div_cla_adj_tst
#  endif
#endif
END MODULE cla_div_tam
