MODULE cla_tam
#ifdef key_tam
   !!==============================================================================
   !!                         ***  MODULE  cla_tam  ***
   !! Cross Land Advection : parameterize ocean exchanges through straits by a
   !!                        specified advection across land. 
   !!                        Tangent and Adjoint module
   !!==============================================================================
#  if defined key_orca_r2
   !!----------------------------------------------------------------------
   !!   'key_orca_r2'   :                             ORCA R2 configuration
   !!----------------------------------------------------------------------
   !!   tra_cla           : update the tracer trend with the horizontal 
   !!                       and vertical advection trends at straits
   !!   tra_bab_el_mandeb : 
   !!   tra_gibraltar     :
   !!   tra_hormuz        :
   !!   tra_cla_init      :
   !!----------------------------------------------------------------------
   !! * Modules used
   USE par_kind,       ONLY: & ! Precision variables
      & wp
   USE par_oce,        ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 & 
      & jpiglo
   USE oce,            ONLY: & ! ocean dynamics and tracers variables
      & tn,                  &
      & sn
   USE sbc_oce_tam,    ONLY: & ! surface variables
      & emp_tl,              &
      & emp_ad
   USE oce_tam,        ONLY: & ! ocean dynamics and tracers variables
      & tn_tl,               &
      & sn_tl,               &
      & ta_tl,               &
      & sa_tl,               &
      & tn_ad,               &
      & sn_ad,               &
      & ta_ad,               &
      & sa_ad
   USE sbc_oce,        ONLY: & ! ocean surface boundary condition (fluxes)
      & emp
   USE phycst,         ONLY: & ! Physical constants
      & rauw
   USE dom_oce,        ONLY: & ! ocean space and time domain
      & mi0,                 &
      & mi1,                 &
      & mj0,                 &
      & mj1,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej,                &
      & rdt,                 &
      & tmask,               &
      & mig,                 &
      & mjg,                 &
      & e2u,                 &
      & e1v,                 &
      & e1t,                 &
      & e2t,                 &
# if defined key_vvl
      & e3t_1,               &
# else
#  if defined key_zco
      & e3t_0,               &
#  else
      & e3t,                 &
#  endif
# endif
# if ! defined key_zco
      & e3u,                 &
      & e3v
# endif
   USE in_out_manager, ONLY: & ! I/O manager 
      & nit000,              &
      & nitend,              &
      & numout,              &
      & lwp
   USE lib_mpp,        ONLY: & ! distributed memory computing library
      & lk_mpp,              &
      & mpp_sum
   USE paresp,         ONLY: &
      & wesp_t,              &
      & wesp_s
   USE gridrandom,     ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld,    ONLY : & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE tstool_tam,     ONLY: &
      & prntst_adj,          & !
      & stdemp,              & ! stdev for evaporation minus precip
      & stdt,                & !           temperature
      & stds                   !           salinity

   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC tra_cla_tan        ! routine called by step_tam.F90
   PUBLIC tra_cla_adj        ! routine called by step_tam.F90
   PUBLIC tra_cla_adj_tst    ! routine called by tst.F90

   !! * Modules variables   
   REAL(wp) :: zempmed, zempred
   REAL(wp) :: zempmed_tl, zempred_tl
   REAL(wp) :: zempmed_ad, zempred_ad

   REAL(wp) :: zisw_rs, zurw_rs, zbrw_rs          ! Imposed transport Red Sea 
   REAL(wp) :: zisw_ms, zmrw_ms, zurw_ms, zbrw_ms ! Imposed transport Med Sea 
   REAL(wp), DIMENSION(jpk) ::      &
      zu1_rs_i, zu2_rs_i, zu3_rs_i, &  ! Red Sea velocities
      zu1_ms_i, zu2_ms_i, zu3_ms_i     ! Mediterranean Sea velocities
   LOGICAL :: lfirst = .TRUE. ! initialisation flag
   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"

CONTAINS

   SUBROUTINE tra_cla_tan( kt )
      !!----------------------------------------------------------------------
      !!                 ***  ROUTINE tra_cla_tan  ***
      !!                   
      !! ** Purpose of the direct reoutine:
      !!      Update the now trend due to the advection of tracers
      !!      and add it to the general trend of passive tracer equations
      !!      at some straits ( Bab el Mandeb, Gibraltar, Hormuz ).
      !!
      !! ** Method  :   ...
      !!         Add this trend now to the general trend of tracer (ta,sa):
      !!            (ta,sa) = (ta,sa) + ( zta , zsa )
      !!
      !! ** Action  :   update (ta,sa) with the now advective tracer trends
      !!
      !! History of the direct method:
      !!        !         (A. Bozec)  original code  
      !!   8.5  !  02-11  (A. Bozec)  F90: Free form and module
      !! History of the TAM:
      !!        !  08-08  (A. Vidard) tangent of the 02-11 vesrsion
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt         ! ocean time-step index
      !!----------------------------------------------------------------------
 
      ! cross land advection for straits

      ! Initialization
      IF( kt == nit000 )   CALL tra_cla_init_tam


      ! Bab el Mandeb strait horizontal advection

      CALL tra_bab_el_mandeb_tan 

      ! Gibraltar strait horizontal advection

      CALL tra_gibraltar_tan

      ! Hormuz Strait ( persian Gulf) horizontal advection

      CALL tra_hormuz_tan 



   END SUBROUTINE tra_cla_tan
   SUBROUTINE tra_cla_adj( kt )
      !!----------------------------------------------------------------------
      !!                 ***  ROUTINE tra_cla_adj  ***
      !!                   
      !! ** Purpose of the direct reoutine:
      !!      Update the now trend due to the advection of tracers
      !!      and add it to the general trend of passive tracer equations
      !!      at some straits ( Bab el Mandeb, Gibraltar, Hormuz ).
      !!
      !! ** Method  :   ...
      !!         Add this trend now to the general trend of tracer (ta,sa):
      !!            (ta,sa) = (ta,sa) + ( zta , zsa )
      !!
      !! ** Action  :   update (ta,sa) with the now advective tracer trends
      !!
      !! History of the direct method:
      !!        !         (A. Bozec)  original code  
      !!   8.5  !  02-11  (A. Bozec)  F90: Free form and module
      !! History:
      !!        !  08-06  (A. Vidard) adjoint of the 02-11 vesrsion
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt         ! ocean time-step index
      !!----------------------------------------------------------------------
 
      ! cross land advection for straits

      ! Initialization
      IF( kt == nitend )   CALL tra_cla_init_tam


      ! Hormuz Strait ( persian Gulf) horizontal advection
      CALL tra_hormuz_adj 
      ! Gibraltar strait horizontal advection

      CALL tra_gibraltar_adj
      ! Bab el Mandeb strait horizontal advection

      CALL tra_bab_el_mandeb_adj 

   END SUBROUTINE tra_cla_adj

   SUBROUTINE tra_bab_el_mandeb_tan
      !!---------------------------------------------------------------------
      !!             ***  ROUTINE tra_bab_el_mandeb_tan  ***
      !!
      !! ** Purpose of the direct routine:
      !!      Update the horizontal advective trend of tracers
      !!      correction in Bab el Mandeb strait and
      !!      add it to the general trend of tracer equations.
      !!
      !! ** Method of the direct routine:
      !!     We impose transport at Bab el Mandeb and knowing T and S in
      !!     surface and depth at each side of the  strait, we deduce T and S
      !!     of the deep outflow of the Red Sea in the Indian ocean . 
      !!                                          |
      !!            |/ \|            N          |\ /|
      !!            |_|_|______      |          |___|______
      !!        88  |   |<-       W - - E    88 |   |<-
      !!        87  |___|______      |       87 |___|->____
      !!             160 161         S           160 161 
      !!       horizontal view                horizontal view
      !!          surface                        depth
      !!    
      !!     The horizontal advection is evaluated by a second order cen-
      !!     tered scheme using now fields (leap-frog scheme). In specific
      !!     areas (vicinity of major river mouths, some straits, or tn
      !!     approaching the freezing point) it is mixed with an upstream 
      !!     scheme for stability reasons. 
      !!
      !!         C A U T I O N : the trend saved is the centered trend only.
      !!      It doesn't take into account the upstream part of the scheme.
      !!
      !! ** history of the direct routine:
      !!           !  02-11  (A. Bozec) Original code 
      !!      8.5  !  02-11  (A. Bozec) F90: Free form and module
      !! ** history of the tangent routine:
      !!           !  08-08  (A. Vidard)
      !!---------------------------------------------------------------------
      !! * Local declarations
      INTEGER ::  ji, jj, jk               ! dummy loop indices
      REAL(wp) :: zsu, zvt
      REAL(wp) :: zsumt, zsumt1, zsumt2, zsumt3, zsumt4
      REAL(wp) :: zsums, zsums1, zsums2, zsums3, zsums4
      REAL(wp) :: zsumttl, zsumt1tl, zsumt2tl, zsumt3tl, zsumt4tl
      REAL(wp) :: zsumstl, zsums1tl, zsums2tl, zsums3tl, zsums4tl
      REAL(wp) :: zt, zs
      REAL(wp) :: zttl, zstl
      REAL(wp) :: zwei
      REAL(wp), DIMENSION (jpk) ::  zu1_rs, zu2_rs, zu3_rs
      REAL(wp), DIMENSION (jpk) ::  zu1_rs_tl, zu2_rs_tl, zu3_rs_tl
      !!---------------------------------------------------------------------

      ! Initialization of vertical sum for T and S transport
      ! ----------------------------------------------------

      zsumt  = 0.e0       ! East  Bab el Mandeb surface north point (T)
      zsums  = 0.e0       ! East  Bab el Mandeb surface north point (S)
      zsumt1 = 0.e0       ! East  Bab el Mandeb depth   south point (T)
      zsums1 = 0.e0       ! East  Bab el Mandeb depth   south point (S)
      zsumt2 = 0.e0       ! West  Bab el Mandeb surface             (T)
      zsums2 = 0.e0       ! West  Bab el Mandeb surface             (S)
      zsumt3 = 0.e0       ! West  Bab el Mandeb depth               (T)
      zsums3 = 0.e0       ! West  Bab el Mandeb depth               (S)
      zsumt4 = 0.e0       ! East  Bab el Mandeb depth   north point (T) 
      zsums4 = 0.e0       ! East  Bab el Mandeb depth   north point (S) 
      zsumttl  = 0.e0       ! East  Bab el Mandeb surface north point (T)
      zsumstl  = 0.e0       ! East  Bab el Mandeb surface north point (S)
      zsumt1tl = 0.e0       ! East  Bab el Mandeb depth   south point (T)
      zsums1tl = 0.e0       ! East  Bab el Mandeb depth   south point (S)
      zsumt2tl = 0.e0       ! West  Bab el Mandeb surface             (T)
      zsums2tl = 0.e0       ! West  Bab el Mandeb surface             (S)
      zsumt3tl = 0.e0       ! West  Bab el Mandeb depth               (T)
      zsums3tl = 0.e0       ! West  Bab el Mandeb depth               (S)
      zsumt4tl = 0.e0       ! East  Bab el Mandeb depth   north point (T) 
      zsums4tl = 0.e0       ! East  Bab el Mandeb depth   north point (S) 
      
      ! EMP of the Red Sea 
      ! ------------------

      zempred_tl = 0.e0
      zempred = 0.e0
      zwei = 0.e0
      DO jj = mj0(87), mj1(96)
         DO ji = mi0(148), mi1(160)
            zwei    = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            zempred    = zempred    + emp(ji,jj)    * zwei
            zempred_tl = zempred_tl + emp_tl(ji,jj) * zwei
         END DO
      END DO
      IF( lk_mpp )   CALL mpp_sum( zempred    )      ! sum with other processors value
      IF( lk_mpp )   CALL mpp_sum( zempred_tl )      ! sum with other processors value

      ! convert in m3
      zempred_tl = zempred_tl * 1.e-3
      zempred    = zempred    * 1.e-3

      ! Velocity profile at each point
      ! ------------------------------

      zu1_rs(:) = zu1_rs_i(:)
      zu2_rs(:) = zu2_rs_i(:)
      zu3_rs(:) = zu3_rs_i(:)
      zu1_rs_tl(:) = 0.0_wp
      zu2_rs_tl(:) = 0.0_wp
      zu3_rs_tl(:) = 0.0_wp

      ! velocity profile at 161,88 East Bab el Mandeb North point 
      ! we imposed zisw_rs + EMP above the Red Sea 
      DO jk = 1, 8                                      
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160) 
               zu1_rs(jk)    = zu1_rs(jk)    &
                  &           - ( zempred / 8. ) / ( e2u(ji,jj) * fse3u(ji,jj,jk) ) 
               zu1_rs_tl(jk) = zu1_rs_tl(jk) &
                  &           - ( zempred_tl / 8. ) / ( e2u(ji,jj) * fse3u(ji,jj,jk) )     
            END DO
         END DO
      END DO

      ! velocity profile at 161, 88 West Bab el Mandeb 
      ! we imposed zisw_rs + EMP above the Red Sea 
      DO jk = 1,  10                                     
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160) 
               zu3_rs(jk)    = zu3_rs(jk)    &
                  &          + ( zempred    / 10. ) / ( e1v(ji,jj) * fse3v(ji,jj,jk) )
               zu3_rs_tl(jk) = zu3_rs_tl(jk) &
                  &          + ( zempred_tl / 10. ) / ( e1v(ji,jj) * fse3v(ji,jj,jk) )
            END DO
         END DO
      END DO
      
      ! Balance of temperature and salinity
      ! -----------------------------------

      ! east Bab el Mandeb surface vertical sum of transport* S,T
      DO jk =  1, 19
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(161), mi1(161) 
               zsumt    = zsumt    &
                  &     + tn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk)
               zsums    = zsums    &
                  &     + sn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk) 
               zsumttl  = zsumttl  &
                  &     + tn_tl(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk) &
                  &     + tn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs_tl(jk)
               zsumstl  = zsumstl  &
                  &     + sn_tl(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk) &
                  &     + sn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs_tl(jk) 
            END DO
         END DO
      END DO

      ! west Bab el Mandeb surface vertical sum of transport* S,T
      DO jk =  1, 10
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(161), mi1(161) 
               zsumt2   = zsumt2     &
                  &     + tn(ji,jj,jk) * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) * zu3_rs(jk)
               zsums2   = zsums2     &
                  &     + sn(ji,jj,jk) * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) * zu3_rs(jk)
               zsumt2tl = zsumt2tl   &
                  &     + tn_tl(ji,jj,jk) * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) * zu3_rs(jk) &
                  &     + tn(ji,jj,jk) * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) * zu3_rs_tl(jk)
               zsums2tl = zsums2tl   &
                  &     + sn_tl(ji,jj,jk) * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) * zu3_rs(jk) &
                  &     + sn(ji,jj,jk) * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) * zu3_rs_tl(jk)
            END DO
         END DO
      END DO

      ! west Bab el Mandeb deeper
      DO jj = mj0(89), mj1(89) 
         DO ji = mi0(160), mi1(160) 
            zsumt3   = tn(ji,jj,16) * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * zu3_rs(16)
            zsums3   = sn(ji,jj,16) * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * zu3_rs(16)
            zsumt3tl = tn_tl(ji,jj,16) * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * zu3_rs(16) &
               &     + tn(ji,jj,16) * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * zu3_rs_tl(16)
            zsums3tl = sn_tl(ji,jj,16) * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * zu3_rs(16) &
               &     + sn(ji,jj,16) * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * zu3_rs_tl(16)
         END DO
      END DO

      ! east  Bab el Mandeb deeper  
      DO jk =  20, 21
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(161), mi1(161) 
               zsumt4   = zsumt4   &
                  &     + tn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk)
               zsums4   = zsums4   &
                  &     + sn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk)
               zsumt4tl = zsumt4tl &
                  &     + tn_tl(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk) &
                  &     + tn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs_tl(jk)
               zsums4tl = zsums4tl &
                  &     + sn_tl(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk) &
                  &     + sn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs_tl(jk)
            END DO
         END DO
      END DO

      ! Total transport  
      zsumt1   = -( zsumt3 + zsumt2 + zsumt + zsumt4 )
      zsums1   = -( zsums3 + zsums2 + zsums + zsums4 )
      zsumt1tl = -( zsumt3tl + zsumt2tl + zsumttl + zsumt4tl )
      zsums1tl = -( zsums3tl + zsums2tl + zsumstl + zsums4tl )

      ! Temperature and Salinity at East Bab el Mandeb, Level 21
      DO jj = mj0(88), mj1(88) 
         DO ji = mi0(160), mi1(160) 
            zt   = zsumt1 / ( zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) )
            zs   = zsums1 / ( zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) )
            zttl = zsumt1tl / ( zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) ) &
               & - zsumt1 / ( 2 * zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) ) * zu2_rs_tl(21)
            zstl = zsums1tl / ( zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) ) &
               & - zsums1 / ( 2 * zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) ) * zu2_rs_tl(21)
         END DO
      END DO
      
      ! New Temperature and Salinity at East Bab el Mandeb
      ! --------------------------------------------------

      ! north point 	
      DO jk = 1, jpk
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(161), mi1(161) 
               zvt             = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu             = e2u(ji-1,jj) * fse3u(ji-1,jj,jk)
               ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) &
                  &             + ( 1. / zvt ) * zsu * zu1_rs_tl(jk) * tn(ji,jj,jk) &
                  &             + ( 1. / zvt ) * zsu * zu1_rs(jk) * tn_tl(ji,jj,jk)
               sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk) &
                  &             + ( 1. / zvt ) * zsu * zu1_rs_tl(jk) * sn(ji,jj,jk) &
                  &             + ( 1. / zvt ) * zsu * zu1_rs(jk) * sn_tl(ji,jj,jk)
            END DO
         END DO
      END DO

      ! south point
      jk =  21
      DO jj = mj0(87), mj1(87) 
         DO ji = mi0(161), mi1(161) 
            zvt             = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
            zsu             = e2u(ji-1,jj) * fse3u(ji-1,jj,jk)
            ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) &
               &            + ( 1. / zvt ) * zsu * zu2_rs_tl(jk) * zt &
               &            + ( 1. / zvt ) * zsu * zu2_rs(jk) * zttl
            sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk) &
               &            + ( 1. / zvt ) * zsu * zu2_rs_tl(jk) * zs &
               &            + ( 1. / zvt ) * zsu * zu2_rs(jk) * zstl
         END DO
      END DO


      ! New Temperature and Salinity at West Bab el Mandeb 
      ! --------------------------------------------------

      ! surface	
      DO jk = 1, 10
         DO jj = mj0(89), mj1(89) 
            DO ji = mi0(160), mi1(160) 
               zvt             = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu             = e1v(ji,jj-1) * fse3v(ji,jj-1,jk)
               ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) &
                  &            + ( 1. / zvt ) * zsu * zu3_rs_tl(jk) * tn(ji+1,jj-1,jk) &
                  &            + ( 1. / zvt ) * zsu * zu3_rs(jk) * tn_tl(ji+1,jj-1,jk)
               sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk) &
                  &            + ( 1. / zvt ) * zsu * zu3_rs_tl(jk) * sn(ji+1,jj-1,jk) &
                  &            + ( 1. / zvt ) * zsu * zu3_rs(jk) * sn_tl(ji+1,jj-1,jk)
            END DO
         END DO
      END DO
      ! deeper
      jk =  16
      DO jj = mj0(89), mj1(89) 
         DO ji = mi0(160), mi1(160) 
            zvt             = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
            zsu             = e1v(ji,jj-1) * fse3v(ji,jj-1,jk)
            ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) &
               &            + ( 1. / zvt ) * zsu * zu3_rs_tl(jk) * tn(ji,jj,jk) &
               &            + ( 1. / zvt ) * zsu * zu3_rs(jk) * tn_tl(ji,jj,jk)
            sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk) &
               &            + ( 1. / zvt ) * zsu * zu3_rs_tl(jk) * sn(ji,jj,jk) &
               &            + ( 1. / zvt ) * zsu * zu3_rs(jk) * sn_tl(ji,jj,jk)
         END DO
      END DO

   END SUBROUTINE tra_bab_el_mandeb_tan
   SUBROUTINE tra_bab_el_mandeb_adj
      !!---------------------------------------------------------------------
      !!             ***  ROUTINE tra_bab_el_mandeb_adj  ***
      !!
      !! ** Purpose of the direct routine:
      !!      Update the horizontal advective trend of tracers
      !!      correction in Bab el Mandeb strait and
      !!      add it to the general trend of tracer equations.
      !!
      !! ** Method of the direct routine:
      !!     We impose transport at Bab el Mandeb and knowing T and S in
      !!     surface and depth at each side of the  strait, we deduce T and S
      !!     of the deep outflow of the Red Sea in the Indian ocean . 
      !!                                          |
      !!            |/ \|            N          |\ /|
      !!            |_|_|______      |          |___|______
      !!        88  |   |<-       W - - E    88 |   |<-
      !!        87  |___|______      |       87 |___|->____
      !!             160 161         S           160 161 
      !!       horizontal view                horizontal view
      !!          surface                        depth
      !!    
      !!     The horizontal advection is evaluated by a second order cen-
      !!     tered scheme using now fields (leap-frog scheme). In specific
      !!     areas (vicinity of major river mouths, some straits, or tn
      !!     approaching the freezing point) it is mixed with an upstream 
      !!     scheme for stability reasons. 
      !!
      !!         C A U T I O N : the trend saved is the centered trend only.
      !!      It doesn't take into account the upstream part of the scheme.
      !!
      !! ** history of the direct routine:
      !!           !  02-11  (A. Bozec) Original code 
      !!      8.5  !  02-11  (A. Bozec) F90: Free form and module
      !! ** history of the adjoint routine:
      !!           !  08-08  (A. Vidard)
      !!---------------------------------------------------------------------
      !! * Local declarations
      INTEGER ::  ji, jj, jk               ! dummy loop indices
      REAL(wp) :: zsu, zvt
      REAL(wp) :: zsumt, zsumt1, zsumt2, zsumt3, zsumt4
      REAL(wp) :: zsums, zsums1, zsums2, zsums3, zsums4
      REAL(wp) :: zsumtad, zsumt1ad, zsumt2ad, zsumt3ad, zsumt4ad
      REAL(wp) :: zsumsad, zsums1ad, zsums2ad, zsums3ad, zsums4ad
      REAL(wp) :: zt, zs
      REAL(wp) :: ztad, zsad
      REAL(wp) :: zwei
      REAL(wp), DIMENSION (jpk) ::  zu1_rs, zu2_rs, zu3_rs
      REAL(wp), DIMENSION (jpk) ::  zu1_rs_ad, zu2_rs_ad, zu3_rs_ad
      !!---------------------------------------------------------------------

      ! Initialization of vertical sum for T and S transport
      ! ----------------------------------------------------

      zsumt    = 0.e0       ! East  Bab el Mandeb surface north point (T)
      zsums    = 0.e0       ! East  Bab el Mandeb surface north point (S)
      zsumt1   = 0.e0       ! East  Bab el Mandeb depth   south point (T)
      zsums1   = 0.e0       ! East  Bab el Mandeb depth   south point (S)
      zsumt2   = 0.e0       ! West  Bab el Mandeb surface             (T)
      zsums2   = 0.e0       ! West  Bab el Mandeb surface             (S)
      zsumt3   = 0.e0       ! West  Bab el Mandeb depth               (T)
      zsums3   = 0.e0       ! West  Bab el Mandeb depth               (S)
      zsumt4   = 0.e0       ! East  Bab el Mandeb depth   north point (T) 
      zsums4   = 0.e0       ! East  Bab el Mandeb depth   north point (S) 
      zsumtad  = 0.e0       ! East  Bab el Mandeb surface north point (T)
      zsumsad  = 0.e0       ! East  Bab el Mandeb surface north point (S)
      zsumt1ad = 0.e0       ! East  Bab el Mandeb depth   south point (T)
      zsums1ad = 0.e0       ! East  Bab el Mandeb depth   south point (S)
      zsumt2ad = 0.e0       ! West  Bab el Mandeb surface             (T)
      zsums2ad = 0.e0       ! West  Bab el Mandeb surface             (S)
      zsumt3ad = 0.e0       ! West  Bab el Mandeb depth               (T)
      zsums3ad = 0.e0       ! West  Bab el Mandeb depth               (S)
      zsumt4ad = 0.e0       ! East  Bab el Mandeb depth   north point (T) 
      zsums4ad = 0.e0       ! East  Bab el Mandeb depth   north point (S) 
      ztad     = 0.e0
      zsad     = 0.e0
      zu1_rs_ad (:) = 0.e0
      zu2_rs_ad (:) = 0.e0
      zu3_rs_ad (:) = 0.e0
      zempred_ad = 0.e0
      !===========================
      ! Direct model recomputation
      !===========================
      ! EMP of the Red Sea 
      ! ------------------

      zempred = 0.e0
      zwei    = 0.e0
      DO jj = mj0(87), mj1(96)
         DO ji = mi0(148), mi1(160)
            zwei    = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            zempred = zempred + emp(ji,jj) * zwei
         END DO
      END DO
      IF( lk_mpp )   CALL mpp_sum( zempred )      ! sum with other processors value

      ! convert in m3
      zempred = zempred * 1.e-3

      ! Velocity profile at each point
      ! ------------------------------

      zu1_rs(:) = zu1_rs_i(:)
      zu2_rs(:) = zu2_rs_i(:)
      zu3_rs(:) = zu3_rs_i(:)

      ! velocity profile at 161,88 East Bab el Mandeb North point 
      ! we imposed zisw_rs + EMP above the Red Sea 
      DO jk = 1, 8                                      
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160) 
               zu1_rs(jk) = zu1_rs(jk) - ( zempred / 8. ) / ( e2u(ji,jj) * fse3u(ji,jj,jk) )     
            END DO
         END DO
      END DO

      ! velocity profile at 161, 88 West Bab el Mandeb 
      ! we imposed zisw_rs + EMP above the Red Sea 
      DO jk = 1,  10                                     
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160) 
               zu3_rs(jk) = zu3_rs(jk) + ( zempred / 10. ) / ( e1v(ji,jj) * fse3v(ji,jj,jk) )
            END DO
         END DO
      END DO
      
      ! Balance of temperature and salinity
      ! -----------------------------------

      ! east Bab el Mandeb surface vertical sum of transport* S,T
      DO jk =  1, 19
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(161), mi1(161) 
               zsumt  = zsumt  + tn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk)  
               zsums  = zsums  + sn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk) 
            END DO
         END DO
      END DO

      ! west Bab el Mandeb surface vertical sum of transport* S,T
      DO jk =  1, 10
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(161), mi1(161) 
               zsumt2 = zsumt2 + tn(ji,jj,jk) * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) * zu3_rs(jk)
               zsums2 = zsums2 + sn(ji,jj,jk) * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) * zu3_rs(jk)
            END DO
         END DO
      END DO

      ! west Bab el Mandeb deeper
      DO jj = mj0(89), mj1(89) 
         DO ji = mi0(160), mi1(160) 
            zsumt3 = tn(ji,jj,16) * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * zu3_rs(16)
            zsums3 = sn(ji,jj,16) * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * zu3_rs(16)
         END DO
      END DO

      ! east  Bab el Mandeb deeper  
      DO jk =  20, 21
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(161), mi1(161) 
               zsumt4 =  zsumt4 + tn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk)
               zsums4 =  zsums4 + sn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk) 
            END DO
         END DO
      END DO

      ! Total transport  
      zsumt1 = -( zsumt3 + zsumt2 + zsumt + zsumt4 )
      zsums1 = -( zsums3 + zsums2 + zsums + zsums4 )

      ! Temperature and Salinity at East Bab el Mandeb, Level 21
      DO jj = mj0(88), mj1(88) 
         DO ji = mi0(160), mi1(160) 
            zt = zsumt1 / ( zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) )
            zs = zsums1 / ( zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) )
         END DO
      END DO
      
      !=============
      ! Adjoint part
      !=============

      ! New Temperature and Salinity at West Bab el Mandeb 
      ! --------------------------------------------------

      ! deeper
      jk =  16
      DO jj = mj1(89), mj0(89), -1
         DO ji = mi1(160), mi0(160), -1
            zvt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
            zsu = e1v(ji,jj-1) * fse3v(ji,jj-1,jk)
            zu3_rs_ad(jk) = zu3_rs_ad(jk) + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * tn(ji,jj,jk)
            tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu3_rs(jk)
            zu3_rs_ad(jk) = zu3_rs_ad(jk) + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * sn(ji,jj,jk)
            sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu3_rs(jk)
         END DO
      END DO
      ! surface	
      DO jk = 10, 1, -1
         DO jj = mj1(89), mj0(89), -1
            DO ji = mi1(160), mi0(160), -1
               zvt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu = e1v(ji,jj-1) * fse3v(ji,jj-1,jk)
               zu3_rs_ad(jk) = zu3_rs_ad(jk) + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * tn(ji+1,jj-1,jk)
               tn_ad(ji+1,jj-1,jk) = tn_ad(ji+1,jj-1,jk) + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu3_rs(jk)
               zu3_rs_ad(jk) = zu3_rs_ad(jk) + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * sn(ji+1,jj-1,jk)
               sn_ad(ji+1,jj-1,jk) = sn_ad(ji+1,jj-1,jk) + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu3_rs(jk)
            END DO
         END DO
      END DO
      ! New Temperature and Salinity at East Bab el Mandeb
      ! --------------------------------------------------

      ! south point
      jk =  21
      DO jj = mj1(87), mj0(87), -1
         DO ji = mi1(161), mi0(161), -1 
            zvt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
            zsu = e2u(ji-1,jj) * fse3u(ji-1,jj,jk)
            zu2_rs_ad(jk) = zu2_rs_ad(jk) + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zt
            ztad = ztad + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu2_rs(jk)
            zu2_rs_ad(jk) = zu2_rs_ad(jk) + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zs
            zsad = zsad + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu2_rs(jk)
         END DO
      END DO

      ! north point 	
      DO jk = jpk, 1, -1
         DO jj = mj1(88), mj0(88), -1
            DO ji = mi1(161), mi0(161), -1
               zvt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu = e2u(ji-1,jj) * fse3u(ji-1,jj,jk)
               zu1_rs_ad(jk) = zu1_rs_ad(jk) + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * tn(ji,jj,jk)
               tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu1_rs(jk) 
               zu1_rs_ad(jk) = zu1_rs_ad(jk) + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * sn(ji,jj,jk)
               sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu1_rs(jk) 
            END DO
         END DO
      END DO

      ! Balance of temperature and salinity
      ! -----------------------------------

      ! Temperature and Salinity at East Bab el Mandeb, Level 21
      DO jj = mj1(88), mj0(88), -1
         DO ji = mi1(160), mi0(160), -1
            zsumt1ad = zsumt1ad + ztad / ( zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) )
            zu2_rs_ad(21) = zu2_rs_ad(21) - ztad * zsumt1 / ( 2 * zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) )
            ztad = 0.0_wp
            zsums1ad = zsums1ad + zsad / ( zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) )
            zu2_rs_ad(21) = zu2_rs_ad(21) - zsad * zsums1 / ( 2 * zu2_rs(21) * e2u(ji,jj-1) * fse3u(ji,jj-1,21) )
            zsad = 0.0_wp
         END DO
      END DO
      ! Total transport  
      zsumt3ad = zsumt3ad - zsumt1ad
      zsumt2ad = zsumt2ad - zsumt1ad
      zsumtad  = zsumtad  - zsumt1ad
      zsumt4ad = zsumt4ad - zsumt1ad
      zsumt1ad = 0.0_wp
      zsums3ad = zsums3ad - zsums1ad
      zsums2ad = zsums2ad - zsums1ad
      zsumsad  = zsumsad  - zsums1ad
      zsums4ad = zsums4ad - zsums1ad
      zsums1ad = 0.0_wp

      ! east  Bab el Mandeb deeper  
      DO jk =  21, 20, -1
         DO jj = mj1(88), mj0(88), -1
            DO ji = mi1(161), mi0(161), -1
               tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) + zsumt4ad * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk)
               zu1_rs_ad(jk)   = zu1_rs_ad(jk)   + zsumt4ad * tn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk)
               sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) + zsums4ad * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk)
               zu1_rs_ad(jk)   = zu1_rs_ad(jk)   + zsums4ad * sn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk)
            END DO
         END DO
      END DO

      ! west Bab el Mandeb deeper
      DO jj = mj1(89), mj0(89), -1
         DO ji = mi1(160), mi0(160), -1
            tn_ad(ji,jj,16) = tn_ad(ji,jj,16) + zsumt3ad * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * zu3_rs(16)
            zu3_rs_ad(16)   = zu3_rs_ad(16)   + zsumt3ad * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * tn(ji,jj,16)
            sn_ad(ji,jj,16) = sn_ad(ji,jj,16) + zsums3ad * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * zu3_rs(16)
            zu3_rs_ad(16)   = zu3_rs_ad(16)   + zsums3ad * e1v(ji,jj-1) * fse3v(ji,jj-1,16) * sn(ji,jj,16)
            zsumt3ad = 0.0_wp
            zsums3ad = 0.0_wp
         END DO
      END DO

      ! west Bab el Mandeb surface vertical sum of transport* S,T
      DO jk =  10, 1, -1
         DO jj = mj1(88), mj0(88), -1
            DO ji = mi1(161), mi0(161), -1
               tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) + zsumt2ad * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) * zu3_rs(jk)
               zu3_rs_ad(jk)   = zu3_rs_ad(jk)   + zsumt2ad * tn(ji,jj,jk) * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) 
               sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) + zsums2ad * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) * zu3_rs(jk)
               zu3_rs_ad(jk)   = zu3_rs_ad(jk)   + zsums2ad * sn(ji,jj,jk) * e1v(ji-1,jj) * fse3v(ji-1,jj,jk) 
            END DO
         END DO
      END DO

      ! east Bab el Mandeb surface vertical sum of transport* S,T
      DO jk =  19, 1, -1
         DO jj = mj1(88), mj0(88), -1
            DO ji = mi1(161), mi0(161), -1
               tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) + zsumtad * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk)
               zu1_rs_ad(jk)   = zu1_rs_ad(jk)   + zsumtad * tn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) 
               sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) + zsumsad * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * zu1_rs(jk)
               zu1_rs_ad(jk)   = zu1_rs_ad(jk)   + zsumsad * sn(ji,jj,jk) * e2u(ji-1,jj) * fse3u(ji-1,jj,jk) 
            END DO
         END DO
      END DO

      ! Velocity profile at each point
      ! ------------------------------
      ! velocity profile at 161, 88 West Bab el Mandeb 
      ! we imposed zisw_rs + EMP above the Red Sea 
      DO jk = 10, 1, -1                                     
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160) 
               zempred_ad = zempred_ad + zu3_rs_ad(jk) / ( 10. * e1v(ji,jj) * fse3v(ji,jj,jk) )
            END DO
         END DO
      END DO

      ! velocity profile at 161,88 East Bab el Mandeb North point 
      ! we imposed zisw_rs + EMP above the Red Sea 
      DO jk = 8, 1, -1                                      
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160) 
               zempred_ad = zempred_ad - zu1_rs_ad(jk) / ( 8. * e2u(ji,jj) * fse3u(ji,jj,jk) )
            END DO
         END DO
      END DO

      zu1_rs(:) = zu1_rs_i(:)
      zu2_rs(:) = zu2_rs_i(:)
      zu3_rs(:) = zu3_rs_i(:)
      zu1_rs_ad(:) = 0.0_wp
      zu2_rs_ad(:) = 0.0_wp
      zu3_rs_ad(:) = 0.0_wp

      ! EMP of the Red Sea 
      ! ------------------

      ! convert in m3
      zempred_ad = zempred_ad * 1.e-3

      IF( lk_mpp )   CALL mpp_sum( zempred_ad )      ! sum with other processors value
      DO jj = mj0(87), mj1(96)
         DO ji = mi0(148), mi1(160)
            zwei    = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            emp_ad(ji,jj) = emp_ad(ji,jj) + zempred_ad * zwei
         END DO
      END DO

      zempred_ad = 0.e0
      zwei = 0.e0
 
   END SUBROUTINE tra_bab_el_mandeb_adj
   SUBROUTINE tra_gibraltar_tan
      !!---------------------------------------------------------------------
      !!               ***  ROUTINE tra_gibraltar_tan  ***
      !!
      !! ** Purpose :
      !!        Update the horizontal advective trend of tracers (t & s)
      !!        correction in Gibraltar  and
      !!        add it to the general trend of tracer equations.
      !!
      !! ** Method :
      !!      We impose transport at Gibraltar and knowing T and S in
      !!      surface and deeper at each side of the strait, we deduce T and S
      !!      of the outflow of the Mediterranean Sea in the Atlantic ocean .
      !!                                
      !!          ________________      N        ________________
      !! 102           |    |->         |           <-|    |<-
      !! 101      ___->|____|_____   W - - E     ___->|____|_____
      !!           139   140  141       |         139   140  141
      !!          horizontal view       S        horizontal view
      !!            surface                          depth
      !!         C A U T I O N : the trend saved is the centered trend only.
      !!      It doesn't take into account the upstream part of the scheme.
      !!
      !! ** history :
      !!           !  02-06  (A. Bozec) Original code 
      !!      8.5  !  02-11  (A. Bozec) F90: Free form and module
      !! ** history of the tangent routine:
      !!           !  08-08  (A. Vidard) tangent of the 02-11 version
      !!---------------------------------------------------------------------
      !! * Local declarations
      INTEGER ::  ji, jj, jk               ! dummy loop indices
      REAL(wp) :: zsu, zvt
      REAL(wp) :: zsumt, zsumt1, zsumt2, zsumt3, zsumt4
      REAL(wp) :: zsums, zsums1, zsums2, zsums3, zsums4
      REAL(wp) :: zsumttl, zsumt1tl, zsumt2tl, zsumt3tl, zsumt4tl
      REAL(wp) :: zsumstl, zsums1tl, zsums2tl, zsums3tl, zsums4tl
      REAL(wp) :: zt, zs
      REAL(wp) :: zttl, zstl
      REAL(wp) :: zwei
      REAL(wp), DIMENSION (jpk) ::  zu1_ms, zu2_ms, zu3_ms
      REAL(wp), DIMENSION (jpk) ::  zu1_ms_tl, zu2_ms_tl, zu3_ms_tl
      !!---------------------------------------------------------------------

      ! Initialization of vertical sum for T and S transport
      ! ----------------------------------------------------

      zsumt  = 0.e0        ! West Gib. surface south point ( T )
      zsums  = 0.e0        ! West Gib. surface south point ( S )
      zsumt1 = 0.e0        ! East Gib. surface north point ( T )
      zsums1 = 0.e0        ! East Gib. surface north point ( S )
      zsumt2 = 0.e0        ! East Gib. depth   north point ( T )
      zsums2 = 0.e0        ! East Gib. depth   north point ( S )
      zsumt3 = 0.e0        ! West Gib. depth   south point ( T ) 
      zsums3 = 0.e0        ! West Gib. depth   south point ( S ) 
      zsumt4 = 0.e0        ! West Gib. depth   north point ( T ) 
      zsums4 = 0.e0        ! West Gib. depth   north point ( S ) 
      zsumttl  = 0.e0        ! West Gib. surface south point ( T )
      zsumstl  = 0.e0        ! West Gib. surface south point ( S )
      zsumt1tl = 0.e0        ! East Gib. surface north point ( T )
      zsums1tl = 0.e0        ! East Gib. surface north point ( S )
      zsumt2tl = 0.e0        ! East Gib. depth   north point ( T )
      zsums2tl = 0.e0        ! East Gib. depth   north point ( S )
      zsumt3tl = 0.e0        ! West Gib. depth   south point ( T ) 
      zsums3tl = 0.e0        ! West Gib. depth   south point ( S ) 
      zsumt4tl = 0.e0        ! West Gib. depth   north point ( T ) 
      zsums4tl = 0.e0        ! West Gib. depth   north point ( S ) 
       
      ! EMP of Mediterranean Sea 
      ! ------------------------
      
      zempmed_tl = 0.e0
      zempmed    = 0.e0
      zwei       = 0.e0
      DO jj = mj0(96), mj1(110)
         DO ji = mi0(141), mi1(181)
            zwei       = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            zempmed    = zempmed    + emp   (ji,jj) * zwei
            zempmed_tl = zempmed_tl + emp_tl(ji,jj) * zwei
         END DO
      END DO
      IF( lk_mpp )   CALL mpp_sum( zempmed    )      ! sum with other processors value
      IF( lk_mpp )   CALL mpp_sum( zempmed_tl )      ! sum with other processors value

      ! minus 2 points in Red Sea and 3 in Atlantic ocean
      DO jj = mj0(96),mj1(96)
         DO ji = mi0(148),mi1(148)
            zempmed_tl = zempmed_tl &
               &       - emp_tl(ji  ,jj) * tmask(ji  ,jj,1) * e1t(ji  ,jj) * e2t(ji  ,jj) &
               &       - emp_tl(ji+1,jj) * tmask(ji+1,jj,1) * e1t(ji+1,jj) * e2t(ji+1,jj)   
            zempmed    = zempmed    &
               &       - emp(ji  ,jj) * tmask(ji  ,jj,1) * e1t(ji  ,jj) * e2t(ji  ,jj)   &
               &       - emp(ji+1,jj) * tmask(ji+1,jj,1) * e1t(ji+1,jj) * e2t(ji+1,jj)   
         END DO
      END DO

      ! convert in m3
      zempmed    = zempmed    * 1.e-3
      zempmed_tl = zempmed_tl * 1.e-3

      ! Velocity profile at each point
      ! ------------------------------

      zu1_ms(:) = zu1_ms_i(:)
      zu2_ms(:) = zu2_ms_i(:)
      zu3_ms(:) = zu3_ms_i(:)
      zu1_ms_tl(:) = 0.0_wp
      zu2_ms_tl(:) = 0.0_wp
      zu3_ms_tl(:) = 0.0_wp

      ! velocity profile at 139,101  South point + emp on surface 
      DO jk = 1, 14                                      
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zu1_ms(jk)    = zu1_ms(jk)    &
                  &          + ( zempmed    / 14. ) / ( e2u(ji-1,jj-1) * fse3u(ji-1,jj-1,jk) )
               zu1_ms_tl(jk) = zu1_ms_tl(jk) &
                  &          + ( zempmed_tl / 14. ) / ( e2u(ji-1,jj-1) * fse3u(ji-1,jj-1,jk) )
            END DO
         END DO
      END DO

      ! profile at East Gibraltar    
      ! velocity profile at 141,102  + emp on surface 
      DO jk = 1,  14                                     
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zu3_ms(jk)    = zu3_ms(jk)    &
                  &          + ( zempmed    / 14. ) / ( e2u(ji,jj) * fse3u(ji,jj,jk) )
               zu3_ms_tl(jk) = zu3_ms_tl(jk) &
                  &          + ( zempmed_tl / 14. ) / ( e2u(ji,jj) * fse3u(ji,jj,jk) )
            END DO
         END DO
      END DO
      
      ! Balance of temperature and salinity
      ! -----------------------------------

      ! west gibraltar surface vertical sum of transport* S,T
      DO jk =  1, 14
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               zsumt   = zsumt   + tn(ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk)
               zsums   = zsums   + sn(ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk) 
               zsumttl = zsumttl  &
                  &    + tn_tl(ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk) &
                  &    + tn(ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms_tl(jk)
               zsumstl = zsumstl  &
                  &    + sn_tl(ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk) &
                  &    + sn(ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms_tl(jk) 
            END DO
         END DO
      END DO

      ! east Gibraltar surface  vertical sum of transport* S,T
      DO jk =  1, 14
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               zsumt1   = zsumt1   &
                  &     + tn(ji,jj,jk) * e2u(ji+1,jj+1) * fse3u(ji+1,jj+1,jk) * zu3_ms(jk)
               zsums1   = zsums1   &
                  &     + sn(ji,jj,jk) * e2u(ji+1,jj+1) * fse3u(ji+1,jj+1,jk) * zu3_ms(jk)
               zsumt1tl = zsumt1tl &
                  &     + tn_tl(ji,jj,jk) * e2u(ji+1,jj+1) * fse3u(ji+1,jj+1,jk) * zu3_ms(jk) &
                  &     + tn(ji,jj,jk) * e2u(ji+1,jj+1) * fse3u(ji+1,jj+1,jk) * zu3_ms_tl(jk)
               zsums1tl = zsums1tl &
                  &     + sn_tl(ji,jj,jk) * e2u(ji+1,jj+1) * fse3u(ji+1,jj+1,jk) * zu3_ms(jk) &
                  &     + sn(ji,jj,jk) * e2u(ji+1,jj+1) * fse3u(ji+1,jj+1,jk) * zu3_ms_tl(jk)
            END DO
         END DO
      END DO

      ! east Gibraltar deeper  vertical sum of transport* S,T
      DO jj = mj0(102), mj1(102) 
         DO ji = mi0(141), mi1(141) 
            zsumt2   = tn(   ji,jj,21) * e2u(ji-1,jj) * fse3u(ji-1,jj,21) * zu3_ms(   21)
            zsums2   = sn(   ji,jj,21) * e2u(ji-1,jj) * fse3u(ji-1,jj,21) * zu3_ms(   21)
            zsumt2tl = tn_tl(ji,jj,21) * e2u(ji-1,jj) * fse3u(ji-1,jj,21) * zu3_ms(   21) &
               &     + tn(   ji,jj,21) * e2u(ji-1,jj) * fse3u(ji-1,jj,21) * zu3_ms_tl(21)
            zsums2tl = sn_tl(ji,jj,21) * e2u(ji-1,jj) * fse3u(ji-1,jj,21) * zu3_ms(   21) &
               &     + sn(   ji,jj,21) * e2u(ji-1,jj) * fse3u(ji-1,jj,21) * zu3_ms_tl(21)
         END DO
      END DO

      ! west Gibraltar deeper vertical sum of transport* S,T
      DO  jk =  21, 22 
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               zsumt3   = zsumt3   &
                  &     + tn(   ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk)
               zsums3   = zsums3   &
                  &     + sn(   ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk)
               zsumt3tl = zsumt3tl &
                  &     + tn_tl(ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk) &
                  &     + tn(   ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms_tl(jk)
               zsums3tl = zsums3tl &
                  &     + sn_tl(ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk) &
                  &     + sn(   ji,jj,jk) * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms_tl(jk)
            END DO
         END DO
      END DO

      ! Total transport = 0.
      zsumt4   = zsumt2   + zsumt1   - zsumt   - zsumt3
      zsums4   = zsums2   + zsums1   - zsums   - zsums3
      zsumt4tl = zsumt2tl + zsumt1tl - zsumttl - zsumt3tl
      zsums4tl = zsums2tl + zsums1tl - zsumstl - zsums3tl

      ! Temperature and Salinity at West gibraltar , Level 22
      DO jj = mj0(102), mj1(102) 
         DO ji = mi0(140), mi1(140) 
            zt   = zsumt4   / ( zu2_ms(22) * e2u(ji-1,jj) * fse3u(ji-1,jj,22) )
            zs   = zsums4   / ( zu2_ms(22) * e2u(ji-1,jj) * fse3u(ji-1,jj,22) )
            zttl = zsumt4tl / ( zu2_ms(22) * e2u(ji-1,jj) * fse3u(ji-1,jj,22) ) &
               & - zsumt4   / ( 2 * zu2_ms(22) * e2u(ji-1,jj) * fse3u(ji-1,jj,22) ) * zu2_ms_tl(22)
            zstl = zsums4tl / ( zu2_ms(22) * e2u(ji-1,jj) * fse3u(ji-1,jj,22) ) &
               & - zsums4   / ( 2 * zu2_ms(22) * e2u(ji-1,jj) * fse3u(ji-1,jj,22) ) * zu2_ms_tl(22)
         END DO
      END DO
      
      ! New Temperature and Salinity trend at West Gibraltar
      ! ----------------------------------------------------

      ! south point 	
      DO jk = 1, 22
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               zvt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu = e2u(ji,jj) * fse3u(ji,jj,jk)
               ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk)                                   &
                  &            - ( 1. / zvt ) * zsu * zu1_ms_tl(jk) * tn(ji,jj,jk) &
                  &            - ( 1. / zvt ) * zsu * zu1_ms(jk) * tn_tl(ji,jj,jk)
               sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk)                                   &
                  &            - ( 1. / zvt ) * zsu * zu1_ms_tl(jk) * sn(ji,jj,jk) &
                  &            - ( 1. / zvt ) * zsu * zu1_ms(jk) * sn_tl(ji,jj,jk)
            END DO
         END DO
      END DO

      ! north point 
      DO jk = 15, 20
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(139), mi1(139) 
               zvt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu = e2u(ji-1,jj) * fse3u(ji-1,jj,jk)
               ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk)                                      &
                  &            - ( 1. / zvt ) * zsu * zu2_ms_tl(jk) * tn(ji, jj-1,jk) &
                  &            - ( 1. / zvt ) * zsu * zu2_ms(jk) * tn_tl(ji, jj-1,jk)
               sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk)                                      &
                  &            - ( 1. / zvt ) * zsu * zu2_ms_tl(jk) * sn(ji, jj-1,jk) &
                  &            - ( 1. / zvt ) * zsu * zu2_ms(jk) * sn_tl(ji, jj-1,jk)
            END DO
         END DO
      END DO

      ! Gibraltar outflow, north point deeper
      jk =  22
      DO jj = mj0(102), mj1(102) 
         DO ji = mi0(139), mi1(139) 
            zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
            zsu = e2u(ji, jj) * fse3u(ji, jj,jk)
            ta_tl(ji, jj,jk) = ta_tl(ji, jj,jk)                        &
               &             - ( 1. / zvt ) * zsu * zu2_ms_tl(jk) * zt &
               &             - ( 1. / zvt ) * zsu * zu2_ms(jk) * zttl
            sa_tl(ji, jj,jk) = sa_tl(ji, jj,jk)                        &
               &             - ( 1. / zvt ) * zsu * zu2_ms_tl(jk) * zs &
               &             - ( 1. / zvt ) * zsu * zu2_ms(jk) * zstl
         END DO
      END DO


      ! New Temperature and Salinity at East Gibraltar 
      ! ----------------------------------------------

      ! surface	
      DO jk = 1, 14
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(141), mi1(141) 
               zvt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu = e2u(ji-1,jj) * fse3u(ji-2,jj,jk)
               ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk)                                       &
                  &            + ( 1. / zvt ) * zsu * zu3_ms_tl(jk) * tn(ji-2,jj-1,jk) &
                  &            + ( 1. / zvt ) * zsu * zu3_ms(jk) * tn_tl(ji-2,jj-1,jk)
               sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk)                                       &
                  &            + ( 1. / zvt ) * zsu * zu3_ms_tl(jk) * sn(ji-2,jj-1,jk) &
                  &            + ( 1. / zvt ) * zsu * zu3_ms(jk) * sn_tl(ji-2,jj-1,jk)
            END DO
         END DO
      END DO
      ! deeper
      jk =  21
      DO jj = mj0(102), mj1(102) 
         DO ji = mi0(141), mi1(141) 
            zvt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
            zsu = e2u(ji-1,jj) * fse3u(ji-1,jj,jk)
            ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk)                                   &
               &            + ( 1. / zvt ) * zsu * zu3_ms_tl(jk) * tn(ji,jj,jk) &
               &            + ( 1. / zvt ) * zsu * zu3_ms(jk) * tn_tl(ji,jj,jk)
            sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk)                                   &
               &            + ( 1. / zvt ) * zsu * zu3_ms_tl(jk) * sn(ji,jj,jk) &
               &            + ( 1. / zvt ) * zsu * zu3_ms(jk) * sn_tl(ji,jj,jk)
         END DO
      END DO

   END SUBROUTINE tra_gibraltar_tan
   SUBROUTINE tra_gibraltar_adj
      !!---------------------------------------------------------------------
      !!               ***  ROUTINE tra_gibraltar_adj  ***
      !!
      !! ** Purpose :
      !!        Update the horizontal advective trend of tracers (t & s)
      !!        correction in Gibraltar  and
      !!        add it to the general trend of tracer equations.
      !!
      !! ** Method :
      !!      We impose transport at Gibraltar and knowing T and S in
      !!      surface and deeper at each side of the strait, we deduce T and S
      !!      of the outflow of the Mediterranean Sea in the Atlantic ocean .
      !!                                
      !!          ________________      N        ________________
      !! 102           |    |->         |           <-|    |<-
      !! 101      ___->|____|_____   W - - E     ___->|____|_____
      !!           139   140  141       |         139   140  141
      !!          horizontal view       S        horizontal view
      !!            surface                          depth
      !!         C A U T I O N : the trend saved is the centered trend only.
      !!      It doesn't take into account the upstream part of the scheme.
      !!
      !! ** history :
      !!           !  02-06  (A. Bozec) Original code 
      !!      8.5  !  02-11  (A. Bozec) F90: Free form and module
      !! ** history of the adjoint routine:
      !!           !  08-08  (A. Vidard) tangent of the 02-11 version
      !!---------------------------------------------------------------------
      !! * Local declarations
      INTEGER ::  ji, jj, jk               ! dummy loop indices
      REAL(wp) :: zsu, zvt
      REAL(wp) :: zsumt, zsumt1, zsumt2, zsumt3, zsumt4
      REAL(wp) :: zsums, zsums1, zsums2, zsums3, zsums4
      REAL(wp) :: zsumtad, zsumt1ad, zsumt2ad, zsumt3ad, zsumt4ad
      REAL(wp) :: zsumsad, zsums1ad, zsums2ad, zsums3ad, zsums4ad
      REAL(wp) :: zt, zs
      REAL(wp) :: ztad, zsad
      REAL(wp) :: zwei
      REAL(wp), DIMENSION (jpk) ::  zu1_ms, zu2_ms, zu3_ms
      REAL(wp), DIMENSION (jpk) ::  zu1_ms_ad, zu2_ms_ad, zu3_ms_ad
      !!---------------------------------------------------------------------

      ! Initialization of vertical sum for T and S transport
      ! ----------------------------------------------------

      zsumt  = 0.e0        ! West Gib. surface south point ( T )
      zsums  = 0.e0        ! West Gib. surface south point ( S )
      zsumt1 = 0.e0        ! East Gib. surface north point ( T )
      zsums1 = 0.e0        ! East Gib. surface north point ( S )
      zsumt2 = 0.e0        ! East Gib. depth   north point ( T )
      zsums2 = 0.e0        ! East Gib. depth   north point ( S )
      zsumt3 = 0.e0        ! West Gib. depth   south point ( T ) 
      zsums3 = 0.e0        ! West Gib. depth   south point ( S ) 
      zsumt4 = 0.e0        ! West Gib. depth   north point ( T ) 
      zsums4 = 0.e0        ! West Gib. depth   north point ( S ) 
      zsumtad  = 0.e0        ! West Gib. surface south point ( T )
      zsumsad  = 0.e0        ! West Gib. surface south point ( S )
      zsumt1ad = 0.e0        ! East Gib. surface north point ( T )
      zsums1ad = 0.e0        ! East Gib. surface north point ( S )
      zsumt2ad = 0.e0        ! East Gib. depth   north point ( T )
      zsums2ad = 0.e0        ! East Gib. depth   north point ( S )
      zsumt3ad = 0.e0        ! West Gib. depth   south point ( T ) 
      zsums3ad = 0.e0        ! West Gib. depth   south point ( S ) 
      zsumt4ad = 0.e0        ! West Gib. depth   north point ( T ) 
      zsums4ad = 0.e0        ! West Gib. depth   north point ( S ) 
      ztad     = 0.e0
      zsad     = 0.e0
      zu1_ms_ad(:) = 0.0_wp
      zu2_ms_ad(:) = 0.0_wp
      zu3_ms_ad(:) = 0.0_wp
      zempmed_ad = 0.e0
       
      !=====================
      ! Direct recomputation
      !=====================

      ! EMP of Mediterranean Sea 
      ! ------------------------
 
      zempmed = 0.e0
      zwei = 0.e0
      DO jj = mj0(96),mj1(110)
         DO ji = mi0(141),mi1(181)
            zwei    = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            zempmed = zempmed + emp(ji,jj) * zwei
         END DO
      END DO
      IF( lk_mpp )   CALL mpp_sum( zempmed )      ! sum with other processors value


      ! minus 2 points in Red Sea and 3 in Atlantic ocean
      DO jj = mj0(96),mj1(96)
         DO ji = mi0(148),mi1(148)
            zempmed = zempmed  -  emp(ji  ,jj) * tmask(ji  ,jj,1) * e1t(ji  ,jj) * e2t(ji  ,jj)   &
                               -  emp(ji+1,jj) * tmask(ji+1,jj,1) * e1t(ji+1,jj) * e2t(ji+1,jj)   
         END DO
      END DO

      ! convert in m3
      zempmed = zempmed * 1.e-3

      ! Velocity profile at each point
      ! ------------------------------

      zu1_ms(:) = zu1_ms_i(:)
      zu2_ms(:) = zu2_ms_i(:)
      zu3_ms(:) = zu3_ms_i(:)

      ! velocity profile at 139,101  South point + emp on surface 
      DO jk = 1, 14                      
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zu1_ms(jk) = zu1_ms(jk) + ( zempmed / 14. ) / ( e2u(ji-1, jj-1) * fse3u(ji-1, jj-1,jk) ) 
            END DO
         END DO
      END DO

      ! profile at East Gibraltar    
      ! velocity profile at 141,102  + emp on surface 
      DO  jk = 1, 14                     
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zu3_ms(jk) = zu3_ms(jk) +  ( zempmed / 14. ) / ( e2u(ji, jj) * fse3u(ji, jj,jk) ) 
            END DO
         END DO
      END DO
    
      ! Balance of temperature and salinity
      ! -----------------------------------

      ! west gibraltar surface vertical sum of transport* S,T
      DO  jk =  1, 14 
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               zsumt  = zsumt + tn(ji, jj,jk) * e2u(ji, jj) * fse3u(ji, jj,jk) * zu1_ms(jk)  
               zsums  = zsums + sn(ji, jj,jk) * e2u(ji, jj) * fse3u(ji, jj,jk) * zu1_ms(jk) 
            END DO
         END DO
      END DO

      ! east Gibraltar surface  vertical sum of transport* S,T
      DO  jk =  1, 14 
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               zsumt1 = zsumt1 + tn(ji, jj,jk) * e2u(ji+1, jj+1) * fse3u(ji+1, jj+1,jk) * zu3_ms(jk)
               zsums1 = zsums1 + sn(ji, jj,jk) * e2u(ji+1, jj+1) * fse3u(ji+1, jj+1,jk) * zu3_ms(jk)
            END DO
         END DO
      END DO

      ! east Gibraltar deeper  vertical sum of transport* S,T
      DO jj = mj0(102), mj1(102) 
         DO ji = mi0(141), mi1(141) 
            zsumt2 = tn(ji, jj,21) * e2u(ji-1, jj) * fse3u(ji-1, jj,21) * zu3_ms(21)
            zsums2 = sn(ji, jj,21) * e2u(ji-1, jj) * fse3u(ji-1, jj,21) * zu3_ms(21)
         END DO
      END DO
      
      ! west Gibraltar deeper vertical sum of transport* S,T
      DO  jk =  21, 22 
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               zsumt3 = zsumt3 + tn(ji, jj,jk) * e2u(ji, jj) * fse3u(ji, jj,jk) * zu1_ms(jk)
               zsums3 = zsums3 + sn(ji, jj,jk) * e2u(ji, jj) * fse3u(ji, jj,jk) * zu1_ms(jk)
            END DO
         END DO
      END DO

      ! Total transport = 0.
      zsumt4 = zsumt2 + zsumt1 - zsumt - zsumt3
      zsums4 = zsums2 + zsums1 - zsums - zsums3

      ! Temperature and Salinity at West gibraltar , Level 22
      DO jj = mj0(102), mj1(102) 
         DO ji = mi0(140), mi1(140) 
            zt = zsumt4 / ( zu2_ms(22) * e2u(ji-1, jj) * fse3u(ji-1, jj, 22) )
            zs = zsums4 / ( zu2_ms(22) * e2u(ji-1, jj) * fse3u(ji-1, jj, 22) )
         END DO
      END DO

      !=============
      ! Adjoint part
      !=============

      ! New Temperature and Salinity at East Gibraltar 
      ! ----------------------------------------------

      ! deeper
      jk =  21
      DO jj = mj0(102), mj1(102) 
         DO ji = mi1(141), mi0(141), -1
            zvt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
            zsu = e2u(ji-1,jj) * fse3u(ji-1,jj,jk)
            zu3_ms_ad(jk)   = zu3_ms_ad(jk)   + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * sn(ji,jj,jk)
            sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu3_ms(jk)
            zu3_ms_ad(jk)   = zu3_ms_ad(jk)   + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * tn(ji,jj,jk)
            tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu3_ms(jk)
         END DO
      END DO

      ! surface	
      DO jk = 14, 1, -1
         DO jj =  mj1(102), mj0(102), -1
            DO ji = mi1(141), mi0(141), -1
               zvt = e1t(ji,jj)   * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu = e2u(ji-1,jj) * fse3u(ji-2,jj,jk)
               zu3_ms_ad(jk)       = zu3_ms_ad(jk)       + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * sn(ji-2,jj-1,jk)
               sn_ad(ji-2,jj-1,jk) = sn_ad(ji-2,jj-1,jk) + sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu3_ms(jk) 
               zu3_ms_ad(jk)       = zu3_ms_ad(jk)       + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * tn(ji-2,jj-1,jk)
               tn_ad(ji-2,jj-1,jk) = tn_ad(ji-2,jj-1,jk) + ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu3_ms(jk) 
            END DO
         END DO
      END DO
      
      ! New Temperature and Salinity trend at West Gibraltar
      ! ----------------------------------------------------
      ! Gibraltar outflow, north point deeper
      jk =  22
      DO jj = mj0(102), mj1(102) 
         DO ji = mi0(139), mi1(139) 
            zvt = e1t(ji, jj) * e2t(ji, jj) * fse3t(ji, jj,jk)
            zsu = e2u(ji, jj) * fse3u(ji, jj,jk)
            zu2_ms_ad(jk) = zu2_ms_ad(jk) - sa_ad(ji, jj,jk) * ( 1. / zvt ) * zsu * zs
            zsad          = zsad          - sa_ad(ji, jj,jk) * ( 1. / zvt ) * zsu * zu2_ms(jk)
            zu2_ms_ad(jk) = zu2_ms_ad(jk) - ta_ad(ji, jj,jk) * ( 1. / zvt ) * zsu * zt
            ztad          = ztad          - ta_ad(ji, jj,jk) * ( 1. / zvt ) * zsu * zu2_ms(jk)
         END DO
      END DO
      ! north point 
      DO jk = 20, 15, -1
         DO jj = mj1(102), mj0(102), -1
            DO ji = mi1(139), mi0(139), -1
               zvt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu = e2u(ji-1,jj) * fse3u(ji-1,jj,jk)
               zu2_ms_ad(jk)      = zu2_ms_ad(jk)      - sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * sn(ji, jj-1,jk)
               sn_ad(ji, jj-1,jk) = sn_ad(ji, jj-1,jk) - sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu2_ms(jk)
               zu2_ms_ad(jk)      = zu2_ms_ad(jk)      - ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * tn(ji, jj-1,jk)
               tn_ad(ji, jj-1,jk) = tn_ad(ji, jj-1,jk) - ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu2_ms(jk)
            END DO
         END DO
      END DO
      ! south point 	
      DO jk = 22, 1, -1
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               zvt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)
               zsu = e2u(ji,jj) * fse3u(ji,jj,jk)
               zu1_ms_ad(jk)   = zu1_ms_ad(jk)   - sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * sn(ji,jj,jk)
               sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) - sa_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu1_ms(jk)
               zu1_ms_ad(jk)   = zu1_ms_ad(jk)   - ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * tn(ji,jj,jk)
               tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) - ta_ad(ji,jj,jk) * ( 1. / zvt ) * zsu * zu1_ms(jk)
            END DO
         END DO
      END DO

      
      ! Balance of temperature and salinity
      ! -----------------------------------

      ! Temperature and Salinity at West gibraltar , Level 22
      DO jj = mj0(102), mj1(102) 
         DO ji = mi1(140), mi0(140), -1
            zsums4ad      = zsums4ad      &
               &          + zsad  / ( zu2_ms(22) * e2u(ji-1,jj) * fse3u(ji-1,jj,22) )
            zu2_ms_ad(22) = zu2_ms_ad(22) &
               &          - zsad * zsums4 / ( 2 * zu2_ms(22) * e2u(ji-1,jj) * fse3u(ji-1,jj,22) )
            zsumt4ad      = zsumt4ad      &
               &          + ztad  / ( zu2_ms(22) * e2u(ji-1,jj) * fse3u(ji-1,jj,22) )
            zu2_ms_ad(22) = zu2_ms_ad(22) &
               &          - ztad * zsumt4 / ( 2 * zu2_ms(22) * e2u(ji-1,jj) * fse3u(ji-1,jj,22) )
            ztad = 0.0_wp
            zsad = 0.0_wp
         END DO
      END DO

      ! Total transport = 0.
      zsums2ad = zsums2ad + zsums4ad
      zsums1ad = zsums1ad + zsums4ad
      zsumsad  = zsumsad  - zsums4ad
      zsums3ad = zsums3ad - zsums4ad
      zsumt2ad = zsumt2ad + zsumt4ad
      zsumt1ad = zsumt1ad + zsumt4ad
      zsumtad  = zsumtad  - zsumt4ad
      zsumt3ad = zsumt3ad - zsumt4ad
      zsumt4ad = 0.0
      zsums4ad = 0.0
      
      ! west Gibraltar deeper vertical sum of transport* S,T
      DO  jk =  22, 21, -1 
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) + zsums3ad * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk)
               zu1_ms_ad(jk)   = zu1_ms_ad(jk)   + zsums3ad * e2u(ji,jj) * fse3u(ji,jj,jk) * sn(ji,jj,jk)
               tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) + zsumt3ad * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk)
               zu1_ms_ad(jk)   = zu1_ms_ad(jk)   + zsumt3ad * e2u(ji,jj) * fse3u(ji,jj,jk) * tn(ji,jj,jk)
            END DO
         END DO
      END DO

      ! east Gibraltar deeper  vertical sum of transport* S,T
      DO jj = mj0(102), mj1(102) 
         DO ji = mi1(141), mi0(141), -1
            sn_ad(ji,jj,21) = sn_ad(ji,jj,21) + zsums2ad * e2u(ji-1,jj) * fse3u(ji-1,jj,21) * zu3_ms(21) 
            zu3_ms_ad(21)   =  zu3_ms_ad(21)  + zsums2ad * e2u(ji-1,jj) * fse3u(ji-1,jj,21) * sn(ji,jj,21)
            tn_ad(ji,jj,21) = tn_ad(ji,jj,21) + zsumt2ad * e2u(ji-1,jj) * fse3u(ji-1,jj,21) * zu3_ms(21) 
            zu3_ms_ad(21)   =  zu3_ms_ad(21)  + zsumt2ad * e2u(ji-1,jj) * fse3u(ji-1,jj,21) * tn(ji,jj,21)
            zsumt2ad = 0.0_wp
            zsums2ad = 0.0_wp
         END DO
      END DO

      ! east Gibraltar surface  vertical sum of transport* S,T
      DO jk =  14, 1, -1
         DO jj = mj1(101), mj0(101), -1
            DO ji = mi1(139), mi0(139), - 1
               sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) + zsums1ad * e2u(ji+1,jj+1) * fse3u(ji+1,jj+1,jk) * zu3_ms(jk) 
               zu3_ms_ad(jk)   = zu3_ms_ad(jk)   + zsums1ad * e2u(ji+1,jj+1) * fse3u(ji+1,jj+1,jk) * sn(ji,jj,jk)
               tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) + zsumt1ad * e2u(ji+1,jj+1) * fse3u(ji+1,jj+1,jk) * zu3_ms(jk) 
               zu3_ms_ad(jk)   = zu3_ms_ad(jk)   + zsumt1ad * e2u(ji+1,jj+1) * fse3u(ji+1,jj+1,jk) * tn(ji,jj,jk)
            END DO
         END DO
      END DO

      ! west gibraltar surface vertical sum of transport* S,T
      DO jk =  14, 1, -1
         DO jj = mj0(101), mj1(101) 
            DO ji = mi0(139), mi1(139) 
               sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) + zsumsad * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk)
               zu1_ms_ad(jk)   = zu1_ms_ad(jk)   + zsumsad * e2u(ji,jj) * fse3u(ji,jj,jk) * sn(ji,jj,jk)
               tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) + zsumtad * e2u(ji,jj) * fse3u(ji,jj,jk) * zu1_ms(jk)
               zu1_ms_ad(jk)   = zu1_ms_ad(jk)   + zsumtad * e2u(ji,jj) * fse3u(ji,jj,jk) * tn(ji,jj,jk)
            END DO
         END DO
      END DO
      ! Velocity profile at each point
      ! ------------------------------
      ! profile at East Gibraltar    
      ! velocity profile at 141,102  + emp on surface 
      DO jk = 14, 1, -1                                     
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zempmed_ad = zempmed_ad + zu3_ms_ad(jk) / ( 14. * e2u(ji,jj) * fse3u(ji,jj,jk) )
            END DO
         END DO
      END DO

      ! velocity profile at 139,101  South point + emp on surface 
      DO jk = 14, 1, -1                            
         DO jj = mj1(102), mj0(102), -1
            DO ji = mi1(140), mi0(140), -1
               zempmed_ad = zempmed_ad + zu1_ms_ad(jk) / ( 14. * e2u(ji-1,jj-1) * fse3u(ji-1,jj-1,jk) )
            END DO
         END DO
      END DO

      zu1_ms_ad(:) = 0.0_wp
      zu2_ms_ad(:) = 0.0_wp
      zu3_ms_ad(:) = 0.0_wp

      ! EMP of Mediterranean Sea 
      ! ------------------------
      
      ! convert in m3
      zempmed_ad = zempmed_ad * 1.e-3

     ! minus 2 points in Red Sea and 3 in Atlantic ocean
      DO jj = mj1(96), mj0(96), -1
         DO ji = mi1(148), mi0(148), -1
            emp_ad(ji  ,jj) = emp_ad(ji  ,jj) - zempmed_ad * tmask(ji  ,jj,1) * e1t(ji  ,jj) * e2t(ji  ,jj)
            emp_ad(ji+1,jj) = emp_ad(ji+1,jj) - zempmed_ad * tmask(ji+1,jj,1) * e1t(ji+1,jj) * e2t(ji+1,jj)
         END DO
      END DO

      IF( lk_mpp )   CALL mpp_sum( zempmed_ad )      ! sum with other processors value

      DO jj = mj0(96), mj1(110)
         DO ji = mi0(141), mi1(181)
            zwei    = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            emp_ad(ji,jj) = emp_ad(ji,jj) + zempmed_ad * zwei
         END DO
      END DO
      zempmed_ad = 0.e0

 
   END SUBROUTINE tra_gibraltar_adj
   SUBROUTINE tra_hormuz_tan
      !!---------------------------------------------------------------------
      !!               ***  ROUTINE tra_hormuz_tan  ***
      !!
      !! ** Purpose of the direct routine:
      !!        Update the horizontal advective trend of tracers
      !!        correction in Hormuz.
      !!
      !! ** Method of the direct routine: We impose transport at Hormuz .
      !!                                
      !! ** history of the direct routine:
      !!           !  02-11  (A. Bozec) Original code 
      !!      8.5  !  02-11  (A. Bozec) F90: Free form and module
      !! ** history of the tangent routine:
      !!           !  08-08  (A. Vidard) tangent of the 02-11 version
      !!---------------------------------------------------------------------
      !! * Local declarations
      !!---------------------------------------------------------------------
      
      !! ... nothing

   END SUBROUTINE tra_hormuz_tan
   SUBROUTINE tra_hormuz_adj
      !!---------------------------------------------------------------------
      !!               ***  ROUTINE tra_hormuz_adj  ***
      !!
      !! ** Purpose of the direct routine:
      !!        Update the horizontal advective trend of tracers
      !!        correction in Hormuz.
      !!
      !! ** Method of the direct routine: We impose transport at Hormuz .
      !!                                
      !! ** history of the direct routine:
      !!           !  02-11  (A. Bozec) Original code 
      !!      8.5  !  02-11  (A. Bozec) F90: Free form and module
      !! ** history of the tangent routine:
      !!           !  08-08  (A. Vidard) adjoint of the 02-11 version
      !!---------------------------------------------------------------------
      !! * Local declarations
      !!---------------------------------------------------------------------
      
      !! ... nothing

   END SUBROUTINE tra_hormuz_adj
   SUBROUTINE tra_cla_init_tam
      !!---------------------------------------------------------------------
      !!               ***  ROUTINE tra_cla_init_tam  ***
      !!
      !! ** Purpose :   Initialization of variables
      !!
      !! ** history :
      !!           !  02-11  (A. Bozec) Original code 
      !!      8.5  !  02-11  (A. Bozec) F90: Free form and module
      !! ** history of the tam version:
      !!      9.0  !  08-08  (A. Vidard) Original code
      !!---------------------------------------------------------------------
      !! * Local declarations
      INTEGER ::  ji, jj, jk              ! dummy loop indices
      !!---------------------------------------------------------------------

      ! Control print
      ! -------------
      IF (lfirst) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_cla_init_tam : cross land advection on tracer '
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~~'

         ! Initialization at Bab el Mandeb
         ! -------------------------------

         ! imposed transport
         zisw_rs = 0.4e6        ! inflow surface water
         zurw_rs = 0.2e6        ! upper recirculation water
         !!Alex      zbrw_rs = 1.2e6        ! bottom  recirculation water
         zbrw_rs = 0.5e6        ! bottom  recirculation water

         ! initialization of the velocity at Bab el Mandeb
         zu1_rs_i(:) = 0.e0      ! velocity profile at 161,88 South point
         zu2_rs_i(:) = 0.e0      ! velocity profile at 161,87 North point
         zu3_rs_i(:) = 0.e0      ! velocity profile at 160,88 East  point

         ! velocity profile at 161,88 East Bab el Mandeb North point 
         ! we imposed zisw_rs + EMP above the Red Sea 
         DO jk = 1, 8                                      
            DO jj = mj0(88), mj1(88) 
               DO ji = mi0(160), mi1(160) 
                  zu1_rs_i(jk) = -( zisw_rs / 8. ) / ( e2u(ji,jj) * fse3u(ji,jj,jk) )     
               END DO
            END DO
         END DO

         ! recirculation water 
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160) 
               zu1_rs_i(20) = -(           zurw_rs ) / ( e2u(ji,jj) * fse3u(ji,jj,20) )
               zu1_rs_i(21) = -( zbrw_rs - zurw_rs ) / ( e2u(ji,jj) * fse3u(ji,jj,21) )
            END DO
         END DO

         ! velocity profile at 161,87 East Bab el Mandeb South point
         DO jj = mj0(87), mj1(87) 
            DO ji = mi0(160), mi1(160) 
               zu2_rs_i(21) =  ( zbrw_rs + zisw_rs ) / ( e2u(ji,jj) * fse3u(ji,jj,21) )
            END DO
         END DO

         ! velocity profile at 161, 88 West Bab el Mandeb 
         ! we imposed zisw_rs + EMP above the Red Sea 
         DO jk = 1,  10                                     
            DO jj = mj0(88), mj1(88) 
               DO ji = mi0(160), mi1(160) 
                  zu3_rs_i(jk) =  ( zisw_rs / 10. ) / ( e1v(ji,jj) * fse3v(ji,jj,jk) )
               END DO
            END DO
         END DO

         ! deeper
         DO jj = mj0(88), mj1(88) 
            DO ji = mi0(160), mi1(160) 
               zu3_rs_i(16)  = - zisw_rs /( e1v(ji,jj) * fse3v(ji,jj,16) )
            END DO
         END DO


         ! Initialization at Gibraltar
         ! ---------------------------

         ! imposed transport
         zisw_ms = 0.8e6         ! atlantic-mediterranean  water
         zmrw_ms = 0.7e6         ! middle recirculation water
         zurw_ms = 2.5e6         ! upper  recirculation water 
         zbrw_ms = 3.5e6         ! bottom recirculation water 

         ! initialization of the velocity
         zu1_ms_i(:) = 0.e0       ! velocity profile at 139,101 South point
         zu2_ms_i(:) = 0.e0       ! velocity profile at 139,102 North point
         zu3_ms_i(:) = 0.e0       ! velocity profile at 141,102 East  point

         ! velocity profile at 139,101  South point
         DO jk = 1, 14                      
            DO jj = mj0(102), mj1(102) 
               DO ji = mi0(140), mi1(140) 
                  zu1_ms_i(jk) = ( zisw_ms / 14. ) / ( e2u(ji-1, jj-1) * fse3u(ji-1, jj-1,jk)) 
               END DO
            END DO
         END DO

         ! middle recirculation ( uncounting in the balance )
         DO jk = 15, 20                      
            DO jj = mj0(102), mj1(102) 
               DO ji = mi0(140), mi1(140) 
                  zu1_ms_i(jk) = ( zmrw_ms / 6. ) / ( e2u(ji-1, jj-1) * fse3u(ji-1, jj-1,jk) ) 
               END DO
            END DO
         END DO

         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zu1_ms_i(21) =  (           zurw_ms ) / ( e2u(ji-1, jj-1) * fse3u(ji-1, jj-1,21) )
               zu1_ms_i(22) =  ( zbrw_ms - zurw_ms ) / ( e2u(ji-1, jj-1) * fse3u(ji-1, jj-1,22) )
            END DO
         END DO

         ! velocity profile at 139,102  North point
         ! middle recirculation ( uncounting in the balance )
         DO jk = 15, 20                      
            DO jj = mj0(102), mj1(102) 
               DO ji = mi0(140), mi1(140) 
                  zu2_ms_i(jk) = -( zmrw_ms / 6. ) / ( e2u(ji-1, jj) * fse3u(ji-1, jj,jk) ) 
               END DO
            END DO
         END DO

         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zu2_ms_i(22) = -( zisw_ms + zbrw_ms ) / ( e2u(ji-1, jj) * fse3u(ji-1, jj,22) )
            END DO
         END DO

         ! profile at East Gibraltar   
         ! velocity profile at 141,102 
         DO  jk = 1, 14                     
            DO jj = mj0(102), mj1(102) 
               DO ji = mi0(140), mi1(140) 
                  zu3_ms_i(jk) =  ( zisw_ms / 14. ) / ( e2u(ji, jj) * fse3u(ji, jj,jk) ) 
               END DO
            END DO
         END DO

         ! deeper 
         DO jj = mj0(102), mj1(102) 
            DO ji = mi0(140), mi1(140) 
               zu3_ms_i(21) = -zisw_ms / ( e2u(ji, jj) * fse3u(ji, jj,21) )
            END DO
         END DO
         lfirst = .FALSE.
      END IF


   END SUBROUTINE tra_cla_init_tam
   SUBROUTINE tra_cla_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE dyn_adv_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !!                   
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit

      INTEGER :: &        ! dummy loop indices
         & ji,   &
         & jj,   &        
         & jk,   &        
         & jt,   &        
         & jii,  &        
         & jis,  &        
         & jji,  &        
         & jjs,  &
         & jpert
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator

      !! * Local declarations
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &      
         & ztn_tlin,               &  ! potential temperature
         & zsn_tlin,               &  ! salinity
         & zta_tlin,               &  ! potential temperature
         & zsa_tlin,               &  ! salinity
         & zta_tlout,              &  ! potential temperature
         & zsa_tlout,              &  ! salinity
         & ztn_adout,              &  ! potential temperature
         & zsn_adout,              &  ! salinity
         & zta_adout,              &  ! potential temperature
         & zsa_adout,              &  ! salinity
         & zta_adin,               &  ! potential temperature
         & zsa_adin,               &  ! salinity
         & z3r                        ! 3D random field 
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &     
         & zemp_tlin,              &  ! evaporation minus precipitation
         & zemp_adout,             &  ! evaporation minus precipitation
         & z2r                        ! 2D random field 

      REAL(KIND=wp) ::       &
         & zsp1,             & ! scalar product involving the tangent routine
         & zsp1_1,           & ! scalar product involving the tangent routine
         & zsp1_2,           & ! scalar product involving the tangent routine
         & zsp2,             & ! scalar product involving the adjoint routine
         & zsp2_1,           & ! scalar product involving the adjoint routine
         & zsp2_2,           & ! scalar product involving the adjoint routine
         & zsp2_3,           & ! scalar product involving the adjoint routine
         & zsp2_4,           & ! scalar product involving the adjoint routine
         & zsp2_5,           & ! scalar product involving the adjoint routine
         & z2dt,             & ! temporary scalars
         & zraur
      CHARACTER(LEN=14) :: cl_name

      ALLOCATE( & 
         & ztn_tlin(   jpi, jpj, jpk ),  &
         & zsn_tlin(   jpi, jpj, jpk ),  &
         & zta_tlin(   jpi, jpj, jpk ),  &
         & zsa_tlin(   jpi, jpj, jpk ),  &
         & zta_tlout(  jpi, jpj, jpk ),  &
         & zsa_tlout(  jpi, jpj, jpk ),  &
         & ztn_adout(  jpi, jpj, jpk ),  &
         & zsn_adout(  jpi, jpj, jpk ),  &
         & zta_adout(  jpi, jpj, jpk ),  &
         & zsa_adout(  jpi, jpj, jpk ),  &
         & zta_adin(   jpi, jpj, jpk ),  &  
         & zsa_adin(   jpi, jpj, jpk ),  &  
         & z3r(        jpi, jpj, jpk ),  &
         & zemp_tlin(  jpi, jpj      ),  &
         & zemp_adout( jpi, jpj      ),  &
         & z2r(        jpi, jpj      )  &
         &  )

      ! Initialize constants

      z2dt  = 2.0_wp * rdt       ! time step: leap-frog
      zraur = 1.0_wp / rauw      ! inverse density of pure water (m3/kg)

      !=============================================================
      ! 1) dx = ( T ) and dy = ( T )
      !=============================================================

      DO jt = 2, 1, -1
      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
         ztn_tlin(  :,:,:) = 0.0_wp  
         zsn_tlin(  :,:,:) = 0.0_wp  
         zta_tlin(  :,:,:) = 0.0_wp  
         zsa_tlin(  :,:,:) = 0.0_wp  
         zta_tlout( :,:,:) = 0.0_wp  
         zsa_tlout( :,:,:) = 0.0_wp  
         ztn_adout( :,:,:) = 0.0_wp  
         zsn_adout( :,:,:) = 0.0_wp  
         zta_adout( :,:,:) = 0.0_wp  
         zsa_adout( :,:,:) = 0.0_wp  
         zta_adin(  :,:,:) = 0.0_wp    
         zsa_adin(  :,:,:) = 0.0_wp    
         zemp_tlin( :,:  ) = 0.0_wp  
         zemp_adout(:,:  ) = 0.0_wp   
         
         SELECT CASE (jt)
         CASE(1) ! Bab el Madeb
            jji = mj0(86)
            jjs = mj1(97)
            jii = mi0(147)
            jis = mi1(162)
         CASE(2) ! Gibraltar
            jji = mj0(95)
            jjs = mj1(111)
            jii = mi0(138)
            jis = mi1(182)
         END SELECT
         !--------------------------------------------------------------------
         ! Initialize the tangent input with random noise: dx
         !--------------------------------------------------------------------
         DO jj = 1, jpj
            DO ji = 1, jpi
               iseed_2d(ji,jj) = - ( 456953 + &
                  &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
            END DO
         END DO
         CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stdt )
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  ztn_tlin(ji,jj,jk) = z3r(ji,jj,jk)  
               END DO
            END DO
         END DO 
         DO jj = 1, jpj
            DO ji = 1, jpi
               iseed_2d(ji,jj) = - ( 395703 + &
                  &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
            END DO
         END DO
         CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stds )
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsn_tlin(ji,jj,jk) = z3r(ji,jj,jk)  
               END DO
            END DO
         END DO
         
         DO jj = 1, jpj
            DO ji = 1, jpi
               iseed_2d(ji,jj) = - ( 536782 + &
                  &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
            END DO
         END DO
         CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stdt )
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zta_tlin(ji,jj,jk) = z3r(ji,jj,jk)  
               END DO
            END DO
         END DO

         DO jj = 1, jpj
            DO ji = 1, jpi
               iseed_2d(ji,jj) = - ( 613925 + &
                  &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
            END DO
         END DO
         CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stds )
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsa_tlin(ji,jj,jk) = z3r(ji,jj,jk)  
               END DO
            END DO
         END DO
         
         DO jj = 1, jpj
            DO ji = 1, jpi
               iseed_2d(ji,jj) = - ( 228401 + &
                  &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
            END DO
         END DO
         CALL grid_random( iseed_2d, z2r, 'T', 0.0_wp, stdemp )
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zemp_tlin(ji,jj) = z2r(ji,jj)  
            END DO
         END DO

         zemp_tlin(:,:) = zemp_tlin(:,:) / ( z2dt * zraur )

         tn_ad(:,:,:) = 0.0_wp
         sn_ad(:,:,:) = 0.0_wp
         ta_ad(:,:,:) = 0.0_wp
         sa_ad(:,:,:) = 0.0_wp
         emp_ad(:,:)  = 0.0_wp
         tn_tl(:,:,:) = 0.0_wp
         sn_tl(:,:,:) = 0.0_wp
         ta_tl(:,:,:) = 0.0_wp
         sa_tl(:,:,:) = 0.0_wp
         emp_tl(:,:)  = 0.0_wp


         DO jk = 1, jpk
            DO jj = jji, jjs
               DO ji = jii, jis
                  tn_tl(ji,jj,jk)  = ztn_tlin(ji,jj,jk)
                  sn_tl(ji,jj,jk)  = zsn_tlin(ji,jj,jk)
                  ta_tl(ji,jj,jk)  = zta_tlin(ji,jj,jk)
                  sa_tl(ji,jj,jk)  = zsa_tlin(ji,jj,jk)
                  emp_tl(ji,jj)    = zemp_tlin(ji,jj)
               END DO
            END DO
         END DO
         CALL tra_cla_tan ( nit000 ) 

         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zta_tlout(ji,jj,jk)  = ta_tl(ji,jj,jk)
                  zsa_tlout(ji,jj,jk)  = sa_tl(ji,jj,jk)
                  zta_adin(ji,jj,jk)   = zta_tlout(ji,jj,jk) &
                     &                 * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)&
                     &                 * tmask(ji,jj,jk) * wesp_t(jk)
                  zsa_adin(ji,jj,jk)   = zsa_tlout(ji,jj,jk) &
                     &                 * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk)&
                     &                 * tmask(ji,jj,jk) * wesp_s(jk)
               END DO
            END DO
         END DO

         !--------------------------------------------------------------------
         ! Compute the scalar product: ( L dx )^T W dy
         !--------------------------------------------------------------------

         zsp1_1 = DOT_PRODUCT( zta_tlout, zta_adin )
         zsp1_2 = DOT_PRODUCT( zsa_tlout, zsa_adin )
         zsp1 = zsp1_1 + zsp1_2

         !--------------------------------------------------------------------
         ! Call the adjoint routine: dx^* = L^T dy^*
         !--------------------------------------------------------------------
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  sa_ad(ji,jj,jk) = zsa_adin(ji,jj,jk)
                  ta_ad(ji,jj,jk) = zta_adin(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL tra_cla_adj ( nit000 ) 
         DO jk = 1, jpk
            DO jj = jji, jjs    ! tlin should be 0 outside these boundaries but is not by construction
               DO ji = jii, jis ! here it would insure that the dot product does not account for it
                  ztn_adout(ji,jj,jk) = tn_ad(ji,jj,jk)
                  zsn_adout(ji,jj,jk) = sn_ad(ji,jj,jk)
                  zta_adout(ji,jj,jk) = ta_ad(ji,jj,jk)
                  zsa_adout(ji,jj,jk) = sa_ad(ji,jj,jk)
                  zemp_adout(ji,jj)  = emp_ad(ji,jj)
               END DO
            END DO
         END DO

         zsp2_1 = DOT_PRODUCT( ztn_tlin , ztn_adout )
         zsp2_2 = DOT_PRODUCT( zsn_tlin , zsn_adout )
         zsp2_3 = DOT_PRODUCT( zta_tlin , zta_adout )
         zsp2_4 = DOT_PRODUCT( zsa_tlin , zsa_adout )
         zsp2_5 = DOT_PRODUCT( zemp_tlin, zemp_adout )
         zsp2   = zsp2_1 + zsp2_2 + zsp2_3 + zsp2_4 + zsp2_5

         ! Compare the scalar products

         ! 14 char:'12345678901234'
         SELECT CASE (jt)
         CASE(1) ! Bab el Madeb
            cl_name = 'tra_cla_adj BM'
         CASE(2) ! Gibraltar
            cl_name = 'tra_cla_adj Gi'
         END SELECT
         CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      END DO
      ! Deallocate memory

      DEALLOCATE(      &
         & ztn_tlin  ,  &
         & zsn_tlin  ,  &
         & zta_tlin  ,  &
         & zsa_tlin  ,  &
         & zta_tlout ,  &
         & zsa_tlout ,  &
         & ztn_adout ,  &
         & zsn_adout ,  &
         & zta_adout ,  &
         & zsa_adout ,  &
         & zta_adin  ,  &  
         & zsa_adin  ,  &  
         & zemp_tlin ,  &
         & zemp_adout   &
         &  )

   END SUBROUTINE tra_cla_adj_tst
   !!======================================================================
#  else
   !!----------------------------------------------------------------------
   !!   Default option                              NO cross land advection
   !!----------------------------------------------------------------------
   USE in_out_manager  ! I/O manager
CONTAINS
   SUBROUTINE tra_cla_tan( kt ) 
      INTEGER, INTENT(in) ::   kt    ! ocean time-step indice
      IF( kt == nit000 .AND. lwp ) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'tra_cla_tan : No use of cross land advection'
         WRITE(numout,*) '~~~~~~~~~~~'
      ENDIF
   END SUBROUTINE tra_cla_tan
   SUBROUTINE tra_cla_adj( kt ) 
      INTEGER, INTENT(in) ::   kt    ! ocean time-step indice
      IF( kt == nit000 .AND. lwp ) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'tra_cla_adj : No use of cross land advection'
         WRITE(numout,*) '~~~~~~~~~~~'
      ENDIF
   END SUBROUTINE tra_cla_adj
   SUBROUTINE tra_cla_adj_tst( kt ) 
      INTEGER, INTENT(in) ::   kt    ! ocean time-step indice
      IF( kt == nit000 .AND. lwp ) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'tra_cla_adj_tst : No use of cross land advection'
         WRITE(numout,*) '~~~~~~~~~~~'
      ENDIF
   END SUBROUTINE tra_cla_adj_tst
#  endif
#endif
END MODULE cla_tam
