#if defined key_mpp_mpi
#if defined key_sp
#define mpivar mpi_real
#else
#define mpivar mpi_double_precision
#endif
#endif
MODULE lib_mpp_tam
   !!======================================================================
   !!                       ***  MODULE  lib_mpp_tam  ***
   !! Ocean numerics:  adjoints of massively parallel processing library
   !!=====================================================================
#if   defined key_mpp_mpi   ||   defined key_mpp_shmem
   !!----------------------------------------------------------------------
   !!   'key_mpp_mpi'     OR      MPI massively parallel processing library
   !!   'key_mpp_shmem'         SHMEM massively parallel processing library
   !!----------------------------------------------------------------------
   !!   mpp_lnkadj  : generic interface (defined in lbclnkadj) for :
   !!                 mpp_lnkadj_2d, mpp_lnkadj_3d
   !!   mpp_lnkadj_3d_gather :  Message passing manadgement for two 3D arrays
   !!   mpp_lnkadj_e   : interface defined in lbclnkadj
   !!----------------------------------------------------------------------
   !! History :
   !!        !  07-07  (K. Mogensen) Original code (lib_mppadj
   !!        !  09-02  (A. Vidard) nemo v3 update
   !! * Modules used
   USE dom_oce                    ! ocean space and time domain 
   USE in_out_manager             ! I/O manager
   USE lib_mpp                    ! Direct MPP library

   IMPLICIT NONE

   PRIVATE
   PUBLIC  mpp_lnk_3d_adj, mpp_lnk_2d_adj, mpp_lnk_3d_gather_adj, mpp_lnk_2d_e_adj
#if defined key_ecmwf_dynmem
   PUBLIC lib_mpp_alloc_adj
#endif
#if defined key_mpp_mpi
   !! ========================= !!
   !!  MPI  variable definition !!
   !! ========================= !!
!$AGRIF_DO_NOT_TREAT
#  include <mpif.h>
!$AGRIF_END_DO_NOT_TREAT
#endif

   INTERFACE mpp_lbc_north_adj
      MODULE PROCEDURE mpp_lbc_north_3d_adj, mpp_lbc_north_2d_adj 
   END INTERFACE

#ifdef key_ecmwf_dynmem
   REAL(wp), ALLOCATABLE, DIMENSION(:,:,:,:) ::   &
       t3ns_ad, t3sn_ad  ! 3d message passing arrays north-south & south-north
   REAL(wp), ALLOCATABLE, DIMENSION(:,:,:,:) ::   &
       t3ew_ad, t3we_ad  ! 3d message passing arrays east-west & west-east
   REAL(wp), ALLOCATABLE, DIMENSION(:,:,:) ::   &
       t2ns_ad, t2sn_ad  ! 2d message passing arrays north-south & south-north
   REAL(wp), ALLOCATABLE, DIMENSION(:,:,:) ::   &
       t2ew_ad, t2we_ad  ! 2d message passing arrays east-west & west-east
   REAL(wp), ALLOCATABLE, DIMENSION(:,:,:) ::   &
       tr2ns_ad, tr2sn_ad  ! 2d message passing arrays north-south & south-north including extra outer halo
   REAL(wp), ALLOCATABLE, DIMENSION(:,:,:) ::   &
       tr2ew_ad, tr2we_ad  ! 2d message passing arrays east-west & west-east including extra outer halo
#else
!!   REAL(wp), DIMENSION(jpi,jprecj,jpk,2,2) ::   &
!!       t4ns, t4sn  ! 3d message passing arrays north-south & south-north
!!   REAL(wp), DIMENSION(jpj,jpreci,jpk,2,2) ::   &
!!       t4ew, t4we  ! 3d message passing arrays east-west & west-east
!!   REAL(wp), DIMENSION(jpi,jprecj,jpk,2,2) ::   &
!!       t4p1, t4p2  ! 3d message passing arrays north fold
   REAL(wp), DIMENSION(jpi,jprecj,jpk,2) ::   &
       t3ns_ad, t3sn_ad  ! 3d message passing arrays north-south & south-north
   REAL(wp), DIMENSION(jpj,jpreci,jpk,2) ::   &
       t3ew_ad, t3we_ad  ! 3d message passing arrays east-west & west-east
!!   REAL(wp), DIMENSION(jpi,jprecj,jpk,2) ::   &
!!       t3p1, t3p2  ! 3d message passing arrays north fold
   REAL(wp), DIMENSION(jpi,jprecj,2) ::   &
       t2ns_ad, t2sn_ad  ! 2d message passing arrays north-south & south-north
   REAL(wp), DIMENSION(jpj,jpreci,2) ::   &
       t2ew_ad, t2we_ad  ! 2d message passing arrays east-west & west-east
!!   REAL(wp), DIMENSION(jpi,jprecj,2) ::   &
!!       t2p1, t2p2  ! 2d message passing arrays north fold
   REAL(wp), DIMENSION(1-jpr2di:jpi+jpr2di,jprecj+jpr2dj,2) ::   &
       tr2ns_ad, tr2sn_ad  ! 2d message passing arrays north-south & south-north including extra outer halo
   REAL(wp), DIMENSION(1-jpr2dj:jpj+jpr2dj,jpreci+jpr2di,2) ::   &
       tr2ew_ad, tr2we_ad  ! 2d message passing arrays east-west & west-east including extra outer halo
#endif

CONTAINS

   SUBROUTINE mpp_lnk_3d_adj( ptab_ad, cd_type, psgn, cd_mpp, pval )
      !!-----------------------------------------------------------------------
      !!
      !!      ***  ROUTINE mpp_lnk_3d_adj : ADJOINT OF ROUTINE mpp_lnkadj_adj  ***
      !!
      !! ** Purpose of direct routine   : Message passing manadgement
      !!
      !! ** Method of direct routine    : Use mppsend and mpprecv function for
      !!            passing mask between processors following neighboring 
      !!            subdomains.
      !!            domain parameters
      !!                    nlci   : first dimension of the local subdomain
      !!                    nlcj   : second dimension of the local subdomain
      !!                    nbondi : mark for "east-west local boundary"
      !!                    nbondj : mark for "north-south local boundary"
      !!                    noea   : number for local neighboring processors 
      !!                    nowe   : number for local neighboring processors
      !!                    noso   : number for local neighboring processors
      !!                    nono   : number for local neighboring processors
      !!
      !! ** Comments on Adjoint Routine :
      !!
      !! ** Action  : ptab_ad with update value at its periphery
      !!                   
      !! References : 
      !!
      !! History :
      !!        ! 07-07 (K. Mogensen) Initial version
      !!        ! 09-02 (A. Vidard) NEMO v3 update 
      !!-----------------------------------------------------------------------
      !! * Modules used
      !! * Arguments
      CHARACTER(len=1) , INTENT( in ) ::   &
         cd_type       ! define the nature of ptab_ad array grid-points
         !             ! = T , U , V , F , W points
         !             ! = S : T-point, north fold treatment ???
         !             ! = G : F-point, north fold treatment ???
      REAL(wp), INTENT( in ) ::   &
         psgn          ! control of the sign change
         !             !   = -1. , the sign is changed if north fold boundary
         !             !   =  1. , the sign is kept  if north fold boundary
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( inout ) ::   &
         ptab_ad          ! 3D array on which the boundary condition is applied
      CHARACTER(len=3), INTENT( in ), OPTIONAL ::    &
         cd_mpp        ! fill the overlap area only 
      REAL(wp)        , INTENT(in   ), OPTIONAL           ::   pval      ! background value (used at closed boundaries)
                                                                         ! only here for compatibility
      !! * Local declarations
      INTEGER ::   ji, jj, jk, jl                        ! dummy loop indices
      INTEGER ::   imigr, iihom, ijhom, iloc, ijt, iju   ! temporary integers
      INTEGER ::   ml_req1, ml_req2, ml_err     ! for key_mpi_isend
      INTEGER ::   ml_stat(MPI_STATUS_SIZE)     ! for key_mpi_isend
      !!----------------------------------------------------------------------

      ! 5. East and west directions exchange
      ! ------------------------------------

      SELECT CASE ( npolj )

      CASE ( 3, 4, 5, 6 )

         ! 5.3 Write Dirichlet lateral conditions

         iihom = nlci-jpreci

         SELECT CASE ( nbondi)
         CASE ( -1 )
            DO jl = 1, jpreci
               t3we_ad(:,jl,:,1)     = ptab_ad(iihom+jl,:,:)
               ptab_ad(iihom+jl,:,:) = 0.0_wp
            END DO
         CASE ( 0 ) 
            DO jl = 1, jpreci
               t3we_ad(:,jl,:,1)     = ptab_ad(iihom+jl,:,:)
               ptab_ad(iihom+jl,:,:) = 0.0_wp
               t3ew_ad(:,jl,:,1)     = ptab_ad(jl      ,:,:)
               ptab_ad(jl      ,:,:) = 0.0_wp
            END DO
         CASE ( 1 )
            DO jl = 1, jpreci
               t3ew_ad(:,jl,:,1)     = ptab_ad(jl      ,:,:)
               ptab_ad(jl      ,:,:) = 0.0_wp
            END DO
         END SELECT

         ! 5.2 Migrations

#if defined key_mpp_shmem
error "key_mpp_shmem not support in nemovar"
#elif defined key_mpp_mpi
         !! MPI version

         imigr=jpreci*jpj*jpk
  
         SELECT CASE ( nbondi )
         CASE ( -1 )
            CALL mppsend( 2, t3we_ad(1,1,1,1), imigr, noea, ml_req1 )
            CALL mpprecv( 1, t3ew_ad(1,1,1,2), imigr )
            IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
         CASE ( 0 )
            CALL mppsend( 1, t3ew_ad(1,1,1,1), imigr, nowe, ml_req1 )
            CALL mppsend( 2, t3we_ad(1,1,1,1), imigr, noea, ml_req2 )
            CALL mpprecv( 1, t3ew_ad(1,1,1,2), imigr )
            CALL mpprecv( 2, t3we_ad(1,1,1,2), imigr )
            IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
            IF(l_isend) CALL mpi_wait(ml_req2,ml_stat,ml_err)
         CASE ( 1 )
            CALL mppsend( 1, t3ew_ad(1,1,1,1), imigr, nowe, ml_req1 )
            CALL mpprecv( 2, t3we_ad(1,1,1,2), imigr )
            IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
         END SELECT
#endif

         ! 5.1 Read Dirichlet lateral conditions

         SELECT CASE ( nbondi )

         CASE ( -1, 0, 1 )
            iihom = nlci-nreci
            DO jl = 1, jpreci
               ptab_ad(iihom +jl,:,:) = ptab_ad(iihom +jl,:,:) + t3ew_ad(:,jl,:,2)
               ptab_ad(jpreci+jl,:,:) = ptab_ad(jpreci+jl,:,:) + t3we_ad(:,jl,:,2)
            END DO

         END SELECT

      END SELECT    ! npolj 
      

      ! 4. north fold treatment
      ! -----------------------

      IF (PRESENT(cd_mpp)) THEN
         ! No north fold treatment (it is assumed to be already OK)
     
      ELSE      

      ! 4.1 treatment without exchange (jpni odd)
      !     T-point pivot  

      SELECT CASE ( jpni )

      CASE ( 1 )  ! only one proc along I, no mpp exchange

         SELECT CASE ( npolj )
  
         CASE ( 3 , 4 )    ! T pivot
            iloc = jpiglo - 2 * ( nimpp - 1 )

            SELECT CASE ( cd_type )

            CASE ( 'T' , 'S', 'W' )
               DO jk = jpk, 1, -1
                  DO ji = nlci, nlci/2+1, -1
                     ijt=iloc-ji+2
                     ptab_ad(ijt,nlcj-1,jk) =        ptab_ad(ijt,nlcj-1,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj-1,jk)
                     ptab_ad(ji ,nlcj-1,jk) = 0.0_wp
                  END DO
                  DO ji = nlci, 2, -1
                     ijt=iloc-ji+2
                     ptab_ad(ijt,nlcj-2,jk) =        ptab_ad(ijt,nlcj-2,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj  ,jk) 
                     ptab_ad(ji, nlcj,  jk) = 0.0_wp
                  END DO
               END DO
          
            CASE ( 'U' )
               DO jk = jpk, 1, -1
                  DO ji = nlci-1, nlci/2, -1
                     iju=iloc-ji+1
                     ptab_ad(iju,nlcj-1,jk) =        ptab_ad(iju,nlcj-1,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj-1,jk)
                     ptab_ad(ji ,nlcj-1,jk) = 0.0_wp
                  END DO
                  DO ji = nlci-1, 1, -1
                     iju=iloc-ji+1
                     ptab_ad(iju,nlcj-2,jk) =        ptab_ad(iju,nlcj-2,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj  ,jk)
                     ptab_ad(ji ,nlcj  ,jk) = 0.0_wp
                  END DO
               END DO

            CASE ( 'V' )
               DO jk = jpk, 1, -1
                  DO ji = nlci, 2, -1 
                     ijt=iloc-ji+2
                     ptab_ad(ijt,nlcj-3,jk) =        ptab_ad(ijt,nlcj-3,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj  ,jk)
                     ptab_ad(ji ,nlcj  ,jk) = 0.0_wp
                     ptab_ad(ijt,nlcj-2,jk) =        ptab_ad(ijt,nlcj-2,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj-1,jk)
                     ptab_ad(ji ,nlcj-1,jk) = 0.0_wp
                  END DO
               END DO

            CASE ( 'F', 'G' )
               DO jk = jpk, 1, -1
                  DO ji = nlci-1, 1, -1
                     iju=iloc-ji+1
                     ptab_ad(iju,nlcj-3,jk) =        ptab_ad(iju,nlcj-3,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj  ,jk)
                     ptab_ad(ji ,nlcj  ,jk) = 0.0_wp
                     ptab_ad(iju,nlcj-2,jk) =        ptab_ad(iju,nlcj-2,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj-1,jk)
                     ptab_ad(ji ,nlcj-1,jk) = 0.0_wp
                  END DO
               END DO
  
          END SELECT
       
         CASE ( 5 , 6 ) ! F pivot
            iloc=jpiglo-2*(nimpp-1)
  
            SELECT CASE ( cd_type )

            CASE ( 'T' , 'S', 'W' )
               DO jk = jpk, 1, -1
                  DO ji = nlci, 1, -1
                     ijt=iloc-ji+1
                     ptab_ad(ijt,nlcj-1,jk) =        ptab_ad(ijt,nlcj-1,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj  ,jk) 
                     ptab_ad(ji ,nlcj  ,jk) = 0.0_wp
                  END DO
               END DO

            CASE ( 'U' )
               DO jk = jpk, 1, -1
                  DO ji = nlci-1, 1, -1
                     iju=iloc-ji
                     ptab_ad(iju,nlcj-1,jk) =        ptab_ad(iju,nlcj-1,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj  ,jk)
                     ptab_ad(ji ,nlcj  ,jk) = 0.0_wp
                  END DO
               END DO

            CASE ( 'V' )
               DO jk = jpk, 1, -1
                  DO ji = nlci, nlci/2+1, -1
                     ijt=iloc-ji+1
                     ptab_ad(ijt,nlcj-1,jk) =        ptab_ad(ijt,nlcj-1,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj-1,jk)
                     ptab_ad(ji ,nlcj-1,jk) = 0.0_wp
                  END DO
                  DO ji = nlci, 1, -1
                     ijt=iloc-ji+1
                     ptab_ad(ijt,nlcj-2,jk) =        ptab_ad(ijt,nlcj-2,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj  ,jk) 
                     ptab_ad(ji, nlcj  ,jk) = 0.0_wp
                  END DO
               END DO

            CASE ( 'F', 'G' )
               DO jk = jpk, 1, -1
                  DO ji = nlci-1, nlci/2+1, -1
                     iju=iloc-ji
                     ptab_ad(iju,nlcj-1,jk) =        ptab_ad(iju,nlcj-1,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj-1,jk)
                     ptab_ad(ji ,nlcj-1,jk) = 0.0_wp
                  END DO
                  DO ji = nlci-1, 1, -1
                     iju=iloc-ji
                     ptab_ad(iju,nlcj-2,jk) =        ptab_ad(iju,nlcj-2,jk) &
                        &                   + psgn * ptab_ad(ji ,nlcj  ,jk)
                     ptab_ad(ji ,nlcj  ,jk) = 0.0_wp
                  END DO
               END DO
            END SELECT  ! cd_type

            ptab_ad(nlci,nlcj,jk) = 0.e0
            ptab_ad( 1  ,nlcj,jk) = 0.e0
            
         END SELECT     !  npolj
  
      CASE DEFAULT ! more than 1 proc along I
         IF ( npolj /= 0 ) CALL mpp_lbc_north_adj (ptab_ad, cd_type, psgn)  ! only for northern procs.

      END SELECT ! jpni 

      ENDIF

      ! 3. North and south directions
      ! -----------------------------

      ! 3.3 Write Dirichlet lateral conditions

      ijhom = nlcj-jprecj

      t3ns_ad(:,:,:,:) = 0.0_wp
      t3sn_ad(:,:,:,:) = 0.0_wp

      SELECT CASE ( nbondj )
      CASE ( -1 )
         DO jl = 1, jprecj
            t3sn_ad(:,jl,:,1)     = ptab_ad(:,ijhom+jl,:)
            ptab_ad(:,ijhom+jl,:) = 0.0_wp
         END DO
      CASE ( 0 ) 
         DO jl = 1, jprecj
             t3sn_ad(:,jl,:,1)     = ptab_ad(:,ijhom+jl      ,:)
             ptab_ad(:,ijhom+jl,:) = 0.0_wp
             t3ns_ad(:,jl,:,1)     = ptab_ad(:,jl,:)
             ptab_ad(:,jl      ,:) = 0.0_wp
         END DO
      CASE ( 1 )
         DO jl = 1, jprecj
            t3ns_ad(:,jl,:,1) = ptab_ad(:,jl,:)
            ptab_ad(:,jl,:)   = 0.0_wp
         END DO
      END SELECT

      ! 3.2 Migrations

#if defined key_mpp_shmem
error "key_mpp_shmem not supported in nemovar"
#elif defined key_mpp_mpi
      !! * Local variables   (MPI version)
  
      imigr=jprecj*jpi*jpk

      SELECT CASE ( nbondj )     
      CASE ( -1 )
         CALL mppsend( 4, t3sn_ad(1,1,1,1), imigr, nono, ml_req1 )
         CALL mpprecv( 3, t3ns_ad(1,1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1, ml_stat, ml_err)
      CASE ( 0 )
         CALL mppsend( 3, t3ns_ad(1,1,1,1), imigr, noso, ml_req1 )
         CALL mppsend( 4, t3sn_ad(1,1,1,1), imigr, nono, ml_req2 )
         CALL mpprecv( 3, t3ns_ad(1,1,1,2), imigr )
         CALL mpprecv( 4, t3sn_ad(1,1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1, ml_stat, ml_err)
         IF(l_isend) CALL mpi_wait(ml_req2, ml_stat, ml_err)
      CASE ( 1 ) 
         CALL mppsend( 3, t3ns_ad(1,1,1,1), imigr, noso, ml_req1 )
         CALL mpprecv( 4, t3sn_ad(1,1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1, ml_stat, ml_err)
      END SELECT

#endif

      ! 3.1 Read Dirichlet lateral conditions

      IF( nbondj /= 2 ) THEN
         ijhom = nlcj-nrecj
         DO jl = 1, jprecj
             ptab_ad(:,jprecj+jl,:) = ptab_ad(:,jprecj+jl,:) + t3sn_ad(:,jl,:,2)
             ptab_ad(:,ijhom +jl,:) = ptab_ad(:,ijhom +jl,:) + t3ns_ad(:,jl,:,2)
         END DO
      ENDIF

      ! 2. East and west directions exchange
      ! ------------------------------------

      ! 2.3 Write Dirichlet lateral conditions

      iihom = nlci-jpreci

      SELECT CASE ( nbondi)
      CASE ( -1 )
         DO jl = 1, jpreci
            t3we_ad(:,jl,:,1)     = ptab_ad(iihom+jl,:,:)
            ptab_ad(iihom+jl,:,:) = 0.0_wp
         END DO
      CASE ( 0 ) 
         DO jl = 1, jpreci
            t3we_ad(:,jl,:,1)     = ptab_ad(iihom+jl,:,:)
            ptab_ad(iihom+jl,:,:) = 0.0_wp
            t3ew_ad(:,jl,:,1)     = ptab_ad(jl      ,:,:)
            ptab_ad(jl      ,:,:) = 0.0_wp
         END DO
      CASE ( 1 )
         DO jl = 1, jpreci
            t3ew_ad(:,jl,:,1)     = ptab_ad(jl      ,:,:)
            ptab_ad(jl      ,:,:) = 0.0_wp
         END DO
      END SELECT

      ! 2.2 Migrations

#if defined key_mpp_shmem
error "key_mpp_shmem not support in nemovar"
#elif defined key_mpp_mpi
      !! * Local variables   (MPI version)

      imigr=jpreci*jpj*jpk
      
      SELECT CASE ( nbondi )
      CASE ( -1 )
         CALL mppsend( 2, t3we_ad(1,1,1,1), imigr, noea, ml_req1 )
         CALL mpprecv( 1, t3ew_ad(1,1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
      CASE ( 0 )
         CALL mppsend( 1, t3ew_ad(1,1,1,1), imigr, nowe, ml_req1 )
         CALL mppsend( 2, t3we_ad(1,1,1,1), imigr, noea, ml_req2 )
         CALL mpprecv( 1, t3ew_ad(1,1,1,2), imigr )
         CALL mpprecv( 2, t3we_ad(1,1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
         IF(l_isend) CALL mpi_wait(ml_req2,ml_stat,ml_err)
      CASE ( 1 )
         CALL mppsend( 1, t3ew_ad(1,1,1,1), imigr, nowe, ml_req1 )
         CALL mpprecv( 2, t3we_ad(1,1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
      END SELECT
#endif

      ! 2.1 Read Dirichlet lateral conditions

     SELECT CASE ( nbondi )
     CASE ( -1, 0, 1 )    ! all exept 2 
        iihom = nlci-nreci
        DO jl = 1, jpreci
           ptab_ad(iihom +jl,:,:) = ptab_ad(iihom +jl,:,:) + t3ew_ad(:,jl,:,2)
           ptab_ad(jpreci+jl,:,:) = ptab_ad(jpreci+jl,:,:) + t3we_ad(:,jl,:,2)
        END DO
     END SELECT

      ! 1. standard boundary treatment
      ! ------------------------------

      IF( PRESENT( cd_mpp ) ) THEN
         DO jj = nlcj+1, jpj   ! only fill extra allows last line
            ptab_ad(1:nlci, jj, :) = ptab_ad(1:nlci, nlej, :)
         END DO
         DO ji = nlci+1, jpi   ! only fill extra allows last column
            ptab_ad(ji    , : , :) = ptab_ad(nlei  , :   , :)
         END DO
      ELSE      

         !                                        ! North-South boundaries
         !                                        ! ======================
          SELECT CASE ( cd_type )
          CASE ( 'T', 'U', 'V', 'W' )
             ptab_ad(:,nlcj-jprecj+1:jpj   ,:) = 0.0_wp
             ptab_ad(:,     1       :jprecj,:) = 0.0_wp
          CASE ( 'F' )
             ptab_ad(:,nlcj-jprecj+1:jpj   ,:) = 0.0_wp
          END SELECT

         !                                        ! East-West boundaries
         !                                        ! ====================
         IF( nbondi == 2 .AND.   &      ! Cyclic east-west
            &   (nperio == 1 .OR. nperio == 4 .OR. nperio == 6) ) THEN
            ptab_ad( 2   ,:,:) = ptab_ad( 2   ,:,:) + ptab_ad(jpi,:,:)
            ptab_ad(jpi  ,:,:) = 0
            ptab_ad(jpim1,:,:) = ptab_ad(jpim1,:,:) + ptab_ad( 1 ,:,:)
            ptab_ad( 1   ,:,:) = 0
         ELSE                           ! closed
            SELECT CASE ( cd_type )
            CASE ( 'T', 'U', 'V', 'W' )
               ptab_ad(nlci-jpreci+1:jpi   ,:,:) = 0.e0
               ptab_ad(     1       :jpreci,:,:) = 0.e0
            CASE ( 'F' )
               ptab_ad(nlci-jpreci+1:jpi   ,:,:) = 0.e0
            END SELECT 
         ENDIF

      ENDIF

   END SUBROUTINE mpp_lnk_3d_adj

   SUBROUTINE mpp_lnk_2d_adj( pt2d_ad, cd_type, psgn, cd_mpp, pval )
      !!----------------------------------------------------------------------
      !!                  ***  routine mpp_lnk_2d_adj  ***
      !!                  
      !! ** Purpose :   Message passing manadgement for 2d array
      !!
      !! ** Method  :   Use mppsend and mpprecv function for passing mask 
      !!      between processors following neighboring subdomains.
      !!            domain parameters
      !!                    nlci   : first dimension of the local subdomain
      !!                    nlcj   : second dimension of the local subdomain
      !!                    nbondi : mark for "east-west local boundary"
      !!                    nbondj : mark for "north-south local boundary"
      !!                    noea   : number for local neighboring processors 
      !!                    nowe   : number for local neighboring processors
      !!                    noso   : number for local neighboring processors
      !!                    nono   : number for local neighboring processors
      !!
      !!----------------------------------------------------------------------
      !! * Arguments
      CHARACTER(len=1) , INTENT( in ) ::   &
         cd_type       ! define the nature of pt2d_ad array grid-points
         !             !  = T , U , V , F , W 
         !             !  = S : T-point, north fold treatment
         !             !  = G : F-point, north fold treatment
         !             !  = I : sea-ice velocity at F-point with index shift
      REAL(wp), INTENT( in ) ::   &
         psgn          ! control of the sign change
         !             !   = -1. , the sign is changed if north fold boundary
         !             !   =  1. , the sign is kept  if north fold boundary
      REAL(wp), DIMENSION(jpi,jpj), INTENT( inout ) ::   &
         pt2d_ad          ! 2D array on which the boundary condition is applied
      CHARACTER(len=3), INTENT( in ), OPTIONAL ::    &
         cd_mpp        ! fill the overlap area only 
      REAL(wp)        , INTENT(in   ), OPTIONAL           ::   pval      ! background value (used at closed boundaries)
                                                                         ! only here for compatibility

      !! * Local variables
      INTEGER  ::   ji, jj, jl      ! dummy loop indices
      INTEGER  ::   &
         imigr, iihom, ijhom,    &  ! temporary integers
         iloc, ijt, iju             !    "          "
      INTEGER  ::   ml_req1, ml_req2, ml_err     ! for key_mpi_isend
      INTEGER  ::   ml_stat(MPI_STATUS_SIZE)     ! for key_mpi_isend
      !!----------------------------------------------------------------------


      ! 5. East and west directions exchange
      ! ------------------------------------

      SELECT CASE ( npolj )

      CASE ( 3, 4, 5, 6 )

         ! 5.3 Write Dirichlet lateral conditions

         iihom = nlci-jpreci

         SELECT CASE ( nbondi)
         CASE ( -1 )
            DO jl = 1, jpreci
               t2we_ad(:,jl,1)     = pt2d_ad(iihom+jl,:)
               pt2d_ad(iihom+jl,:) = 0.0_wp
            END DO
         CASE ( 0 ) 
            DO jl = 1, jpreci
               t2we_ad(:,jl,1)     = pt2d_ad(iihom+jl,:)
               pt2d_ad(iihom+jl,:) = 0.0_wp
               t2ew_ad(:,jl,1)     = pt2d_ad(jl      ,:)
               pt2d_ad(jl      ,:) = 0.0_wp
            END DO
         CASE ( 1 )
            DO jl = 1, jpreci
               t2ew_ad(:,jl,1)     = pt2d_ad(jl      ,:)
               pt2d_ad(jl      ,:) = 0.0_wp
            END DO
         END SELECT

         ! 5.2 Migrations

#if defined key_mpp_shmem
error "key_mpp_shmem not support in nemovar"
#elif defined key_mpp_mpi
         !! MPI version

         imigr=jpreci*jpj
  
       SELECT CASE ( nbondi )
         CASE ( -1 )
            CALL mppsend( 2, t2we_ad(1,1,1), imigr, noea, ml_req1 )
            CALL mpprecv( 1, t2ew_ad(1,1,2), imigr )
            IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
         CASE ( 0 )
            CALL mppsend( 1, t2ew_ad(1,1,1), imigr, nowe, ml_req1 )
            CALL mppsend( 2, t2we_ad(1,1,1), imigr, noea, ml_req2 )
            CALL mpprecv( 1, t2ew_ad(1,1,2), imigr )
            CALL mpprecv( 2, t2we_ad(1,1,2), imigr )
            IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
            IF(l_isend) CALL mpi_wait(ml_req2,ml_stat,ml_err)
         CASE ( 1 )
            CALL mppsend( 1, t2ew_ad(1,1,1), imigr, nowe, ml_req1 )
            CALL mpprecv( 2, t2we_ad(1,1,2), imigr )
            IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
         END SELECT
#endif

         ! 5.1 Read Dirichlet lateral conditions

         SELECT CASE ( nbondi )

         CASE ( -1, 0, 1 )
            iihom = nlci-nreci 
            DO jl = 1, jpreci
               DO jj=1, jpj
                  pt2d_ad(iihom +jl,jj) = pt2d_ad(iihom +jl,jj) + t2ew_ad(jj,jl,2)
                  pt2d_ad(jpreci+jl,jj) = pt2d_ad(jpreci+jl,jj) + t2we_ad(jj,jl,2)
               END DO
            END DO

         END SELECT

      END SELECT    ! npolj 
      

      ! 4. north fold treatment
      ! -----------------------

      IF (PRESENT(cd_mpp)) THEN
         ! No north fold treatment (it is assumed to be already OK)
     
      ELSE      

      ! 4.1 treatment without exchange (jpni odd)
      !     T-point pivot  

      SELECT CASE ( jpni )

      CASE ( 1 )  ! only one proc along I, no mpp exchange

         SELECT CASE ( npolj )
  
         CASE ( 3 , 4 )    ! T pivot
            iloc = jpiglo - 2 * ( nimpp - 1 )

            SELECT CASE ( cd_type )

            CASE ( 'T' , 'S', 'W' )
               DO ji = nlci, nlci/2+1, -1
                  ijt=iloc-ji+2
                  pt2d_ad(ijt,nlcj-1) =        pt2d_ad(ijt,nlcj-1) &
                     &                + psgn * pt2d_ad(ji ,nlcj-1)
                  pt2d_ad(ji ,nlcj-1) = 0.0_wp
               END DO
               DO ji = nlci, 2, -1
                  ijt=iloc-ji+2
                  pt2d_ad(ijt,nlcj-2) =        pt2d_ad(ijt,nlcj-2) &
                     &                + psgn * pt2d_ad(ji ,nlcj  ) 
                  pt2d_ad(ji, nlcj  ) = 0.0_wp
               END DO
          
            CASE ( 'U' )
               DO ji = nlci-1, nlci/2, -1
                  iju=iloc-ji+1
                  pt2d_ad(iju,nlcj-1) =        pt2d_ad(iju,nlcj-1) &
                     &                + psgn * pt2d_ad(ji ,nlcj-1)
                  pt2d_ad(ji ,nlcj-1) = 0.0_wp
               END DO
               DO ji = nlci-1, 1, -1
                  iju=iloc-ji+1
                  pt2d_ad(iju,nlcj-2) =        pt2d_ad(iju,nlcj-2) &
                     &                + psgn * pt2d_ad(ji ,nlcj  )
                  pt2d_ad(ji ,nlcj  ) = 0.0_wp
               END DO

            CASE ( 'V' )
               DO ji = nlci, 2, -1 
                  ijt=iloc-ji+2
                  pt2d_ad(ijt,nlcj-3) =        pt2d_ad(ijt,nlcj-3) &
                     &                + psgn * pt2d_ad(ji ,nlcj  )
                  pt2d_ad(ji ,nlcj  ) = 0.0_wp
                  pt2d_ad(ijt,nlcj-2) =        pt2d_ad(ijt,nlcj-2) &
                     &                + psgn * pt2d_ad(ji ,nlcj-1)
                  pt2d_ad(ji ,nlcj-1) = 0.0_wp
               END DO

            CASE ( 'F', 'G' )
               DO ji = nlci-1, 1, -1
                  iju=iloc-ji+1
                  pt2d_ad(iju,nlcj-3) =        pt2d_ad(iju,nlcj-3) &
                     &                + psgn * pt2d_ad(ji ,nlcj  )
                  pt2d_ad(ji ,nlcj  ) = 0.0_wp
                  pt2d_ad(iju,nlcj-2) =        pt2d_ad(iju,nlcj-2) &
                     &                + psgn * pt2d_ad(ji ,nlcj-1)
                  pt2d_ad(ji ,nlcj-1) = 0.0_wp
               END DO

            CASE ( 'I' )                                  ! ice U-V point
               DO ji = nlci, 3, -1
                  iju = iloc - ji + 3
                  pt2d_ad(iju,nlcj-1) = pt2d_ad(iju,nlcj-1) + psgn * pt2d_ad(ji,nlcj)
                  pt2d_ad(ji,nlcj)    = 0.0_wp
               END DO
               pt2d_ad(3,nlcj-1) = pt2d_ad(3,nlcj-1) + psgn * pt2d_ad(2,nlcj)
               pt2d_ad(2,nlcj)   = 0.0_wp
  
          END SELECT
       
         CASE ( 5 , 6 ) ! F pivot
            iloc=jpiglo-2*(nimpp-1)
  
            SELECT CASE ( cd_type )

            CASE ( 'T' , 'S', 'W' )
               DO ji = nlci, 1, -1
                  ijt=iloc-ji+1
                  pt2d_ad(ijt,nlcj-1) =        pt2d_ad(ijt,nlcj-1) &
                     &                + psgn * pt2d_ad(ji ,nlcj  ) 
                  pt2d_ad(ji ,nlcj  ) = 0.0_wp
               END DO

            CASE ( 'U' )
               DO ji = nlci-1, 1, -1
                  iju=iloc-ji
                  pt2d_ad(iju,nlcj-1) =        pt2d_ad(iju,nlcj-1) &
                     &                + psgn * pt2d_ad(ji ,nlcj  )
                  pt2d_ad(ji ,nlcj  ) = 0.0_wp
               END DO

            CASE ( 'V' )
               DO ji = nlci, nlci/2+1, -1
                  ijt=iloc-ji+1
                  pt2d_ad(ijt,nlcj-1) =        pt2d_ad(ijt,nlcj-1) &
                     &                + psgn * pt2d_ad(ji ,nlcj-1)
                  pt2d_ad(ji ,nlcj-1) = 0.0_wp
               END DO
               DO ji = nlci, 1, -1
                  ijt=iloc-ji+1
                  pt2d_ad(ijt,nlcj-2) =        pt2d_ad(ijt,nlcj-2) &
                     &                + psgn * pt2d_ad(ji ,nlcj  ) 
                  pt2d_ad(ji, nlcj  ) = 0.0_wp
               END DO

            CASE ( 'F', 'G' )
               DO ji = nlci-1, nlci/2+1, -1
                  iju=iloc-ji
                  pt2d_ad(iju,nlcj-1) =        pt2d_ad(iju,nlcj-1) &
                     &                + psgn * pt2d_ad(ji ,nlcj-1)
                  pt2d_ad(ji ,nlcj-1) = 0.0_wp
               END DO
               DO ji = nlci-1, 1, -1
                  iju=iloc-ji
                  pt2d_ad(iju,nlcj-2) =        pt2d_ad(iju,nlcj-2) &
                     &                + psgn * pt2d_ad(ji ,nlcj  )
                  pt2d_ad(ji ,nlcj  ) = 0.0_wp
               END DO

            CASE ( 'I' )                                  ! ice U-V point
               DO ji = nlci-1, 2 , -1
                  ijt = iloc - ji + 2
                  pt2d_ad(ji,nlcj-1)  = pt2d_ad(ji,nlcj-1)  +  0.5 * pt2d_ad(ji,nlcj)
                  pt2d_ad(ijt,nlcj-1) = pt2d_ad(ijt,nlcj-1) + psgn * pt2d_ad(ji,nlcj)
                  pt2d_ad(ji,nlcj)    = 0.0_wp
               END DO
               pt2d_ad( 2 ,nlcj) = 0.0_wp
  
            END SELECT  ! cd_type

         END SELECT     !  npolj
  
         pt2d_ad(nlci,nlcj) = 0.e0
         pt2d_ad( 1  ,nlcj) = 0.e0

      CASE DEFAULT ! more than 1 proc along I
         IF ( npolj /= 0 ) CALL mpp_lbc_north_adj (pt2d_ad, cd_type, psgn)  ! only for northern procs.

      END SELECT ! jpni 

      ENDIF

      ! 3. North and south directions
      ! -----------------------------

      ! 3.3 Write Dirichlet lateral conditions

      ijhom = nlcj-jprecj

      t2ns_ad(:,:,:) = 0.0_wp
      t2sn_ad(:,:,:) = 0.0_wp

      SELECT CASE ( nbondj )
      CASE ( -1 )
         DO jl = 1, jprecj
            t2sn_ad(:,jl,1)     = pt2d_ad(:,ijhom+jl)
            pt2d_ad(:,ijhom+jl) = 0.0_wp
         END DO
      CASE ( 0 ) 
         DO jl = 1, jprecj
            t2sn_ad(:,jl,1)     = pt2d_ad(:,ijhom+jl) 
            pt2d_ad(:,ijhom+jl) = 0.0_wp
            t2ns_ad(:,jl,1)     = pt2d_ad(:,jl      )
            pt2d_ad(:,jl      ) = 0.0_wp
         END DO
      CASE ( 1 )
         DO jl = 1, jprecj
            t2ns_ad(:,jl,1) = pt2d_ad(:,jl)
            pt2d_ad(:,jl)   = 0.0_wp
         END DO
      END SELECT

      ! 3.2 Migrations

#if defined key_mpp_shmem
error "key_mpp_shmem not support in nemovar"
#elif defined key_mpp_mpi
      !! * Local variables   (MPI version)
  
      imigr=jprecj*jpi

      SELECT CASE ( nbondj )     
      CASE ( -1 )
         CALL mppsend( 4, t2sn_ad(1,1,1), imigr, nono, ml_req1 )
         CALL mpprecv( 3, t2ns_ad(1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1, ml_stat, ml_err)
      CASE ( 0 )
         CALL mppsend( 3, t2ns_ad(1,1,1), imigr, noso, ml_req1 )
         CALL mppsend( 4, t2sn_ad(1,1,1), imigr, nono, ml_req2 )
         CALL mpprecv( 3, t2ns_ad(1,1,2), imigr )
         CALL mpprecv( 4, t2sn_ad(1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1, ml_stat, ml_err)
         IF(l_isend) CALL mpi_wait(ml_req2, ml_stat, ml_err)
      CASE ( 1 ) 
         CALL mppsend( 3, t2ns_ad(1,1,1), imigr, noso, ml_req1 )
         CALL mpprecv( 4, t2sn_ad(1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1, ml_stat, ml_err)
      END SELECT

#endif

      ! 3.1 Read Dirichlet lateral conditions

      IF( nbondj /= 2 ) THEN
         ijhom = nlcj-nrecj
         DO jl = 1, jprecj
            DO ji = 1, jpi
               pt2d_ad(ji,ijhom +jl) = pt2d_ad(ji,ijhom +jl) + t2ns_ad(ji,jl,2) 
               pt2d_ad(ji,jprecj+jl) = pt2d_ad(ji,jprecj+jl) + t2sn_ad(ji,jl,2)
            END DO
         END DO
      ENDIF

      ! 2. East and west directions exchange
      ! ------------------------------------

      ! 2.3 Write Dirichlet lateral conditions

      iihom = nlci-jpreci

      SELECT CASE ( nbondi )
      CASE ( -1 )
         DO jl = 1, jpreci
            t2we_ad(:,jl,1) = pt2d_ad(iihom+jl,:)
            pt2d_ad(iihom+jl,:)=0.0_wp
         END DO
      CASE ( 0 ) 
         DO jl = 1, jpreci
            t2we_ad(:,jl,1) = pt2d_ad(iihom+jl,:)
            pt2d_ad(iihom+jl,:)=0.0_wp
            t2ew_ad(:,jl,1) = pt2d_ad(jl      ,:)
            pt2d_ad(jl      ,:)=0.0_wp
         END DO
      CASE ( 1 )
         DO jl = 1, jpreci
            t2ew_ad(:,jl,1) = pt2d_ad(jl      ,:)
            pt2d_ad(jl      ,:)=0.0_wp
         END DO
      END SELECT
      
      ! 2.2 Migrations

#if defined key_mpp_shmem
error "key_mpp_shmem not support in nemovar"
#elif defined key_mpp_mpi
      !! * Local variables   (MPI version)

      imigr = jpreci * jpj

      SELECT CASE ( nbondi ) 
      CASE ( -1 )
         CALL mppsend( 2, t2we_ad(1,1,1), imigr, noea, ml_req1 )
         CALL mpprecv( 1, t2ew_ad(1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
      CASE ( 0 )
         CALL mppsend( 1, t2ew_ad(1,1,1), imigr, nowe, ml_req1 )
         CALL mppsend( 2, t2we_ad(1,1,1), imigr, noea, ml_req2 )
         CALL mpprecv( 1, t2ew_ad(1,1,2), imigr )
         CALL mpprecv( 2, t2we_ad(1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
         IF(l_isend) CALL mpi_wait(ml_req2,ml_stat,ml_err)
      CASE ( 1 )
         CALL mppsend( 1, t2ew_ad(1,1,1), imigr, nowe, ml_req1 )
         CALL mpprecv( 2, t2we_ad(1,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
      END SELECT
#endif

      ! 2.1 Read Dirichlet lateral conditions

      SELECT CASE ( nbondi )
      CASE ( -1, 0, 1 )    ! all exept 2 
         iihom = nlci-nreci
         DO jl = 1, jpreci
            pt2d_ad(iihom +jl,:) = pt2d_ad(iihom +jl,:) + t2ew_ad(:,jl,2)
            pt2d_ad(jpreci+jl,:) = pt2d_ad(jpreci+jl,:) + t2we_ad(:,jl,2)
         END DO
      END SELECT

      ! 1. standard boundary treatment
      ! ------------------------------

      IF( PRESENT( cd_mpp ) ) THEN
         DO jj = nlcj+1, jpj   ! only fill extra allows last line
            pt2d_ad(1:nlci, jj) = pt2d_ad(1:nlci, nlej)
         END DO
         DO ji = nlci+1, jpi   ! only fill extra allows last column
            pt2d_ad(ji    , : ) = pt2d_ad(nlei  , :   )
         END DO
      ELSE      

         !                                        ! East-West boundaries
         !                                        ! ====================
         IF( nbondi == 2 .AND.   &      ! Cyclic east-west
            &   (nperio == 1 .OR. nperio == 4 .OR. nperio == 6) ) THEN
            pt2d_ad( 2   ,:) = pt2d_ad( 2   ,:) + pt2d_ad(jpi,:)
            pt2d_ad(jpi  ,:) = 0
            pt2d_ad(jpim1,:) = pt2d_ad(jpim1,:) + pt2d_ad( 1 ,:)
            pt2d_ad( 1   ,:) = 0

         ELSE                           ! closed
            SELECT CASE ( cd_type )
            CASE ( 'T', 'U', 'V', 'W' )
               pt2d_ad(nlci-jpreci+1:jpi   ,:) = 0.e0
               pt2d_ad(     1       :jpreci,:) = 0.e0
            CASE ( 'F' )
               pt2d_ad(nlci-jpreci+1:jpi   ,:) = 0.e0
            END SELECT 
         ENDIF

         !                                        ! North-South boundaries
         !                                        ! ======================
         SELECT CASE ( cd_type )
         CASE ( 'T', 'U', 'V', 'W' )
            pt2d_ad(:,nlcj-jprecj+1:jpj   ) = 0.e0
            pt2d_ad(:,     1       :jprecj) = 0.e0
         CASE ( 'F' )
            pt2d_ad(:,nlcj-jprecj+1:jpj   ) = 0.e0
         END SELECT
     
      ENDIF

   END SUBROUTINE mpp_lnk_2d_adj

   SUBROUTINE mpp_lnk_3d_gather_adj( ptab1_ad, cd_type1, ptab2_ad, cd_type2, psgn )
      !!-----------------------------------------------------------------------
      !!
      !! ***  ROUTINE mpp_lnk_3d_gather_adj : ADJOINT OF ROUTINE mpp_lnk_3d_gather  ***
      !!
      !! ** Purpose of direct routine   : Message passing manadgement for two 
      !!              3D arrays
      !!
      !! ** Method of direct routine    : Use mppsend and mpprecv function for
      !!              passing mask between processors following neighboring 
      !!              subdomains.
      !!
      !! ** Comments on Adjoint Routine :
      !!
      !! ** Action  : ptab_ad1 and ptab_ad2  with update value at its periphery
      !!                   
      !! References : 
      !!
      !! History :
      !!        ! 07-07 (K. Mogensen) skeleton
      !!        ! 09-03 (A. Vidard) nemo V3
      !!-----------------------------------------------------------------------
      !! * Modules used
      !! * Arguments
      CHARACTER(len=1) , INTENT( in ) ::   &
         cd_type1, cd_type2       ! define the nature of ptab_ad array grid-points
         !                        ! = T , U , V , F , W points
         !                        ! = S : T-point, north fold treatment ???
         !                        ! = G : F-point, north fold treatment ???
      REAL(wp), INTENT( in ) ::   &
         psgn          ! control of the sign change
         !             !   = -1. , the sign is changed if north fold boundary
         !             !   =  1. , the sign is kept  if north fold boundary
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( inout ) ::   &
         ptab1_ad, ptab2_ad             ! 3D array on which the boundary condition is applied
      !! * Local variables
      INTEGER ::   ji, jk, jl   ! dummy loop indices
      INTEGER ::   imigr, iihom, ijhom, iloc, ijt, iju   ! temporary integers
      INTEGER ::   ml_req1, ml_req2, ml_err     ! for key_mpi_isend
      INTEGER ::   ml_stat(MPI_STATUS_SIZE)     ! for key_mpi_isend
      !!----------------------------------------------------------------------

      CALL ctl_stop('mpp_lnk_3d_gather_adj is not done yet')

   END SUBROUTINE mpp_lnk_3d_gather_adj

   SUBROUTINE mpp_lnk_2d_e_adj( pt2d_ad, cd_type, psgn, cd_mpp, pval )
      !!----------------------------------------------------------------------
      !!                  ***  routine mpp_lnk_2d_e_adj  ***
      !!                  
      !! ** Purpose of the direct routine:
      !!                Message passing manadgement for 2d array (with halo)
      !!
      !! ** Method  :   Use mppsend and mpprecv function for passing mask 
      !!      between processors following neighboring subdomains.
      !!            domain parameters
      !!                    nlci   : first dimension of the local subdomain
      !!                    nlcj   : second dimension of the local subdomain
      !!                    jpr2di : number of rows for extra outer halo
      !!                    jpr2dj : number of columns for extra outer halo
      !!                    nbondi : mark for "east-west local boundary"
      !!                    nbondj : mark for "north-south local boundary"
      !!                    noea   : number for local neighboring processors 
      !!                    nowe   : number for local neighboring processors
      !!                    noso   : number for local neighboring processors
      !!                    nono   : number for local neighboring processors
      !!   
      !! History of the direct routine:
      !!       
      !!   9.0  !  05-09  (R. Benshila, G. Madec)  original code
      !!
      !! History of the adjoint routine:
      !!       
      !!   9.0  !  09-03  (A. Vidard) Adjoint of the 05-09 version
      !!
      !!----------------------------------------------------------------------
      !! * Arguments
      CHARACTER(len=1) , INTENT( in ) ::   &
         cd_type       ! define the nature of pt2d_ad array grid-points
         !             !  = T , U , V , F , W 
         !             !  = S : T-point, north fold treatment
         !             !  = G : F-point, north fold treatment
         !             !  = I : sea-ice velocity at F-point with index shift
      REAL(wp), INTENT( in ) ::   &
         psgn          ! control of the sign change
         !             !   = -1. , the sign is changed if north fold boundary
         !             !   =  1. , the sign is kept  if north fold boundary
      REAL(wp), DIMENSION(jpi,jpj), INTENT( inout ) ::   &
         pt2d_ad          ! 2D array on which the boundary condition is applied
      CHARACTER(len=3), INTENT( in ), OPTIONAL ::    &
         cd_mpp        ! fill the overlap area only 
      REAL(wp)        , INTENT(in   ), OPTIONAL           ::   pval      ! background value (used at closed boundaries)
                                                                         ! only here for compatibility

      !! * Local variables
      INTEGER  ::   ji, jj, jl      ! dummy loop indices
      INTEGER  ::   &
         imigr, iihom, ijhom,    &  ! temporary integers
         iloc, ijt, iju             !    "          "
      INTEGER  ::   ml_req1, ml_req2, ml_err     ! for key_mpi_isend
      INTEGER  ::   ml_stat(MPI_STATUS_SIZE)     ! for key_mpi_isend
      INTEGER :: ipreci, iprecj
      !!----------------------------------------------------------------------

      ! take into account outer extra 2D overlap area
      ipreci = jpreci + jpr2di
      iprecj = jprecj + jpr2dj

      ! 5. East and west directions exchange
      ! ------------------------------------

      SELECT CASE ( npolj )

      CASE ( 3, 4, 5, 6 )

         ! 5.3 Write Dirichlet lateral conditions

         iihom = nlci-jpreci

         SELECT CASE ( nbondi)
         CASE ( -1 )
            DO jl = ipreci, 1, -1
               tr2we_ad(:,jl,1)     = pt2d_ad(iihom+jl,:)
               pt2d_ad(iihom+jl,:)  = 0.0_wp
            END DO
         CASE ( 0 ) 
            DO jl = ipreci, 1, -1
               tr2we_ad(:,jl,1)      = pt2d_ad(iihom+jl,:)
               pt2d_ad(iihom+jl,:)   = 0.0_wp
               tr2ew_ad(:,jl,1)      = pt2d_ad(jl- jpr2di,:)
               pt2d_ad(jl- jpr2di,:) = 0.0_wp
            END DO
         CASE ( 1 )
            DO jl = ipreci, 1, -1
               tr2ew_ad(:,jl,1)     = pt2d_ad(jl-jpr2di,:)
               pt2d_ad(jl-jpr2di,:) = 0.0_wp
            END DO
         END SELECT

         ! 5.2 Migrations

#if defined key_mpp_shmem
error "key_mpp_shmem not support in nemovar"
#elif defined key_mpp_mpi
         !! MPI version

         imigr=ipreci* ( jpj + 2*jpr2dj )
  
         SELECT CASE ( nbondi )
         CASE ( -1 )
            CALL mppsend( 2, tr2we_ad(1-jpr2dj,1,1), imigr, noea, ml_req1 )
            CALL mpprecv( 1, tr2ew_ad(1-jpr2dj,1,2), imigr )
            IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
         CASE ( 0 )
            CALL mppsend( 1, tr2ew_ad(1-jpr2dj,1,1), imigr, nowe, ml_req1 )
            CALL mppsend( 2, tr2we_ad(1-jpr2dj,1,1), imigr, noea, ml_req2 )
            CALL mpprecv( 1, tr2ew_ad(1-jpr2dj,1,2), imigr )
            CALL mpprecv( 2, tr2we_ad(1-jpr2dj,1,2), imigr )
            IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
            IF(l_isend) CALL mpi_wait(ml_req2,ml_stat,ml_err)
         CASE ( 1 )
            CALL mppsend( 1, tr2ew_ad(1-jpr2dj,1,1), imigr, nowe, ml_req1 )
            CALL mpprecv( 2, tr2we_ad(1-jpr2dj,1,2), imigr )
            IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
         END SELECT 
#endif

         ! 5.1 Read Dirichlet lateral conditions

         SELECT CASE ( nbondi )

         CASE ( -1, 0, 1 )
            iihom = nlci-nreci -jpr2di
            DO jl = 1, ipreci
               pt2d_ad(iihom +jl,:) = pt2d_ad(iihom +jl,:) + tr2ew_ad(:,jl,2)
               pt2d_ad(jpreci+jl,:) = pt2d_ad(jpreci+jl,:) + tr2we_ad(:,jl,2)
            END DO

         END SELECT

      END SELECT    ! npolj 
      

      ! 4. north fold treatment
      ! -----------------------

      ! 4.1 treatment without exchange (jpni odd)
      !     T-point pivot  

      SELECT CASE ( jpni )

      CASE ( 1 )  ! only one proc along I, no mpp exchange

         SELECT CASE ( npolj )
  
         CASE ( 3 , 4 )    ! T pivot
            iloc = jpiglo - 2 * ( nimpp - 1 )

            SELECT CASE ( cd_type )

            CASE ( 'T' , 'S', 'W' )
               DO ji = nlci+jpr2di, nlci/2+1, -1
                  ijt=iloc-ji+2
                  pt2d_ad(ijt,nlcj-1) =        pt2d_ad(ijt,nlcj-1) &
                     &                + psgn * pt2d_ad(ji ,nlcj-1)
                  pt2d_ad(ji ,nlcj-1) = 0.0_wp
               END DO
               DO jl = iprecj-1, 0, -1
                  DO ji = nlci+jpr2di, 2-jpr2di, -1
                     ijt=iloc-ji+2
                     pt2d_ad(ijt,nlcj-2-jl) =        pt2d_ad(ijt,nlcj-2-jl) &
                        &                   + psgn * pt2d_ad(ji ,nlcj+jl  ) 
                     pt2d_ad(ji, nlcj+jl)   = 0.0_wp
                  END DO
               END DO
          
            CASE ( 'U' )
               DO ji = nlci-1+jpr2di, nlci/2, -1
                  iju=iloc-ji+1
                  pt2d_ad(iju,nlcj-1) =        pt2d_ad(iju,nlcj-1) &
                     &                + psgn * pt2d_ad(ji ,nlcj-1)
                  pt2d_ad(ji ,nlcj-1) = 0.0_wp
               END DO
               DO jl =iprecj-1, 0, -1
                  DO ji = nlci-1-jpr2di, 1-jpr2di, -1
                     iju=iloc-ji+1
                     pt2d_ad(iju,nlcj-2-jl) =        pt2d_ad(iju,nlcj-2-jl) &
                        &                   + psgn * pt2d_ad(ji ,nlcj+jl)
                     pt2d_ad(ji ,nlcj+jl)   = 0.0_wp
                  END DO
               END DO

            CASE ( 'V' )
               DO jl = iprecj-1, -1, -1
                  DO ji = nlci+jpr2di, 2-jpr2di, -1
                     ijt=iloc-ji+2
                     pt2d_ad(ijt,nlcj-3-jl) =        pt2d_ad(ijt,nlcj-3-jl) &
                        &                   + psgn * pt2d_ad(ji ,nlcj+jl  )
                     pt2d_ad(ji ,nlcj+jl  ) = 0.0_wp
                  END DO
               END DO

            CASE ( 'F', 'G' )
               DO jl = iprecj-1, -1, -1
                  DO ji = nlci-1+jpr2di, 1-jpr2di, -1
                     iju=iloc-ji+1
                     pt2d_ad(iju,nlcj-3-jl) =        pt2d_ad(iju,nlcj-3-jl) &
                        &                   + psgn * pt2d_ad(ji ,nlcj+jl)
                     pt2d_ad(ji ,nlcj+jl)   = 0.0_wp
                  END DO
               END DO

            CASE ( 'I' )                                  ! ice U-V point
               DO jl = iprecj-1, 0, -1
                  DO ji = nlci+jpr2di, 3, -1
                     iju = iloc - ji + 3
                     pt2d_ad(iju,nlcj-1-jl) = pt2d_ad(iju,nlcj-1-jl) + psgn * pt2d_ad(ji,nlcj+jl)
                     pt2d_ad(ji,nlcj+jl)    = 0.0_wp
                  END DO
                  pt2d_ad(3,nlcj-1-jl)      = pt2d_ad(3,nlcj-1-jl) + psgn * pt2d_ad(2,nlcj+jl)
                  pt2d_ad(2,nlcj+jl)        = 0.0_wp
               END DO
   
          END SELECT
       
         CASE ( 5 , 6 ) ! F pivot
            iloc=jpiglo-2*(nimpp-1)
  
            SELECT CASE ( cd_type )

            CASE ( 'T' , 'S', 'W' )
               DO jl = iprecj-1, 0, -1
                  DO ji = nlci+jpr2di, 1-jpr2di, -1
                     ijt=iloc-ji+1
                     pt2d_ad(ijt,nlcj-1-jl) =        pt2d_ad(ijt,nlcj-1-jl) &
                        &                   + psgn * pt2d_ad(ji ,nlcj+jl) 
                     pt2d_ad(ji ,nlcj+jl)   = 0.0_wp
                  END DO
               END DO

            CASE ( 'U' )
              DO jl = iprecj-1, 0, -1
                 DO ji = nlci-1+jpr2di, 1-jpr2di, -1
                    iju=iloc-ji
                    pt2d_ad(iju,nlcj-1-jl) =        pt2d_ad(iju,nlcj-1-jl) &
                       &                   + psgn * pt2d_ad(ji ,nlcj + jl)
                    pt2d_ad(ji ,nlcj+jl)   = 0.0_wp
                 END DO
              END DO

            CASE ( 'V' )
               DO ji = nlci+jpr2di, nlci/2+1, -1
                  ijt=iloc-ji+1
                  pt2d_ad(ijt,nlcj-1) =        pt2d_ad(ijt,nlcj-1) &
                     &                + psgn * pt2d_ad(ji ,nlcj-1)
                  pt2d_ad(ji ,nlcj-1) = 0.0_wp
               END DO
               DO jl = iprecj-1, 0, -1
                  DO ji =  nlci+jpr2di, 1-jpr2di, -1
                     ijt=iloc-ji+1
                     pt2d_ad(ijt,nlcj-2-jl) =        pt2d_ad(ijt,nlcj-2-jl) &
                        &                   + psgn * pt2d_ad(ji ,nlcj+jl) 
                     pt2d_ad(ji, nlcj+jl)   = 0.0_wp
                  END DO
               END DO

            CASE ( 'F', 'G' )
               DO ji = nlci-1+jpr2di, nlci/2+1, -1
                  iju=iloc-ji
                  pt2d_ad(iju,nlcj-1) =        pt2d_ad(iju,nlcj-1) &
                     &                + psgn * pt2d_ad(ji ,nlcj-1)
                  pt2d_ad(ji ,nlcj-1) = 0.0_wp
               END DO
               DO jl =  iprecj-1, 0, -1
                  DO ji = nlci-1+jpr2di, 1-jpr2di, -1
                     iju=iloc-ji
                     pt2d_ad(iju,nlcj-2-jl) =        pt2d_ad(iju,nlcj-2-jl) &
                        &                   + psgn * pt2d_ad(ji ,nlcj+jl)
                     pt2d_ad(ji ,nlcj+jl)   = 0.0_wp
                  END DO
               END DO
            CASE ( 'I' )                                  ! ice U-V point
               DO jl = iprecj-1,0, -1
                  DO ji = nlci-1+jpr2di, 2, -1
                     ijt = iloc - ji + 2
                     pt2d_ad(ji,nlcj-1-jl)  = pt2d_ad(ji,nlcj-1-jl)  + 0.5  * pt2d_ad(ji,nlcj+jl)
                     pt2d_ad(ijt,nlcj-1-jl) = pt2d_ad(ijt,nlcj-1-jl) + psgn * pt2d_ad(ji,nlcj+jl)
                     pt2d_ad(ji,nlcj+jl)    = 0.0_wp
                  END DO
               END DO
               pt2d_ad( 2 ,nlcj) = 0.0_wp
              END SELECT  ! cd_type

         END SELECT     !  npolj
  
      CASE DEFAULT ! more than 1 proc along I
         IF ( npolj /= 0 ) CALL mpp_lbc_north_adj (pt2d_ad, cd_type, psgn)  ! only for northern procs.

      END SELECT ! jpni 

      ! 3. North and south directions
      ! -----------------------------

      ! 3.3 Write Dirichlet lateral conditions

      ijhom = nlcj-jprecj

      tr2ns_ad(:,:,:) = 0.0_wp
      tr2sn_ad(:,:,:) = 0.0_wp

      SELECT CASE ( nbondj )
      CASE ( -1 )
         DO jl = 1, iprecj
            tr2sn_ad(:,jl,1   ) = pt2d_ad(:,ijhom+jl)
            pt2d_ad(:,ijhom+jl) = 0.0_wp
         END DO
      CASE ( 0 ) 
         DO jl = 1, iprecj
            tr2ns_ad(:,jl,1    ) = pt2d_ad(:,jl-jpr2dj)
            pt2d_ad(:,jl-jpr2dj) = 0.0_wp
            tr2sn_ad(:,jl,1    ) = pt2d_ad(:,ijhom+jl ) 
            pt2d_ad(:,ijhom+jl ) = 0.0_wp
         END DO
      CASE ( 1 )
         DO jl = 1, iprecj
            tr2ns_ad(:,jl,1    ) = pt2d_ad(:,jl-jpr2dj)
            pt2d_ad(:,jl-jpr2dj) = 0.0_wp
         END DO
      END SELECT

      ! 3.2 Migrations

#if defined key_mpp_shmem
error "key_mpp_shmem not support in nemovar"
#elif defined key_mpp_mpi
      !! * Local variables   (MPI version)
  
      imigr = iprecj * ( jpi + 2*jpr2di )

      SELECT CASE ( nbondj )
      CASE ( -1 )
         CALL mppsend( 4, tr2sn_ad(1-jpr2di,1,1), imigr, nono, ml_req1 )
         CALL mpprecv( 3, tr2ns_ad(1-jpr2di,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
      CASE ( 0 )
         CALL mppsend( 3, tr2ns_ad(1-jpr2di,1,1), imigr, noso, ml_req1 )
         CALL mppsend( 4, tr2sn_ad(1-jpr2di,1,1), imigr, nono, ml_req2 )
         CALL mpprecv( 3, tr2ns_ad(1-jpr2di,1,2), imigr )
         CALL mpprecv( 4, tr2sn_ad(1-jpr2di,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
         IF(l_isend) CALL mpi_wait(ml_req2,ml_stat,ml_err)
      CASE ( 1 )
         CALL mppsend( 3, tr2ns_ad(1-jpr2di,1,1), imigr, noso, ml_req1 )
         CALL mpprecv( 4, tr2sn_ad(1-jpr2di,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
      END SELECT

#endif

      ! 3.1 Read Dirichlet lateral conditions

      IF( nbondj /= 2 ) THEN
         ijhom = nlcj-nrecj-jpr2dj
         DO jl = 1, iprecj
            pt2d_ad(:,ijhom +jl) = pt2d_ad(:,ijhom +jl) + tr2ns_ad(:,jl,2) 
            pt2d_ad(:,jprecj+jl) = pt2d_ad(:,jprecj+jl) + tr2sn_ad(:,jl,2)
         END DO
      ENDIF

      ! 2. East and west directions exchange
      ! ------------------------------------

      ! 2.3 Write Dirichlet lateral conditions

      iihom = nlci-jpreci

      SELECT CASE ( nbondi )
      CASE ( -1 )
         DO jl = 1, ipreci
            tr2we_ad(:,jl,1) = pt2d_ad( iihom+jl,:)
            pt2d_ad( iihom+jl,:)=0.0_wp
         END DO
      CASE ( 0 ) 
         DO jl = 1, ipreci
            tr2we_ad(:,jl,1) = pt2d_ad( iihom+jl,:)
            pt2d_ad( iihom+jl,:)=0.0_wp
            tr2ew_ad(:,jl,1) = pt2d_ad(jl-jpr2di,:)
            pt2d_ad(jl-jpr2di,:)=0.0_wp
         END DO
      CASE ( 1 )
         DO jl = 1, ipreci
            tr2ew_ad(:,jl,1) = pt2d_ad(jl-jpr2di,:)
            pt2d_ad(jl-jpr2di,:)=0.0_wp
         END DO
      END SELECT
      
      ! 2.2 Migrations

#if defined key_mpp_shmem
error "key_mpp_shmem not support in nemovar"
#elif defined key_mpp_mpi
      !! * Local variables   (MPI version)

      imigr = ipreci * ( jpj + 2*jpr2dj)

      SELECT CASE ( nbondi )
      CASE ( -1 )
         CALL mppsend( 2, tr2we_ad(1-jpr2dj,1,1), imigr, noea, ml_req1 )
         CALL mpprecv( 1, tr2ew_ad(1-jpr2dj,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
      CASE ( 0 )
         CALL mppsend( 1, tr2ew_ad(1-jpr2dj,1,1), imigr, nowe, ml_req1 )
         CALL mppsend( 2, tr2we_ad(1-jpr2dj,1,1), imigr, noea, ml_req2 )
         CALL mpprecv( 1, tr2ew_ad(1-jpr2dj,1,2), imigr )
         CALL mpprecv( 2, tr2we_ad(1-jpr2dj,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
         IF(l_isend) CALL mpi_wait(ml_req2,ml_stat,ml_err)
      CASE ( 1 )
         CALL mppsend( 1, tr2ew_ad(1-jpr2dj,1,1), imigr, nowe, ml_req1 )
         CALL mpprecv( 2, tr2we_ad(1-jpr2dj,1,2), imigr )
         IF(l_isend) CALL mpi_wait(ml_req1,ml_stat,ml_err)
      END SELECT
#endif

      ! 2.1 Read Dirichlet lateral conditions

      SELECT CASE ( nbondi )
      CASE ( -1, 0, 1 )    ! all exept 2 
         iihom = nlci-nreci-jpr2di
         DO jl = 1, ipreci
            pt2d_ad(iihom +jl,:) = pt2d_ad(iihom +jl,:) + tr2ew_ad(:,jl,2)
            pt2d_ad(jpreci+jl,:) = pt2d_ad(jpreci+jl,:) + tr2we_ad(:,jl,2)
         END DO
      END SELECT

      ! 1. standard boundary treatment
      ! ------------------------------


      !                                        ! North-South boundaries
      !                                        ! ======================
      SELECT CASE ( cd_type )
      CASE ( 'T', 'U', 'V', 'W' , 'I' )
         pt2d_ad(:,nlcj-jprecj+1:jpj+jpr2dj) = 0.e0
         pt2d_ad(:,  1-jpr2dj   :  jprecj  ) = 0.e0
      CASE ( 'F' )
         pt2d_ad(:,nlcj-jprecj+1:jpj+jpr2dj ) = 0.e0
      END SELECT

      !                                        ! East-West boundaries
      !                                        ! ====================
      IF( nbondi == 2 .AND.   &      ! Cyclic east-west
	 &   (nperio == 1 .OR. nperio == 4 .OR. nperio == 6) ) THEN
	  pt2d_ad(     2      :2+jpr2di,:) = pt2d_ad(     2      :2+jpr2di,:) + pt2d_ad(   jpi  :jpi+jpr2di,:)
	  pt2d_ad(    jpi   :jpi+jpr2di,:) = 0
	  pt2d_ad(jpim1-jpr2di:  jpim1 ,:) = pt2d_ad(jpim1-jpr2di:  jpim1 ,:) + pt2d_ad(1-jpr2di:     1    ,:)
	  pt2d_ad(1-jpr2di:      1     ,:) = 0

      ELSE                           ! closed
	 SELECT CASE ( cd_type )
	 CASE ( 'T', 'U', 'V', 'W' , 'I' )
	    pt2d_ad(nlci-jpreci+1:jpi+jpr2di,:) = 0.e0
	    pt2d_ad(  1-jpr2di   :jpreci    ,:) = 0.e0
	 CASE ( 'F' )
	    pt2d_ad(nlci-jpreci+1:jpi+jpr2di,:) = 0.e0
	 END SELECT
      ENDIF
     
   END SUBROUTINE mpp_lnk_2d_e_adj

   SUBROUTINE mpp_lbc_north_3d_adj ( pt3d_ad, cd_type, psgn )
      !!-----------------------------------------------------------------------
      !!
      !! ** ROUTINE mpp_lbc_northadj_3d : ADJOINT OF ROUTINE mpp_lbc_north_3d  ***
      !!
      !! ** Purpose of direct routine   : Ensure proper north fold horizontal
      !!              bondary condition in mpp configuration in case of 
      !!              jpn1 > 1 (for 3d array)
      !!
      !! ** Method of direct routine    : Gather the 4 northern lines of the 
      !!              global domain on 1 processor and  apply lbc north-fold 
      !!              on this sub array. Then scatter the fold array back 
      !!              to the processors.
      !!
      !! ** Comments on Adjoint Routine :
      !!
      !! ** Action  : 
      !!                   
      !! References : 
      !!
      !! History :
      !!        ! 07-11 (K. Mogensen) Initial version
      !!        ! 09-03 (A. Vidard) V3
      !!-----------------------------------------------------------------------
      !! * Arguments
      CHARACTER(len=1), INTENT( in ) ::   &
         cd_type       ! nature of pt3d_ad grid-points
         !             !   = T ,  U , V , F or W  gridpoints
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( inout ) ::   &
         pt3d_ad          ! 3D array on which the boundary condition is applied
      REAL(wp), INTENT( in ) ::   &
         psgn          ! control of the sign change
         !             !   = -1. , the sign is changed if north fold boundary
         !             !   =  1. , the sign is kept  if north fold boundary

      !! * Local declarations
      INTEGER :: ji, jj, jk, jr, jproc
      INTEGER :: ierr
      INTEGER :: ildi,ilei,iilb
      INTEGER :: ijpj,ijpjm1,ij,ijt,iju
      INTEGER :: itaille
      REAL(wp), DIMENSION(jpiglo,4,jpk) :: ztabad
      REAL(wp), DIMENSION(jpi,4,jpk,jpni) :: znorthgloioad
      REAL(wp), DIMENSION(jpi,4,jpk) :: znorthlocad

      ! If we get in this routine it s because : North fold condition and mpp with more
      !   than one proc across i : we deal only with the North condition
      znorthlocad(  :,:,:  )=0.0_wp   
      znorthgloioad(:,:,:,:)=0.0_wp   

      ! 0. Sign setting
      ! ---------------

      ijpj=4
      ijpjm1=3
      
      ! put in znorthlocad the last 4 jlines of pt3d_ad
      DO jk = 1, jpk 
         DO jj = nlcj - ijpj +1, nlcj
            ij = jj - nlcj + ijpj
            znorthlocad(:,ij,jk) = pt3d_ad(:,jj,jk)
         END DO
      END DO
      

      IF (npolj /= 0 ) THEN
         ! Build in proc 0 of ncomm_north the znorthgloioad
         
#ifdef key_mpp_shmem
         not done : compiler error
#elif defined key_mpp_mpi
         itaille=jpi*jpk*ijpj
         CALL mpi_gather(znorthlocad,itaille,mpivar,znorthgloioad,itaille,mpivar,0,ncomm_north,ierr)
#endif

      ENDIF
      
      IF (narea == north_root+1 ) THEN
       ! recover the global north array
         ztabad(:,:,:) = 0_wp
         
         DO jr = 1, ndim_rank_north
            jproc = nrank_north(jr) + 1
            ildi  = nldit (jproc)
            ilei  = nleit (jproc)
            iilb  = nimppt(jproc)
            DO jk = 1, jpk 
                DO jj = 1, ijpj
                  DO ji = ildi, ilei
                     ztabad(ji+iilb-1,jj,jk) = znorthgloioad(ji,jj,jk,jr)
                     znorthgloioad(ji,jj,jk,jr)=0.0_wp
                  END DO
               END DO
            END DO
         END DO
         
         ! 2. North-Fold boundary conditions
         ! ----------------------------------
         
         SELECT CASE ( npolj )
            
         CASE ( 3 , 4 )    ! T pivot
            
            SELECT CASE ( cd_type )
               
            CASE ( 'T' , 'S', 'W' )
               DO jk = jpk, 1, -1
                  DO ji = jpiglo, jpiglo/2+1, -1
                     ijt=jpiglo-ji+2
                     ztabad(ijt,ijpj-1,jk) =        ztabad(ijt,ijpj-1,jk) &
                        &                  + psgn * ztabad(ji ,ijpj-1,jk)
                     ztabad(ji ,ijpj-1,jk) = 0.0_wp
                  END DO
                  DO ji = jpiglo, 2, -1
                     ijt=jpiglo-ji+2
                     ztabad(ijt,ijpj-2,jk) =        ztabad(ijt,ijpj-2,jk) &
                        &                  + psgn * ztabad(ji ,ijpj  ,jk) 
                     ztabad(ji, ijpj,  jk) = 0.0_wp
                  END DO
               END DO
               
            CASE ( 'U' )
               DO jk = jpk, 1, -1
                  DO ji = jpiglo-1, jpiglo/2, -1
                     iju=jpiglo-ji+1
                     ztabad(iju,ijpj-1,jk) =        ztabad(iju,ijpj-1,jk) &
                        &                  + psgn * ztabad(ji ,ijpj-1,jk)
                     ztabad(ji ,ijpj-1,jk) = 0.0_wp
                  END DO
                  DO ji = jpiglo-1, 1, -1
                     iju=jpiglo-ji+1
                     ztabad(iju,ijpj-2,jk) =        ztabad(iju,ijpj-2,jk) &
                        &                  + psgn * ztabad(ji ,ijpj  ,jk)
                     ztabad(ji ,ijpj  ,jk) = 0.0_wp
                  END DO
               END DO

            CASE ( 'V' )
               DO jk = jpk, 1, -1
                  DO ji = jpiglo, 2, -1 
                     ijt=jpiglo-ji+2
                     ztabad(ijt,ijpj-3,jk) =        ztabad(ijt,ijpj-3,jk) &
                        &                  + psgn * ztabad(ji ,ijpj  ,jk)
                     ztabad(ji ,ijpj  ,jk) = 0.0_wp
                     ztabad(ijt,ijpj-2,jk) =        ztabad(ijt,ijpj-2,jk) &
                        &                  + psgn * ztabad(ji ,ijpj-1,jk)
                     ztabad(ji ,ijpj-1,jk) = 0.0_wp
                  END DO
               END DO

            CASE ( 'F', 'G' )
               DO jk = jpk, 1, -1
                  DO ji = jpiglo-1, 1, -1
                     iju=jpiglo-ji+1
                     ztabad(iju,ijpj-3,jk) =        ztabad(iju,ijpj-3,jk) &
                        &                  + psgn * ztabad(ji ,ijpj  ,jk)
                     ztabad(ji ,ijpj  ,jk) = 0.0_wp
                     ztabad(iju,ijpj-2,jk) =        ztabad(iju,ijpj-2,jk) &
                        &                  + psgn * ztabad(ji ,ijpj-1,jk)
                     ztabad(ji ,ijpj-1,jk) = 0.0_wp
                  END DO
               END DO
  
            END SELECT

            ztabad( 1    ,ijpj,:) = 0.0_wp
            ztabad(jpiglo,ijpj,:) = 0.0_wp 
      
         CASE ( 5 , 6 ) ! F pivot
            
            SELECT CASE ( cd_type )
               
            CASE ( 'T' , 'S', 'W' )
               DO jk = jpk, 1, -1
                  DO ji = jpiglo, 1, -1
                     ijt=jpiglo-ji+1
                     ztabad(ijt,ijpj-1,jk) =        ztabad(ijt,ijpj-1,jk) &
                        &                  + psgn * ztabad(ji ,ijpj  ,jk) 
                     ztabad(ji ,ijpj  ,jk) = 0.0_wp
                  END DO
               END DO

            CASE ( 'U' )
               DO jk = jpk, 1, -1
                  DO ji = jpiglo-1, 1, -1
                     iju=jpiglo-ji
                     ztabad(iju,ijpj-1,jk) =        ztabad(iju,ijpj-1,jk) &
                        &                  + psgn * ztabad(ji ,ijpj  ,jk)
                     ztabad(ji ,ijpj  ,jk) = 0.0_wp
                  END DO
               END DO

            CASE ( 'V' )
               DO jk = jpk, 1, -1
                  DO ji = jpiglo, jpiglo/2+1, -1
                     ijt=jpiglo-ji+1
                     ztabad(ijt,ijpj-1,jk) =        ztabad(ijt,ijpj-1,jk) &
                        &                  + psgn * ztabad(ji ,ijpj-1,jk)
                     ztabad(ji ,ijpj-1,jk) = 0.0_wp
                  END DO
                  DO ji = jpiglo, 1, -1
                     ijt=jpiglo-ji+1
                     ztabad(ijt,ijpj-2,jk) =        ztabad(ijt,ijpj-2,jk) &
                        &                  + psgn * ztabad(ji ,ijpj  ,jk) 
                     ztabad(ji, ijpj  ,jk) = 0.0_wp
                  END DO
               END DO

            CASE ( 'F', 'G' )
               DO jk = jpk, 1, -1
                  DO ji = jpiglo-1, jpiglo/2+1, -1
                     iju=jpiglo-ji
                     ztabad(iju,ijpj-1,jk) =        ztabad(iju,ijpj-1,jk) &
                        &                  + psgn * ztabad(ji ,ijpj-1,jk)
                     ztabad(ji ,ijpj-1,jk) = 0.0_wp
                  END DO
                  DO ji = jpiglo-1, 1, -1
                     iju=jpiglo-ji
                     ztabad(iju,ijpj-2,jk) =        ztabad(iju,ijpj-2,jk) &
                        &                  + psgn * ztabad(ji ,ijpj  ,jk)
                     ztabad(ji ,ijpj  ,jk) = 0.0_wp
                  END DO
               END DO
            END SELECT  ! cd_type

            ztabad(jpiglo,ijpj,:) = 0.e0
            ztabad( 1  ,ijpj,:) = 0.e0

         CASE DEFAULT                           ! *  closed

            SELECT CASE ( cd_type) 

            CASE ( 'T' , 'U' , 'V' , 'W' )             ! T-, U-, V-, W-points
               ztabad(:, 1  ,jk) = 0.e0
               ztabad(:,ijpj,jk) = 0.e0

            CASE ( 'F' )                               ! F-point
               ztabad(:,ijpj,jk) = 0.e0

            END SELECT
            
         END SELECT     !  npolj

         !! Scatter back to pt3d_ad
         DO jr = 1, ndim_rank_north
            jproc=nrank_north(jr)+1
            ildi=nldit (jproc)
            ilei=nleit (jproc)
            iilb=nimppt(jproc)
            DO jk=  1, jpk
                DO jj=1,4
                  DO ji=ildi,ilei
                     znorthgloioad(ji,jj,jk,jr)=ztabad(ji+iilb-1,jj,jk)
                  END DO
               END DO
            END DO
         END DO
         
      ENDIF      ! only done on proc 0 of ncomm_north

#ifdef key_mpp_shmem
      not done yet in shmem : compiler error
#elif key_mpp_mpi
      IF ( npolj /= 0 ) THEN
         itaille=jpi*jpk*ijpj
         CALL mpi_scatter(znorthgloioad,itaille,mpivar,znorthlocad,itaille,mpivar,0,ncomm_north,ierr)
      ENDIF
#endif

      ! put in the last ijpj jlines of pt3d_ad znorthlocad
      DO jk = 1 , jpk 
         DO jj = nlcj - ijpj + 1 , nlcj
            ij = jj - nlcj + ijpj
            pt3d_ad(:,jj,jk)= znorthlocad(:,ij,jk)
         END DO
      END DO
       
   END SUBROUTINE mpp_lbc_north_3d_adj

   SUBROUTINE mpp_lbc_north_2d_adj ( pt2d_ad, cd_type, psgn )
      !!-----------------------------------------------------------------------
      !!
      !! ** ROUTINE mpp_lbc_north_2d_adj : ADJOINT OF ROUTINE mpp_lbc_north_2d  ***
      !!
      !! ** Purpose of direct routine   : Ensure proper north fold horizontal
      !!              bondary condition in mpp configuration in case of 
      !!              jpn1 > 1 (for 2d array)
      !!
      !! ** Method of direct routine    : Gather the 4 northern lines of the 
      !!              global domain on 1 processor and  apply lbc north-fold 
      !!              on this sub array. Then scatter the fold array back 
      !!              to the processors.
      !!
      !! ** Comments on Adjoint Routine :
      !!
      !! ** Action  : 
      !!                   
      !! References : 
      !!
      !! History :
      !!        ! 07-11 (K. Mogensen) Initial version
      !!        ! 09-03 (A. Vidard) nemo V3
      !!-----------------------------------------------------------------------
      !! * Arguments
      CHARACTER(len=1), INTENT( in ) ::   &
         cd_type       ! nature of pt2d_ad grid-points
         !             !   = T ,  U , V , F or W  gridpoints
      REAL(wp), DIMENSION(jpi,jpj), INTENT( inout ) ::   &
         pt2d_ad          ! 2D array on which the boundary condition is applied
      REAL(wp), INTENT( in ) ::   &
         psgn          ! control of the sign change
         !             !   = -1. , the sign is changed if north fold boundary
         !             !   =  1. , the sign is kept  if north fold boundary

      !! * Local declarations
      INTEGER :: ji, jj, jk, jr, jproc
      INTEGER :: ierr
      INTEGER :: ildi,ilei,iilb
      INTEGER :: ijpj,ijpjm1,ij,ijt,iju
      INTEGER :: itaille
      REAL(wp), DIMENSION(jpiglo,4) :: ztabad
      REAL(wp), DIMENSION(jpi,4,jpni) :: znorthgloioad
      REAL(wp), DIMENSION(jpi,4) :: znorthlocad

      ! If we get in this routine it s because : North fold condition and mpp with more
      !   than one proc across i : we deal only with the North condition
      
      ! 0. Sign setting
      ! ---------------
      
      ijpj=4
      ijpjm1=3
      
      znorthlocad(    :,:) = 0.0_wp
      znorthgloioad(:,:,:) = 0.0_wp
 
      ! put in znorthlocad the last 4 jlines of pt2d_ad
      DO jj = nlcj - ijpj +1, nlcj
         ij = jj - nlcj + ijpj
         znorthlocad(:,ij)=pt2d_ad(:,jj)
      END DO
      
      IF (npolj /= 0 ) THEN
         ! Build in proc 0 of ncomm_north the znorthgloioad
#ifdef key_mpp_shmem
         not done : compiler error
#elif defined key_mpp_mpi
         itaille=jpi*ijpj
         CALL MPI_GATHER(znorthlocad,itaille,mpivar,znorthgloioad,itaille,mpivar,0,ncomm_north,ierr)
#endif
      ENDIF
      
      IF (narea == north_root+1 ) THEN
       ! recover the global north array
         ztabad(:,:) = 0_wp
         
         DO jr = 1, ndim_rank_north
            jproc=nrank_north(jr)+1
            ildi=nldit (jproc)
            ilei=nleit (jproc)
            iilb=nimppt(jproc)
            DO jj=1,ijpj
               DO ji=ildi,ilei
                  ztabad(ji+iilb-1,jj)=znorthgloioad(ji,jj,jr)
                  znorthgloioad(ji,jj,jr)=0.0_wp
               END DO
            END DO
         END DO
         
       
         ! 2. North-Fold boundary conditions
         ! ----------------------------------
         
         SELECT CASE ( npolj )
  
         CASE ( 3 , 4 )    ! T pivot

            SELECT CASE ( cd_type )

            CASE ( 'T' , 'S', 'W' )
               DO ji = jpiglo, jpiglo/2+1, -1
                  ijt=jpiglo-ji+2
                  ztabad(ijt,ijpj-1) =        ztabad(ijt,ijpj-1) &
                     &               + psgn * ztabad(ji ,ijpj-1)
                  ztabad(ji ,ijpj-1) = 0.0_wp
               END DO
               DO ji = jpiglo, 2, -1
                  ijt=jpiglo-ji+2
                  ztabad(ijt,ijpj-2) =        ztabad(ijt,ijpj-2) &
                     &               + psgn * ztabad(ji ,ijpj  ) 
                  ztabad(ji, ijpj  ) = 0.0_wp
               END DO
          
            CASE ( 'U' )
               DO ji = jpiglo-1, jpiglo/2, -1
                  iju=jpiglo-ji+1
                  ztabad(iju,ijpj-1) =        ztabad(iju,ijpj-1) &
                     &               + psgn * ztabad(ji ,ijpj-1)
                  ztabad(ji ,ijpj-1) = 0.0_wp
               END DO
               DO ji = jpiglo-1, 1, -1
                  iju=jpiglo-ji+1
                  ztabad(iju,ijpj-2) =        ztabad(iju,ijpj-2) &
                     &               + psgn * ztabad(ji ,ijpj  )
                  ztabad(ji ,ijpj  ) = 0.0_wp
               END DO

            CASE ( 'V' )
               DO ji = jpiglo, 2, -1 
                  ijt=jpiglo-ji+2
                  ztabad(ijt,ijpj-3) =        ztabad(ijt,ijpj-3) &
                     &               + psgn * ztabad(ji ,ijpj  )
                  ztabad(ji ,ijpj  ) = 0.0_wp
                  ztabad(ijt,ijpj-2) =        ztabad(ijt,ijpj-2) &
                     &               + psgn * ztabad(ji ,ijpj-1)
                  ztabad(ji ,ijpj-1) = 0.0_wp
               END DO

            CASE ( 'F', 'G' )
               DO ji = jpiglo-1, 1, -1
                  iju=jpiglo-ji+1
                  ztabad(iju,ijpj-3) =        ztabad(iju,ijpj-3) &
                     &               + psgn * ztabad(ji ,ijpj  )
                  ztabad(ji ,ijpj  ) = 0.0_wp
                  ztabad(iju,ijpj-2) =        ztabad(iju,ijpj-2) &
                     &               + psgn * ztabad(ji ,ijpj-1)
                  ztabad(ji ,ijpj-1) = 0.0_wp
               END DO
  
          END SELECT

          ztabad(jpiglo,ijpj) = 0.e0
          ztabad( 1  ,ijpj)   = 0.e0
       
         CASE ( 5 , 6 ) ! F pivot
  
            SELECT CASE ( cd_type )

            CASE ( 'T' , 'S', 'W' )
               DO ji = jpiglo, 1, -1
                  ijt=jpiglo-ji+1
                  ztabad(ijt,ijpj-1) =        ztabad(ijt,ijpj-1) &
                     &               + psgn * ztabad(ji ,ijpj  ) 
                  ztabad(ji ,ijpj  ) = 0.0_wp
               END DO

            CASE ( 'U' )
               DO ji = jpiglo-1, 1, -1
                  iju=jpiglo-ji
                  ztabad(iju,ijpj-1) =        ztabad(iju,ijpj-1) &
                     &               + psgn * ztabad(ji ,ijpj  )
                     ztabad(ji ,ijpj  ) = 0.0_wp
               END DO

            CASE ( 'V' )
               DO ji = jpiglo, jpiglo/2+1, -1
                  ijt=jpiglo-ji+1
                  ztabad(ijt,ijpj-1) =        ztabad(ijt,ijpj-1) &
                     &               + psgn * ztabad(ji ,ijpj-1)
                  ztabad(ji ,ijpj-1) = 0.0_wp
               END DO
               DO ji = jpiglo, 1, -1
                  ijt=jpiglo-ji+1
                  ztabad(ijt,ijpj-2) =        ztabad(ijt,ijpj-2) &
                     &               + psgn * ztabad(ji ,ijpj  ) 
                  ztabad(ji, ijpj  ) = 0.0_wp
               END DO

            CASE ( 'F', 'G' )
               DO ji = jpiglo-1, jpiglo/2+1, -1
                  iju=jpiglo-ji
                  ztabad(iju,ijpj-1) =        ztabad(iju,ijpj-1) &
                     &               + psgn * ztabad(ji ,ijpj-1)
                  ztabad(ji ,ijpj-1) = 0.0_wp
               END DO
               DO ji = jpiglo-1, 1, -1
                  iju=jpiglo-ji
                  ztabad(iju,ijpj-2) =        ztabad(iju,ijpj-2) &
                     &               + psgn * ztabad(ji ,ijpj  )
                  ztabad(ji ,ijpj  ) = 0.0_wp
               END DO
            END SELECT  ! cd_type

            ztabad(jpiglo,ijpj) = 0.e0
            ztabad( 1  ,ijpj) = 0.e0

         END SELECT     !  npolj
       
         !     End of slab
         !     ===========

         !! Scatter back to pt2d_ad
         DO jr = 1, ndim_rank_north
            jproc=nrank_north(jr)+1
            ildi=nldit (jproc)
            ilei=nleit (jproc)
            iilb=nimppt(jproc)
            DO jj=1,4
               DO ji=ildi,ilei
                  znorthgloioad(ji,jj,jr)=ztabad(ji+iilb-1,jj)
               END DO
            END DO
         END DO

      ENDIF      ! only done on proc 0 of ncomm_north

#ifdef key_mpp_shmem
      not done yet in shmem : compiler error
#elif key_mpp_mpi
      IF ( npolj /= 0 ) THEN
         itaille=jpi*ijpj
         CALL mpi_scatter(znorthgloioad,itaille,mpivar,znorthlocad,itaille,mpivar,0,ncomm_north,ierr)
      ENDIF
#endif

      ! put in the last ijpj jlines of pt2d_ad znorthlocad
      DO jj = nlcj - ijpj + 1 , nlcj
         ij = jj - nlcj + ijpj
         pt2d_ad(:,jj)= znorthlocad(:,ij)
      END DO

   END SUBROUTINE mpp_lbc_north_2d_adj

#if defined key_ecmwf_dynmem   
   SUBROUTINE lib_mpp_alloc_adj      
      !!---------------------------------------------------------------------
      !!                   ***  routine lib_mpp_alloc_adj  ***      
      !!       
      !! ** Purpose :: Allocate memory for dynamic memory version      
      !!      
      !! ** Method  ::       
      !!       
      !! History :: 08/10 :: K. Mogensen initial version      
      !!      
      !!--------------------------------------------------------------------- 
      ALLOCATE( &         
         & t3ns_ad(jpi,jprecj,jpk,2), &         
         & t3sn_ad(jpi,jprecj,jpk,2), &         
         & t3ew_ad(jpj,jpreci,jpk,2), &         
         & t3we_ad(jpj,jpreci,jpk,2), &         
         & t2ns_ad(jpi,jprecj,2), &         
         & t2sn_ad(jpi,jprecj,2), &         
         & t2ew_ad(jpj,jpreci,2), &         
         & t2we_ad(jpj,jpreci,2), &         
         & tr2ns_ad(1-jpr2di:jpi+jpr2di,jprecj+jpr2dj,2), &         
         & tr2sn_ad(1-jpr2di:jpi+jpr2di,jprecj+jpr2dj,2), &         
         & tr2ew_ad(1-jpr2dj:jpj+jpr2dj,jpreci+jpr2di,2), &         
         & tr2we_ad(1-jpr2dj:jpj+jpr2dj,jpreci+jpr2di,2)  &         
         & )   
   END SUBROUTINE lib_mpp_alloc_adj
#endif
#else
   !!----------------------------------------------------------------------
   !!   Default case:            Dummy module        share memory computing
   !!----------------------------------------------------------------------
#endif
END MODULE lib_mpp_tam
