MODULE mppsumtam
   !!======================================================================
   !!                       ***  MODULE mppsumtam  ***
   !! NEMOTAM: Summation of arrays across processors
   !!======================================================================

   !!----------------------------------------------------------------------
   !!   mpp_sum_inter  : Interface for depending on nmppsum
   !!   mpp_sum_simple : Simple sum
   !!   loc_sum_inter  : Interface for depending on nmppsum (local version)
   !!   loc_sum_simple : Simple sum (local version)
   !!   loc_sum_indep  : Order independent MPP reproducible sum
   !!----------------------------------------------------------------------
   !! * Modules used   
   USE par_kind, ONLY : &   ! Precision variables
      & wp
   USE dom_oce, ONLY : &    ! Ocean space and time domain variables
      & nproc
   USE par_oce, ONLY : &    ! Ocean parameters
      & jpnij
#if defined key_mpp_mpi
   USE lib_mpp, ONLY : &    ! MPP library
      & mpi_comm_opa
#endif
   USE lib_mpp
   USE mpp_tam
   USE mppsum
   USE in_out_manager

   IMPLICIT NONE

   !! * Routine accessibility
   PRIVATE

   PUBLIC &
      & mpp_sum_inter,  & ! Interface for depending on nmppsum
      & mpp_sum_simple, & ! Simple sum
      & loc_sum_inter,  & ! Interface for depending on nmppsum
      & loc_sum_simple, & ! Simple sum
      & loc_sum_indep,  & ! Order independent local reproducible sum
      & nmppsum           ! Summation control

   !! * Module variables
   INTEGER :: nmppsum = 2 ! Choice of MPP sum for 
                          ! (1 = simple)
                          ! (2 = order independent)
   
CONTAINS

   FUNCTION mpp_sum_inter( pval, kn )
      !!----------------------------------------------------------------------
      !!               ***  ROUTINE mpp_sum_inter ***
      !!          
      !! ** Purpose : Summation of arrays across processors
      !!
      !! ** Method  : Call either mpp_sum_simple or mpp_sum_indep depending on
      !!              nmppsum
      !!
      !! ** Action  : This does only work for MPI. 
      !!              It does not work for SHMEM.
      !!
      !! References : http://www.mpi-forum.org
      !!
      !! History :
      !!        !  07-07  (K. Mogensen)  Original code
      !!----------------------------------------------------------------------
      !! * Function return
      REAL(wp) mpp_sum_inter
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kn
      REAL(wp), DIMENSION(kn), INTENT(IN) :: &
         & pval
      !! * Local declarations
     
      IF ( nmppsum == 1 ) THEN
         
         mpp_sum_inter = mpp_sum_simple( pval, kn )
         
      ELSEIF ( nmppsum == 2 ) THEN

         mpp_sum_inter = mpp_sum_indep( pval, kn )

      ELSE

         CALL ctl_stop( ' mpp_sum: Invalid nmppsum ' )

      ENDIF

   END FUNCTION mpp_sum_inter

   FUNCTION mpp_sum_simple( pval, kn )
      !!----------------------------------------------------------------------
      !!               ***  ROUTINE mpp_sum_simple ***
      !!          
      !! ** Purpose : Summation of arrays across processors
      !!
      !! ** Method  : This routine is the naive implementation using
      !!              mpi_allreduce for the sum of local data 
      !!
      !! ** Action  : This does only work for MPI. 
      !!              It does not work for SHMEM.
      !!
      !! References : http://www.mpi-forum.org
      !!
      !! History :
      !!        !  07-07  (K. Mogensen)  Original code
      !!----------------------------------------------------------------------
      !! * Function return
      REAL(wp) mpp_sum_simple
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kn
      REAL(wp), DIMENSION(kn), INTENT(IN) :: &
         & pval
      !! * Local declarations
      REAL(wp) :: &
         & zres
      INTEGER :: &
         & ierr
      INTEGER :: &
         & ji

      ! Compute local sum   
      zres = SUM( pval(1:kn) )    
      ! Global sum

      CALL mpp_sum( zres )

      mpp_sum_simple = zres

   END FUNCTION mpp_sum_simple

   FUNCTION loc_sum_inter( pval, kn )
      !!----------------------------------------------------------------------
      !!               ***  ROUTINE loc_sum_inter ***
      !!          
      !! ** Purpose : Summation of arrays across processors
      !!
      !! ** Method  : Call either loc_sum_simple or loc_sum_indep depending on
      !!              nmppsum
      !!
      !! ** Action  : 
      !!
      !! References : 
      !!
      !! History :
      !!        !  07-09  (K. Mogensen)  Original code
      !!----------------------------------------------------------------------
      !! * Function return
      REAL(wp) loc_sum_inter
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kn
      REAL(wp), DIMENSION(kn), INTENT(IN) :: &
         & pval
      !! * Local declarations
      
      IF ( nmppsum == 1 ) THEN
         
         loc_sum_inter = loc_sum_simple( pval, kn )
         
      ELSEIF ( nmppsum == 2 ) THEN

         loc_sum_inter = loc_sum_indep( pval, kn )

      ELSE

         CALL ctl_stop( ' loc_sum: Invalid nmppsum = ' )

      ENDIF

   END FUNCTION loc_sum_inter

   FUNCTION loc_sum_simple( pval, kn )
      !!----------------------------------------------------------------------
      !!               ***  ROUTINE loc_sum_simple ***
      !!          
      !! ** Purpose : Summation of arrays across processors
      !!
      !! ** Method  : This routine is the naive implementation for the sum 
      !!              of local data 
      !!
      !! ** Action  : 
      !!
      !! References : 
      !!
      !! History :
      !!        !  07-09  (K. Mogensen)  Original code
      !!----------------------------------------------------------------------
      !! * Function return
      REAL(wp) loc_sum_simple
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kn
      REAL(wp), DIMENSION(kn), INTENT(IN) :: &
         & pval
      !! * Local declarations
      REAL(wp) :: &
         & zloc
      INTEGER :: &
         & ierr
      INTEGER :: &
         & ji

      ! Compute local sum

      zloc = SUM( pval(1:kn) )
      
      loc_sum_simple = zloc

   END FUNCTION loc_sum_simple

   FUNCTION loc_sum_indep( pval, kn )
      !!----------------------------------------------------------------------
      !!               ***  ROUTINE loc_sum_indep ***
      !!          
      !! ** Purpose : Sum all elements in the pval array in
      !!              an accurate order-independent way.
      !!
      !! ** Method  : The code iterates the compensated summation until the 
      !!              result is guaranteed to be within 4*eps of the true sum.
      !!              It then rounds the result to the nearest floating-point 
      !!              number whose last three bits are zero, thereby 
      !!              guaranteeing an order-independent result.
      !!
      !!              This version is used for local arrays. It is identical
      !!              to the MPP version for clarity.
      !!
      !! ** Action  : 
      !!
      !! References : M. Fisher (ECMWF): IFS code + personal communication
      !!              The algorithm is based on Ogita et al. (2005) 
      !!              SIAM J. Sci. Computing, Vol.26, No.6, pp1955-1988. 
      !!              This is based in turn on an algorithm
      !!              by Knuth (1969, seminumerical algorithms).
      !!
      !! History :
      !!        !  07-09  (K. Mogensen)  Original code heavily based on IFS.
      !!----------------------------------------------------------------------
      !! * Function return
      REAL(wp) loc_sum_indep
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kn
      REAL(wp), DIMENSION(kn), INTENT(IN) :: &
         & pval
      !! * Local declarations
      REAL(wp), DIMENSION(3) ::&
         & zbuffl
      REAL(wp), DIMENSION(:), ALLOCATABLE :: &
         & zpsums, &
         & zperrs, &
         & zpcors, &
         & zbuffg, &
         & zp
      REAL(wp) :: &
         & zcorr,   &
         & zerr,    &
         & zolderr, &
         & zbeta,   &
         & zres
      INTEGER :: &
         & jj

      ! Check that the the algorithm can work 
      
      IF ( ( REAL( 2 * kn ) * EPSILON( zres ) ) >= 1.0 ) THEN

         CALL ctl_stop( 'kn is too large to guarantee error bounds' )

      ENDIF
      
      ALLOCATE( &
         & zp(MAX(kn,1)),                & 
         & zbuffg(1*SIZE(zbuffl)), &
         & zpsums(1),              &
         & zperrs(1),              &
         & zpcors(1)               &
         & )

      zolderr = HUGE(zerr)

      ! Copy the input array. This avoids some tricky indexing, at the
      ! expense of some inefficency.

      IF ( kn > 0 ) THEN

         zp(:) = pval(:)

      ELSE

         zp(1) = 0.0_wp

      ENDIF

      k_loop: DO

         ! Transform local arrays
         
         IF ( kn > 0 ) THEN

            CALL comp_sum ( zp, kn, zcorr, zerr )

         ENDIF

         ! Gather partial sums and error bounds to all processors

         zbuffl(1) = zp(MAX(kn,1))
         
         IF ( kn > 0 ) THEN

            zbuffl(2) = zerr
            zbuffl(3) = zcorr

         ELSE
            
            zbuffl(2) = 0.0_wp
            zbuffl(3) = 0.0_wp

         ENDIF

         zpsums(1) = zbuffl(1)
         zperrs(1) = zbuffl(2)
         zpcors(1) = zbuffl(3)

         ! Transform partial sums

         CALL comp_sum( zpsums, 1, zcorr, zerr )
         zerr  = zerr  + SUM(zperrs)
         zcorr = zcorr + SUM(zpcors)
         
         ! Calculate final result
         
         zres = zpsums(1) + zcorr

         ! Calculate error bound. This is corollary 4.7 from Ogita et al. 
         ! (2005)
         
         zbeta = zerr *( REAL( 2*kn, wp ) * EPSILON(zres) ) &
            &  /(1.0_wp - REAL( 2*kn, wp ) * EPSILON(zres) )

         zerr = EPSILON(zres) * ABS(zres) &
            & +(zbeta + ( 2.0_wp * EPSILON(zres) * EPSILON(zres) * ABS(zres) &
            &            +3.0_wp * TINY(zres) ) )

         ! Update the last element of the local array

         zp(MAX(kn,1)) = zpsums(nproc+1)

         ! Exit if the global error is small enough

         IF ( zerr < 4.0_wp * SPACING(zres) ) EXIT k_loop
         
         ! Take appropriate action if ZRES cannot be sufficiently refined.
         
         IF (zerr >= zolderr) THEN

            CALL ctl_stop( 'Failed to refine sum', &
               &           'Warning: Possiblity of non-reproducible results' )

         ENDIF

         zolderr = zerr
         
      ENDDO k_loop

      ! At this stage, we have guaranteed that ZRES less than 4*EPS
      ! away from the exact sum. There are only four floating point
      ! numbers in this range. So, if we find the nearest number that
      ! has its last three bits zero, then we have a reproducible result.

      loc_sum_indep = fround(zres)

      DEALLOCATE( &
         & zpcors, &
         & zperrs, &
         & zpsums, &
         & zbuffg, &
         & zp      &
         & )

   END FUNCTION loc_sum_indep


END MODULE mppsumtam
